/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#include "stdlib.h"
#include "stddef.h"
#include "string.h"
#include "ctype.h"
#include "stdio.h"

#include "tokenizer.h"

static char const *ptr, *first_ptr, *reader_ptr, *hlptr, *nextptr=0;
static char statement[128+31]; // debugstring 
static int  statelen, hexval, decval;
static char tok[4]; //

#define MAX_VARNUM 64 // 2*26 'a'-'Z' +#0-#9 +## + #@=loop 
#define MAX_NUMLEN 10 //:...i....1....;....2....;....3....;....4..

struct keyword_token {
  const char *keyword;
  short token;
};
/* static int   lastptr, last_token, last_line=1; */ 
static ubasic_token current_token = TOKENIZER_ERROR;
static unsigned short first_line=1, current_line = 1;
static short fn_stack_ptr=0; /*/ :)
/*:...i....1....;....2....;....3....;....4..*/
static short sep_memtoken = 0;
short debug = 0;
 
#define DEF_FN
#ifdef  DEF_FN
#define MAX_FN_STACK_DEPTH 16
static char const *fn_stack[MAX_FN_STACK_DEPTH];
static char const *fn_place[MAX_VARNUM];
// forward references
static char const *fn_insert(int fn); // replace token with def line
#endif

#define PUSH_POP  // && #def in ubasic.c ~169 too
#define MAX_PTR_STACK_DEPTH 32
static char const    *ptr_stack[MAX_PTR_STACK_DEPTH];
static unsigned short lin_stack[MAX_PTR_STACK_DEPTH];
static short ptr_stack_ptr=0;

static int looptrap=0, looperr=0;
int tokenizer_curptr(void) { int p=ptr-reader_ptr;
    if (p==looptrap) {looperr++; 
       if (looperr==3) debug=1;
       if (looperr==6) { tokenizer_next(); current_token = TOKENIZER_ENDOFINPUT;}
    }
    else { looptrap=p; looperr=0;}
  return p;
}

void bugstr(char c);
inline void bugstr(char c)
{ statement[statelen]=c; statelen++; statelen&=127; statement[statelen]='\0';}
//:.,....1....,....2....,....3....,....4....,....5....,....6....,....7....,.... */
// atoi(ptr);}//
inline long  tokenizer_value(void)  {return decval ? decval:hexval;}//atoi(ptr);}//
inline unsigned short tokenizer_line_number(void)
{  return current_line;}
inline short tokenizer_token(void)  {return current_token;}
inline short is_cr()    {return current_token==TOKENIZER_CR;}
inline short is_paren() {return current_token==TOKENIZER_LEFTPAREN;}
inline short is_comma() {return current_token==TOKENIZER_COMMA;}
inline short is_eof()   {return current_token==TOKENIZER_ENDOFINPUT;}
inline short is_token(short token)  {return current_token==token;}
short tokenizer_finished(void)
{  return *ptr == 0 || current_token == TOKENIZER_ENDOFINPUT || current_token == TOKENIZER_ERROR;}
inline void  tokenizer_sep(void)
{  sep_memtoken=current_token;  current_token=TOKENIZER_SEP;}

/*---------------------------------------------------------------------------*/

static const struct keyword_token keywords[] =
 {  // lowercase ONLY!!  // Case ignored in script now.
  {"not",                     TOKENIZER_LNOT},
  {"or",                      TOKENIZER_LOR},
  {"and",                     TOKENIZER_LAND},
  {"mod",                     TOKENIZER_MOD},
  {"xor",                 TOKENIZER_LXOR},
  {"nor",                 TOKENIZER_NOR}, // if a nor b : (a=0) and (b=0)
  {"nand",                TOKENIZER_NAND},  // if a nand b : (a!=0) and (b=0)
  {"abs",                 TOKENIZER_ABS}, // // -x -->> +x
  {"odd",                 TOKENIZER_ODD}, // // 0 2 4 6 8 => 1

  {"ifkey",               TOKENIZER_IFKEY}, // 1:is_pressed == if_key | -1 v-v |0
  {"if",                      TOKENIZER_IF},
  {"then",                    TOKENIZER_THEN},
  {"elseif",              TOKENIZER_ELSEIF},
  {"else",                    TOKENIZER_ELSE},
  {"endif",                   TOKENIZER_ENDIF},
  {"select",                  TOKENIZER_SELECT},
  {"case_else",               TOKENIZER_CASE_ELSE},
  {"case",                    TOKENIZER_CASE}, // autobreak
  {"end_select",              TOKENIZER_END_SELECT},
  {"exec",                    TOKENIZER_EXEC},
  {"force_analog_av",         TOKENIZER_FORCE_ANALOG_AV}, // has to come before "for"
  {"for",                     TOKENIZER_FOR},
  {"to",                      TOKENIZER_TO},
  {"next",                    TOKENIZER_NEXT},
  {"step",                    TOKENIZER_STEP},
  {"do",                      TOKENIZER_DO}, // ++ bubble behind then, else, case
  {"repeat",                  TOKENIZER_DO}, // LUAlike 
  {"until",                   TOKENIZER_UNTIL},
  {"while",                   TOKENIZER_WHILE},
  {"wend",                    TOKENIZER_WEND},
  {"break",               TOKENIZER_BREAK}, // // do for while & see data statement;....4..
  {"cont",                TOKENIZER_CONT}, // // jump to -> wend, until, next
  {"goto",                    TOKENIZER_GOTO}, // ++ calculated jump method: goto 100*i+x 
  {"gosub",                   TOKENIZER_GOSUB}, //++ gosub 100*i+x 
  {"return",                  TOKENIZER_RETURN},
  {"skip",                TOKENIZER_SKIP}, // >> Cracy Jump >> token
  {"data",                TOKENIZER_DATA}, //feeds var #0,..,..,#9
  {"shift",               TOKENIZER_SHIFT}, // << 
  {"let",                     TOKENIZER_LET}, // speeds up on 1. Position
  {"def",                 TOKENIZER_DEF_FN},
  {"fn",                  TOKENIZER_FN},
  {"rem",                     TOKENIZER_REM},
  {"cls",                     TOKENIZER_CLS},
  {"print_screen",            TOKENIZER_PRINT_SCREEN},
  {"print",                   TOKENIZER_PRINT},
  {"playsound",               TOKENIZER_PLAY_SOUND},
  {"is_key",                  TOKENIZER_IS_KEY}, // ++ is_key K "up" 100+x -->> K=100+x
  {"is_pressed",              TOKENIZER_IS_PRESSED}, // ++ is_pressed K "up" 100*x -->> gosub K data ....,...,...
  {"click",                   TOKENIZER_CLICK}, //  
  {"press",                   TOKENIZER_PRESS},  // ==... ("set")  for compat Lua
  {"release",                 TOKENIZER_RELEASE},   //
  ///{"shot",                    TOKENIZER_SHOOT}, // for compatibility
  {"shoot",                   TOKENIZER_SHOOT}, // ++ shoot ~ // rawproof?
  {"md_detect_motion",        TOKENIZER_MD_DETECT_MOTION},

// new string-related statements and functions
  {"abc$",                  TOKENIZER_ABC$},
  {"left$",                   TOKENIZER_LEFT$},
  {"mid$",                    TOKENIZER_MID$},
  {"right$",                  TOKENIZER_RIGHT$},
  {"mem$",                  TOKENIZER_MEM$},  // addr[,len]
  {"up$",                   TOKENIZER_UP$},
  {"low$",                  TOKENIZER_LOW$},
  {"read$",                 TOKENIZER_READ$}, // read$(A$[,"A/readme.txt"])
  {"rev$",                  TOKENIZER_REV$}, // reverse
  {"rep$",                  TOKENIZER_REP$}, // replace (a$,"?!")
  {"strol$",                TOKENIZER_STROL$}, // rotate left,1  abcd -> bcda
  {"stror$",                TOKENIZER_STROR$}, // rotate right,1 abcd -> dabc
  {"spc$",                  TOKENIZER_SPC$}, // spc$(3)="   "
  {"str$",                    TOKENIZER_STR$}, // str$(123*10+x)
  {"trim$",                 TOKENIZER_TRIM$}, // cut framing " '\"\10\r\n"
  {"arg$",                  TOKENIZER_ARG$}, // a$="a,b,c,d" ~> arg$(a$,0)="a"
  {"chr$",                    TOKENIZER_CHR$}, // =chr$(64,4) ~> @@@@

  {"args",                  TOKENIZER_ARGS}, // args(a$[,',']) ~> =4
  {"val",                     TOKENIZER_VAL},
  {"len",                     TOKENIZER_LEN},
  {"instr",                   TOKENIZER_INSTR},
  {"asc",                     TOKENIZER_ASC},
  {"open",                  TOKENIZER_OPEN},  // if open read$ a$
  {"close",                 TOKENIZER_CLOSE}, // ?read$~ close. //readfile
  {"file_browser",          TOKENIZER_FILE_BROWSER},
  {"textbox",               TOKENIZER_TEXTBOX},
// end of string additions

  {"in",                    TOKENIZER_IN}, // if x in 2,4,6,8... = if (x=2) or (x=4) or ...
  /* WARNING due to tokenizer limitation longest match must be first */
// GET  
  {"get_aflock",              TOKENIZER_GET_AFLOCK},
  {"get_af_frame",            TOKENIZER_GET_AF_FRAME},
  {"get_shooting",            TOKENIZER_GET_SHOOTING},
  {"get_av96",                TOKENIZER_GET_AV96},
  {"get_av",                  TOKENIZER_GET_USER_AV_ID}, //FOR COMPATIBILITY
  {"get_bv96",                TOKENIZER_GET_BV96},
  {"get_capture_mode",        TOKENIZER_GET_CAPTURE_MODE},
  {"get_config_value",        TOKENIZER_GET_CONFIG_VALUE},
  {"get_current_av96",        TOKENIZER_GET_CURRENT_AV96},
  {"get_current_tv96",        TOKENIZER_GET_CURRENT_TV96},
  {"get_current_base_sv96",   TOKENIZER_GET_CURRENT_BASE_SV96},
  {"get_current_delta_sv96",  TOKENIZER_GET_CURRENT_DELTA_SV96},
  {"get_day_seconds",         TOKENIZER_GET_DAY_SECONDS},
  {"get_digic",               TOKENIZER_GET_DIGIC},
  {"get_disk_size",           TOKENIZER_GET_DISK_SIZE},
  {"get_display_mode",        TOKENIZER_GET_DISPLAY_MODE},
  {"get_dof",                 TOKENIZER_GET_DOF},
  {"get_draw_title_line",     TOKENIZER_GET_DRAW_TITLE_LINE},
  {"get_drive_mode",          TOKENIZER_GET_DRIVE_MODE},
  {"get_exp_count",           TOKENIZER_GET_EXP_COUNT},
  {"get_flash_mode",          TOKENIZER_GET_FLASH_MODE},
  {"get_flash_ready",         TOKENIZER_GET_FLASH_READY},
  {"get_far_limit",           TOKENIZER_GET_FAR_LIMIT},
  {"get_free_disk_space",     TOKENIZER_GET_FREE_DISK_SPACE},
  {"get_eff_focal_length",TOKENIZER_GET_EFF_FOCAL_LENGTH},  // eq 35mm
  {"get_focal_length",        TOKENIZER_GET_FOCAL_LENGTH},
  {"get_focus_mode",          TOKENIZER_GET_FOCUS_MODE},
  {"get_focus_ok",            TOKENIZER_GET_FOCUS_OK},
  {"get_focus_state",         TOKENIZER_GET_FOCUS_STATE},
  {"get_focus",               TOKENIZER_GET_FOCUS},
  {"get_hyp_dist",            TOKENIZER_GET_HYPERFOCAL_DIST},
  {"get_imager_active",       TOKENIZER_GET_IMAGER_ACTIVE},
  {"get_iso_market",          TOKENIZER_GET_ISO_MARKET},
  {"get_iso_mode",            TOKENIZER_GET_ISO_MODE},
  {"get_iso_real",            TOKENIZER_GET_ISO_REAL},
  {"get_iso",                 TOKENIZER_GET_ISO_MODE}, //FOR COMPATIBILITY
  {"get_is_mode",             TOKENIZER_GET_IS_MODE}, // =get_IS_mode ; upper|lower IGNORED NOW in script
  {"get_jpg_count",           TOKENIZER_GET_JPG_COUNT},
  {"get_luminance",       TOKENIZER_GET_LUMINANCE},
  {"get_metering_mode",   TOKENIZER_GET_METERING_MODE},
  {"get_max_av96",            TOKENIZER_GET_MAX_AV96},
  {"get_min_av96",            TOKENIZER_GET_MIN_AV96},// get min. valid av96 value (widest aperture), or -1 if in playback or no iris
  {"get_min_stack_dist",      TOKENIZER_GET_MIN_STACK_DIST},
  {"get_movie_status",        TOKENIZER_GET_MOVIE_STATUS},
  {"get_mode",                TOKENIZER_GET_MODE},  // Returns 0 in recordmode, 1 in playmode
  {"get_nd_current_ev96",     TOKENIZER_GET_ND_CURRENT_EV96},// get the current ND value: 0 if out or not present, or nd_value if in
  {"get_nd_value_ev96",       TOKENIZER_GET_ND_VALUE_EV96},// get the exposure value of the ND filter, or 0 if not present
  {"get_nd_present",          TOKENIZER_GET_ND_PRESENT},
  {"get_near_limit",          TOKENIZER_GET_NEAR_LIMIT},
  {"get_platform_id",         TOKENIZER_GET_PLATFORM_ID},
  {"get_propset",             TOKENIZER_GET_PROPSET},
  {"get_prop",                TOKENIZER_GET_PROP},
  {"get_quality",             TOKENIZER_GET_QUALITY},
  {"get_raw_support",         TOKENIZER_GET_RAW_SUPPORT},
  {"get_raw_count",           TOKENIZER_GET_RAW_COUNT},
  {"get_raw_nr",              TOKENIZER_GET_RAW_NR},
  {"get_raw",                 TOKENIZER_GET_RAW},
  {"get_resolution",          TOKENIZER_GET_RESOLUTION},
  {"get_sd_over_modes",       TOKENIZER_GET_SD_OVER_MODES},
  {"get_sd_over_value",       TOKENIZER_GET_SD_OVER_VALUE},
  {"get_sv96",                TOKENIZER_GET_SV96},
  {"get_temperature",         TOKENIZER_GET_TEMPERATURE},
  {"get_tick_count",          TOKENIZER_GET_TICK_COUNT},
  {"gtc",                   TOKENIZER_GET_TICK_COUNT},
  {"get_time",                TOKENIZER_GET_TIME},
  {"get_tv96",                TOKENIZER_GET_TV96},
  {"get_user_av_id",          TOKENIZER_GET_USER_AV_ID},
  {"get_user_av96",           TOKENIZER_GET_USER_AV96},
  {"get_user_tv_id",          TOKENIZER_GET_USER_TV_ID},
  {"get_user_tv96",           TOKENIZER_GET_USER_TV96},
  {"get_video_recording",     TOKENIZER_GET_VIDEO_RECORDING},
  {"get_video_button",        TOKENIZER_GET_VIDEO_BUTTON},
  {"get_vbatt",               TOKENIZER_GET_VBATT},
  {"get_wheel",           TOKENIZER_GET_WHEEL},
  {"get_zoom_steps",          TOKENIZER_GET_ZOOM_STEPS},
  {"get_zoom",                TOKENIZER_GET_ZOOM},
  {"is_capture_mode_valid",   TOKENIZER_IS_CAPTURE_MODE_VALID},
  {"is_infinity",         TOKENIZER_IS_INFINITY},
  {"out_of_focus",        TOKENIZER_OUT_OF_FOCUS},
//SET
  {"set_aelock",              TOKENIZER_SET_AELOCK},
  {"set_aflock",              TOKENIZER_SET_AFLOCK},
  {"set_af_frame",            TOKENIZER_SET_AF_FRAME}, // readonly :(
  {"set_av96_direct",         TOKENIZER_SET_AV96_DIRECT},
  {"set_av96",                TOKENIZER_SET_AV96},
  {"set_av_rel",              TOKENIZER_SET_USER_AV_BY_ID_REL}, //FOR COMPATIBILITY
  {"set_av",                  TOKENIZER_SET_USER_AV_BY_ID}, //FOR COMPATIBILITY
  {"set_backlight",           TOKENIZER_SET_BACKLIGHT},
  {"set_capture_mode_canon",  TOKENIZER_SET_CAPTURE_MODE_CANON},
  {"set_capture_mode",        TOKENIZER_SET_CAPTURE_MODE},
  {"set_config_value",        TOKENIZER_SET_CONFIG_VALUE},
  {"set_draw_title_line",     TOKENIZER_SET_DRAW_TITLE_LINE},
  {"set_exit_key",            TOKENIZER_SET_EXIT_KEY},
  {"set_focus",               TOKENIZER_SET_FOCUS},
  {"set_iso_mode",            TOKENIZER_SET_ISO_MODE},
  {"set_iso_real",            TOKENIZER_SET_ISO_REAL},
  {"set_iso",                 TOKENIZER_SET_ISO_MODE}, //FOR COMPATIBILITY
  {"set_lcd_display",         TOKENIZER_SET_LCDDISPLAY},
  {"set_led",                 TOKENIZER_SET_LED},
  {"set_mf",                  TOKENIZER_SET_MF},
  {"set_metering_mode",   TOKENIZER_GET_METERING_MODE}, // readonly :(
  {"set_movie_status",        TOKENIZER_SET_MOVIE_STATUS},  
  {"set_nd_filter",           TOKENIZER_SET_ND_FILTER},
  {"set_prop",                TOKENIZER_SET_PROP},
  {"set_quality",             TOKENIZER_SET_QUALITY},
  {"set_raw_nr",              TOKENIZER_SET_RAW_NR},
  {"set_raw",                 TOKENIZER_SET_RAW},
  {"set_record",              TOKENIZER_SET_RECORD},
  {"set_remote_timing",       TOKENIZER_SET_REMOTE_TIMING},
  {"set_resolution",          TOKENIZER_SET_RESOLUTION},
  {"set_sv96",		          TOKENIZER_SET_SV96},
  {"set_tv96_direct",         TOKENIZER_SET_TV96_DIRECT},
  {"set_tv96",                TOKENIZER_SET_TV96},
  {"set_tv_rel",              TOKENIZER_SET_USER_TV_BY_ID_REL}, //FOR COMPATIBILITY
  {"set_tv",                  TOKENIZER_SET_USER_TV_BY_ID}, //FOR COMPATIBILITY
  {"set_user_av_by_id_rel",   TOKENIZER_SET_USER_AV_BY_ID_REL},
  {"set_user_av_by_id",       TOKENIZER_SET_USER_AV_BY_ID},
  {"set_user_av96",           TOKENIZER_SET_USER_AV96},
  {"set_user_tv_by_id_rel",   TOKENIZER_SET_USER_TV_BY_ID_REL},
  {"set_user_tv_by_id",       TOKENIZER_SET_USER_TV_BY_ID},
  {"set_user_tv96",           TOKENIZER_SET_USER_TV96},
  {"set_yield",               TOKENIZER_SET_YIELD},
  {"set_zoom_speed",          TOKENIZER_SET_ZOOM_SPEED},
  {"set_zoom_rel",            TOKENIZER_SET_ZOOM_REL},
  {"set_zoom",                TOKENIZER_SET_ZOOM},
  
  {"wait_click",              TOKENIZER_WAIT_CLICK},
  {"wheel_right",           TOKENIZER_WHEEL_RIGHT}, // ++ if wheel_right
  {"wheel_left",            TOKENIZER_WHEEL_LEFT}, // ++ if ...
  /*
  {"@debug",                  TOKENIZER_REM}, // opt. 1st-3rd! line to build a string, slows down a bit
  {"@title",                  TOKENIZER_REM},
  {"@chdk_version",           TOKENIZER_REM},
  {"@subtitle",               TOKENIZER_REM},
  {"@param",                  TOKENIZER_REM},
  {"@default",                TOKENIZER_REM},
  {"@range",                  TOKENIZER_REM},
  {"@values",                 TOKENIZER_REM},  //:...i....1....;....2....;....3....;....4..*/
  /// important
  // @@    is token to avoid gui_script.c processing #v=.. singleline below uBasic header !!!!!!!!!!!!!!!!!!

  {"md_get_cell_diff",        TOKENIZER_MD_GET_CELL_DIFF},
  {"md_get_cell_val",         TOKENIZER_MD_GET_CELL_VAL},
  {"md_af_led_control",       TOKENIZER_MD_AF_LED_CONTROL},
  {"md_detect_tick",          TOKENIZER_MD_DETECT_TICK},
  {"exit_alt",                TOKENIZER_EXIT_ALT},
  {"enter_alt",               TOKENIZER_ENTER_ALT}, 
  {"get_alt_mode",            TOKENIZER_GET_ALT_MODE}, 
  
  {"set_ev",                  TOKENIZER_SET_EV},
  {"get_ev",                  TOKENIZER_GET_EV},
  {"get_orientation_sensor",  TOKENIZER_GET_ORIENTATION_SENSOR},
  {"get_histo_range",         TOKENIZER_GET_HISTO_RANGE},
  {"autostarted",             TOKENIZER_SCRIPT_AUTOSTARTED},
  {"get_autostart",           TOKENIZER_GET_SCRIPT_AUTOSTART},
  {"set_autostart",           TOKENIZER_SET_SCRIPT_AUTOSTART},
  {"get_canon_raw_support",   TOKENIZER_GET_CANON_RAW_SUPPORT},
// return current canon image format as bitmask 1 = jpg, 2 = raw, 3 = raw+jpg 
  {"get_canon_image_format",  TOKENIZER_GET_CANON_IMAGE_FORMAT},
  {"set_canon_image_format",  TOKENIZER_SET_CANON_IMAGE_FORMAT},
  {"get_usb_power",           TOKENIZER_GET_USB_POWER},
  {"usb_force_active",        TOKENIZER_FORCE_USB_PRESENT},
  {"usb_sync_wait",           TOKENIZER_USB_SYNC_WAIT}, 
  {"shot_histo_enable",       TOKENIZER_SHOT_HISTO_ENABLE},
  {"set_console_layout",      TOKENIZER_SET_CONSOLE_LAYOUT},
  {"set_console_autoredraw",  TOKENIZER_SET_CONSOLE_AUTOREDRAW},
  {"show",                    TOKENIZER_SET_CONSOLE_AUTOREDRAW},
  {"console_redraw",          TOKENIZER_CONSOLE_REDRAW},
  {"swap_partitions",         TOKENIZER_SWAP_PARTITIONS},
  {"sleep",                   TOKENIZER_SLEEP}, 
  {"random",                  TOKENIZER_RANDOM},
  {"reboot",                  TOKENIZER_REBOOT},
  {"shut_down",               TOKENIZER_SHUT_DOWN},
  
  // APEX functions
  {"iso_to_sv96",             TOKENIZER_ISO_TO_SV96},
  {"sv96_to_iso",             TOKENIZER_SV96_TO_ISO},
  {"iso_real_to_market",      TOKENIZER_ISO_REAL_TO_MARKET},
  {"iso_market_to_real",      TOKENIZER_ISO_MARKET_TO_REAL},
  {"sv96_real_to_market",     TOKENIZER_SV96_REAL_TO_MARKET},
  {"sv96_market_to_real",     TOKENIZER_SV96_MARKET_TO_REAL},
  {"aperture_to_av96",        TOKENIZER_APERTURE_TO_AV96},
  {"av96_to_aperture",        TOKENIZER_AV96_TO_APERTURE},
  {"usec_to_tv96",            TOKENIZER_USEC_TO_TV96},
  {"tv96_to_usec",            TOKENIZER_TV96_TO_USEC},
  {"seconds_to_tv96",         TOKENIZER_SECONDS_TO_TV96},
  
   //:...i....1....;....2....;....3....;....4..
  {"peek",              TOKENIZER_PEEK},
  {"poke",              TOKENIZER_POKE},
  {"crc",               TOKENIZER_CRC},
  {"echo",              TOKENIZER_ECHO}, // not logged by print_screen
  {"write",             TOKENIZER_WRITE}, // Not console;  write [="A/Romlog.Log",]....
  {"strlen",            TOKENIZER_STRLEN}, // No output
  {"beep",              TOKENIZER_BEEP}, //:...i....1....;....2....;....3....;....4..
  {"varlist",           TOKENIZER_VARLIST}, // varlist[$]  vars<>0|"";
  {"mem_dump",          TOKENIZER_MEM_DUMP}, // 0xaddr,size
  {"lsec",              TOKENIZER_LSEC}, // linetime,looptime
  {"msec",              TOKENIZER_MSEC}, // since Start:
  {"sec",               TOKENIZER_SEKUNDE},//
  {"minute",            TOKENIZER_MINUTE},//
  {"hour",              TOKENIZER_STUNDE},//
  {"min",               TOKENIZER_MIN},//
  {"max",               TOKENIZER_MAX},//
  {"average",           TOKENIZER_AVERAGE},//
  {"sgn",               TOKENIZER_SGN},
  {"uns",               TOKENIZER_UNS},
  {"true",              TOKENIZER_TRUE}, // !0
  {"false",             TOKENIZER_FALSE}, // 0
  {"off",               TOKENIZER_OFF}, // 0
  {"on",                TOKENIZER_ON}, // 1  || on x=1,2,3 goto "a","b","c"
  {"swap",              TOKENIZER_SWAP}, // i,j ; k,l ; #7,m...
  {"loop",              TOKENIZER_LOOP}, // 1 = restart if error|eof // |for i=0|1,loop ...
  {"list",              TOKENIZER_LIST}, // Line prefix: show & do
  {"tronly",            TOKENIZER_TRONLY}, // 'list' !do >>// 'list' >>
  {"test",              TOKENIZER_TEST},
  {"tron",              TOKENIZER_TRON}, // 'list' >>
  {"troff",             TOKENIZER_TROFF},// <<
  {"bug",               TOKENIZER_BUG}, // preprogrammable Error // ?bug,bug gives linenr 
  {"debug",             TOKENIZER_DEBUG}, 
  {"clean",             TOKENIZER_CLEAN},  // #V..#7 // 42[..62]
  {"reset",             TOKENIZER_RESET},
  {"display",           TOKENIZER_UBASIC_READER},
  {"exit",              TOKENIZER_EXIT}, //== goto + break all loops
  {"is",                TOKENIZER_EQ},  // write a letter to Your cam...:)
  {"fi",                TOKENIZER_ENDIF},
  {"sqrt",              TOKENIZER_SQRT},
  {"even",              TOKENIZER_EVEN},
  {"memory",            TOKENIZER_MEMORY},
  
  {"exception",         TOKENIZER_EXCEPTION}, // Romlog on demand 20 secs later
  {"end",               TOKENIZER_END},  // ++ end X :show console X%10 sec.
  {NULL,                TOKENIZER_ERROR}
};

/*---------------------------------------------------------------------------*/

static short singlechar(void)
{
  if(*ptr == '\r') ptr++;
  switch (*ptr) {
  case  '=':  return TOKENIZER_EQ;
  case  '+':  return TOKENIZER_PLUS;
  case  '-':  return TOKENIZER_MINUS; // at linebegin a signal like '#'|'?'|'@' ++ speeds up loops...
  case  '*':  return TOKENIZER_ASTR;
  case  '/':  return TOKENIZER_SLASH;
  case  ',':  return TOKENIZER_COMMA;
  case  '%':  return TOKENIZER_MOD;
  case  '(':  return TOKENIZER_LEFTPAREN;
  case  ')':  return TOKENIZER_RIGHTPAREN;
  case '\n':  bugstr('@'); return TOKENIZER_CR;
  case  '<'://:...i....1....;....2....;....3....;....4..
		if(*(ptr+1) == '=') { ptr++; return TOKENIZER_LE; }
		if(*(ptr+1) == '>') { ptr++; return TOKENIZER_NE; }
		if(*(ptr+1) == '<') { ptr++; return TOKENIZER_SHL; }
		return TOKENIZER_LT;
  case  '>':
		if(*(ptr+1) == '=') { ptr++; return TOKENIZER_GE; } // "<>" == "!="
		if(*(ptr+1) == '<') { ptr++; return TOKENIZER_NE; }
		if(*(ptr+1) == '>') { ptr++; return TOKENIZER_SHR; }
		return TOKENIZER_GT;
  case  '!':
		if(*(ptr+1) == '=') { ptr++; return TOKENIZER_NE; }
		if(*(ptr+1) == '&') { ptr++; return TOKENIZER_NAND; }
		if(*(ptr+1) == '|') { ptr++; return TOKENIZER_NOR; }
		return TOKENIZER_NOT;
  case  '~':
		if(*(ptr+1) == '=') { ptr++; return TOKENIZER_NE; }
		return TOKENIZER_SEP;  // '~' allows multi statement lines
  case  '&':
		if(*(ptr+1) == '&') { ptr++; return TOKENIZER_LAND;}
		return TOKENIZER_AND;
  case  '|':
		if(*(ptr+1) == '|') { ptr++; return TOKENIZER_LOR;}
		return TOKENIZER_OR;
  case  '^':  return TOKENIZER_XOR;
  case  ';':  return TOKENIZER_SEMICOLON;
  case  '?':  return TOKENIZER_PRINT;  // Caefix, 6.4.19
// string addition
  case  '$':  return TOKENIZER_STRINGVARIABLE;
// end of string addition
  case  '@':
		if(*(ptr+1) == '(') { ptr++; return TOKENIZER_FIELD;}
		tokenizer_eol();             return TOKENIZER_CR;
  case '\\': // while (abs(r)>9) {r=r/10;}; '\-4210'=-4
		while (*(ptr+1) == '\\') {ptr++;} return TOKENIZER_CAP; // avoid RegExp ->.Log
  case  '`':  return TOKENIZER_SEKUNDE; // a sleep 1000 
  case  '.':
		 if(*(ptr+1) == '.') { ptr++; return TOKENIZER_TO;}
		 else return TOKENIZER_MOMENT; // a sleep (50)
  case  176:  return TOKENIZER_BEEP; //'' Major update! // Yes, Sir! // Wheres the joke?
  case  ' ':  return TOKENIZER_ERROR;
  }
  //return 0;
  return TOKENIZER_BUG;
}
/*---------------------------------------------------------------------------*/
static void wskip(void)  // eat all {whitespace}
{
  if (*ptr == '{') ptr++;
  while(*ptr == ' ' || *ptr == '\t'  || *ptr == '}') {ptr++;
    if (*ptr == '_') do ptr++; while (*ptr>' '); //  _ignore
    if (*ptr == '{') {//3....,....4.. comment 
	  do {ptr++;
		if (*ptr =='\n') current_line++;
	  } while (*ptr != '}' && *ptr);
      if (*ptr == '}') ptr++;
    }
    if (!*ptr) { return;}
    if (*ptr == '}') bugstr('}');  // exec  {   
  }
  nextptr=ptr;
  if (*ptr == '@' && *(ptr+1)!='(') {tokenizer_eol();}
}
/*---------------------------------------------------------------------------*/ 
 // Caefix, :...,....1....,....2....,....3....,....4..
void tokenizer_eol(void)
{
	if (current_token != TOKENIZER_CR)
    ptr=nextptr;
	while (*ptr != '\n' && *ptr) {++ptr;}
	current_token= (*ptr == '\n') ? TOKENIZER_CR : TOKENIZER_ENDOFINPUT;
  nextptr=ptr;
}
/* 
*/ 
void tokenizer_inc(void)
{ 
  ptr=nextptr;
	current_token=singlechar();
	if (current_token==TOKENIZER_BUG) ptr--;
  nextptr=ptr+1;
}

	// if ( t=tokenizer_char() != TOKENIZER_BUG)
short tokenizer_char(void)
{ const char *saveptr=ptr;
  int  t=current_token;
  if (t!=TOKENIZER_VARIABLE) t=singlechar();
  ptr=saveptr;
  return t;
}

static int hex2u()
{ int c=0;
  register unsigned u=0; 
  while (   (*nextptr>'/'    &&  *nextptr<':' )
        || ((*nextptr|32)>96 && (*nextptr|32)<'g')) {
    c++; if (c==9) return 0x80000000;
    u<<=4;
    u |= (*nextptr<64) ? (*nextptr-48):(*nextptr|32)-87;
    nextptr++;}
  return u;
}
/* 
int tokenizer_last(void)
{
  nextptr=ptr;
  ptr=lastptr;
  current_line=last_line;
  current_token=last_token;
  return last_token;
return ptr_stack_ptr;
} 
*/ 
/*---------------------------------------------------------------------------*/
static short get_next_token(void)
{
  struct keyword_token const *kt;
  register int t=TOKENIZER_ENDOFINPUT;
  register short  i,x;

  DEBUG_PRINTF("get_next_token(): '%s'\n", ptr);
  
  if (*ptr == 32 || *ptr == 10 || *ptr == 123) wskip();
  if (*ptr == 0) { return TOKENIZER_ENDOFINPUT; }
    // UnknStatement should have size, otherwise hanging-up in ubasic.c possible for some cases 
  //nextptr = ptr + 1; 
  if (*ptr=='[') { t=abs(atoi(nextptr));
	 if (!t || t>999 /*TOKENIZER_END*/) return TOKENIZER_BUG;
     else ptr++;  } 	// save s, count [token] in tokenizer.h //

  if (isdigit(*ptr)) { 
    hexval=0; decval=0;
	if (*ptr=='0' && (*(ptr+1)|32)=='x') { nextptr=ptr+2;
        hexval=hex2u();
		//while (strpbrk(nextptr,"0123456789abcdefABCDEF")==nextptr) nextptr++;
		if (nextptr>ptr+2) return TOKENIZER_NUMBER;
    }
    for (i = 0; i <= (MAX_NUMLEN); ++i) {
    if (isdigit(ptr[i])) {
      t=decval<<3; decval<<=1; decval+=t+ptr[i]-48;}
    else {
	  if (i > 0) {
	  nextptr = ptr + i;
      if (*nextptr==']') { ptr--; nextptr++;
		if (debug) { tok[0]='\0';
			x=statelen;
			if (x<120) {
			sprintf(tok, "%d%d%d",t/100,t%100/10,t%10);
			statement[x]='[';
			statement[x+1]='\0';
			strcat(statement,tok);
			statement[x+4]=']';
			statement[x+5]='\0';
			statelen+=5;}
		} return t;
      }
	  if (debug) {bugstr(ptr[0]);}
	  return TOKENIZER_NUMBER;
	} else {
	  DEBUG_PRINTF("get_next_token: error due to too short number\n");
	  bugstr(''); return TOKENIZER_ERROR;}
	DEBUG_PRINTF("get_next_token: error due to malformed number\n");
	bugstr('');   return TOKENIZER_ERROR;}
    } // for
    DEBUG_PRINTF("get_next_token: error due to too long number\n");
    bugstr('');   return TOKENIZER_ERROR;
  } // isdigit

  if (*ptr==':') {
		if (current_token != TOKENIZER_CR)          // a=a>0 ? a:-a
		{nextptr = ptr + 1; return TOKENIZER_LABEL;} //  ?"line 1",a:"line 2=";b,c:"line 3".
    // label
      nextptr = ptr;
      do {++nextptr;} while (*nextptr > ' ');
	//  while(*nextptr != ' ' && *nextptr != '\r' && *nextptr != '\n'
	//	 && *nextptr != '\t' && *nextptr != '\0');
	  return TOKENIZER_LABEL;
	} else if(*ptr=='"' || *ptr=='\'') { x=*ptr;
      nextptr = ptr;
      do { ++nextptr; }
	  while(*nextptr != x //&& *nextptr != '\r' && *nextptr != '\n'
		 && *nextptr != '\0'); // avoid open string error
	  if (  *nextptr == '\0' || nextptr-ptr>=255)
	     {bugstr('"'); return TOKENIZER_ERROR;}
      while(*nextptr == x) ++nextptr;
	  return TOKENIZER_STRING;
	
  } // else
  if((i=singlechar()) != TOKENIZER_BUG) { 
	  if (i == TOKENIZER_CR) { nextptr = ptr;
	} else {
      nextptr = ptr + 1; bugstr(*ptr);
	}
    return i;
	
  } else {	if (*(ptr+1) != '.') 
    for (kt = keywords; kt->keyword != NULL; ++kt)
	{   x=strlen(kt->keyword);
		// *:...;....1....;....2....;
		hlptr=kt->keyword; i=0;
		while ( i<x ) {
			if (*(ptr+i) == '.') {if(i+1<x) x=i+1; break;} // ++ Autovervollstndigung: 'Wai. 0' // 1. Treffer in Liste!!!
			if ( ((*(ptr+i) | (*(ptr+i)==95?95:32)) != *(hlptr+i)) ) //I  ++ Ignore_CASE  
				{x=0; break;}
			++i;
		}	 ///*/
		//:...;....1....;....2....;
		if (x) {
		nextptr = ptr + x;
		if (current_token==TOKENIZER_CR) {
			strcpy(statement,hlptr); statelen=x;}
		else if (debug>1) {
		   if (x<8) {strcat(statement,hlptr); statelen+=x;} else { // [
			x=statelen;
			if (x<120) { tok[0]='\0'; t= kt->token;
				sprintf(tok, "%d%d%d",t/100,t%100/10,t%10);
				statement[x]='[';
				statement[x+1]='\0';
				strcat(statement,tok);
				statement[x+4]=']';
				statement[x+5]='\0';
				statelen+=5;} } // ]
			} else bugstr('_');
        if (kt->token == TOKENIZER_REM) {
			if (*nextptr=='~')  // old uBasic ignores, new not.
				{nextptr++; return TOKENIZER_SEP;}
			tokenizer_eol(); return TOKENIZER_CR;}
		return kt->token;
		} // if x 
    } // for kt
  }

#ifdef DEF_FN
  if (*ptr == '') {nextptr = ptr + 1; return TOKENIZER_FN;}
#endif
  if((*ptr >= 'a' && *ptr <= 'z') || (*ptr >= 'A' && *ptr <= 'Z')) {
	if (debug>1) bugstr(ptr[0]);

// string addition
    if (*(ptr+1) == '$') { nextptr=ptr+2;
		bugstr('$');
	  return TOKENIZER_STRINGVARIABLE;}
// end of string addition

    nextptr = ptr + 1;
	if (current_token==TOKENIZER_VARIABLE)	{debug=2;}
    return TOKENIZER_VARIABLE; // each '#' avoids this way. 
  }
  debug=1;
  return TOKENIZER_ERROR;
}

/*---------------------------------------------------------------------------*/
void tokenizer_def_fn(void)  //  def fna .. fnZ , fn#0-fn#9
{ 
#ifdef DEF_FN
  int fn=-999;
  ptr = nextptr; wskip(); 
  if ((*ptr=='f' || *ptr=='F') && (*(ptr+1)=='n' || *(ptr+1)=='N')) {
    ptr+=2;
    fn=tokenizer_variable_num();
    if (fn<0) {debug=2; return;}
    ptr++; wskip(); 
    fn_place[fn&63] = ptr;
	tokenizer_eol(); 
  }
}

static char const *fn_insert(int fn) // replace token with def line
{
  if (fn_stack_ptr < MAX_FN_STACK_DEPTH) {
     if (fn_place[fn]) {
         fn_stack[fn_stack_ptr] = ptr;
         fn_stack_ptr++;
         ptr = fn_place[fn];
     } else { debug=2; current_token=TOKENIZER_DEF_FN;} // :)
  }
  return ptr;
#else 
  tokenizer_eol();
#endif
}
/*---------------------------------------------------------------------------*/
/*    tokenizer_line_number_inc();*/
/*---------------------------------------------------------------------------*/
void tokenizer_next(void)
{
 /* lastptr=ptr; last_line=current_line; last_token=current_token;
*/ 

  while (sep_memtoken) {ptr = nextptr;
	  current_token = sep_memtoken;
	  sep_memtoken =0;
	  if (current_token == TOKENIZER_CR) break;
	return;}
  
  if (current_token == TOKENIZER_CR) { // end of line
    statement[0]='\0';
	do {    // move to next line, and skip all following empty lines as well
		ptr++; current_line++; wskip();}
	while (*ptr == '\n');
    if (*ptr=='-') { nextptr=ptr+1; // signal to speed up gotos in long scripts
        current_token=TOKENIZER_MINUS; return;}
  
 } else { ptr = nextptr; wskip(); }
  
        DEBUG_PRINTF("tokenizer_next: %p\n", nextptr);
  nextptr=ptr+1;
  if (*ptr == '#') { //#a=a //++ speed up jumps & loops in long scripts!, Caefix 
// string addition
      if (*(ptr+2)=='$' || (*(ptr+1)=='$' && *(ptr+2)=='('))
	  { do nextptr++; while (*nextptr!='$'); nextptr++;
		current_token = TOKENIZER_STRINGVARIABLE;
      } else
// end of string addition
        current_token = TOKENIZER_VARIABLE; }
  else  current_token = get_next_token();
  
#ifdef DEF_FN
  if (current_token == TOKENIZER_FN) { ptr=nextptr;
      int fn=tokenizer_variable_num();
      ptr++; wskip(); bugstr('[');
      ptr = fn_insert(fn); wskip();
      current_token = get_next_token();
  }
  if (fn_stack_ptr && current_token == TOKENIZER_CR) { // end of fn_
      fn_stack_ptr--; bugstr(']'); // continue invoking line
      ptr = fn_stack[fn_stack_ptr]; wskip();
      current_token = 0;
      current_token = get_next_token();
  }
#endif
        DEBUG_PRINTF("tokenizer_next: '%s' %d\n", ptr, current_token);

  tokenizer_curptr(); // -INF_LOOPERROR
  return;
}
/*---------------------------------------------------------------------------*/
void tokenizer_exec(void)  /// switch multilinecomment  {to function		}
{  /// then exec { else "ignore"  block
while(*ptr != '{' &&  *ptr != '\n' && *ptr != '\0') ptr++;
  if (*ptr == '{') {ptr++; ///{3....,....4.. first '}' closes All '{' !!
       bugstr('{');}
  nextptr=ptr;
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
void tokenizer_skip(short token)
{
	while (current_token != token &&
	current_token != TOKENIZER_ENDOFINPUT)
	{tokenizer_next();}
}

/*---------------------------------------------------------------------------*/
void climb_cr()
{
	do ptr--; while (*ptr!='\n');
	current_line--; current_token = TOKENIZER_CR;
}

void tokenizer_push()
{
#ifdef PUSH_POP
  if (ptr_stack_ptr < MAX_PTR_STACK_DEPTH) {
      ptr_stack_ptr++;
      ptr_stack[ptr_stack_ptr] = nextptr;
      lin_stack[ptr_stack_ptr] = current_line;
  } else debug=2;  // :)
#endif
}

void tokenizer_pop(int p)
{
#ifdef PUSH_POP
  if (p==0x1000) {ptr_stack_ptr=0; return;}
  if (!ptr_stack_ptr) {debug=2; bugstr(169); tokenizer_eol();
    //  current_token=TOKENIZER_ENDOFINPUT;
      return;}
  if (p!=1) {
      current_line = lin_stack[ptr_stack_ptr];
      ptr = ptr_stack[ptr_stack_ptr];
      if (p==-1 && *ptr!='\n') climb_cr(); //  wend -1-> while 
  }
  if (p!=0) ptr_stack_ptr--;
  /* swallow the CR that would be read next */
  tokenizer_next();
#endif
}
/*---------------------------------------------------------------------------*/
void tokenizer_begin(void)
{
	tokenizer_eol();
	tokenizer_next();
	nextptr=ptr-1;
	first_line=current_line-1;
	first_ptr=ptr-1;
#ifdef DEF_FN
    fn_stack_ptr=0;
#endif
}
/*---------------------------------------------------------------------------*/
const char *strfind(const char *ptr, const char *find)
{	register unsigned short len=strlen(find);
	while (*ptr) {
      while (*ptr && *ptr!=find[0]) ptr++;
      if (!strncmp(ptr,find,len)) return ptr;
      ptr++; }
    return 0;
}

void tokenizer_init(const char *program)
{  
  if (current_line==1) { // no need to read "twice", nice. Caefix 
  ptr = program; reader_ptr = program; debug=0;
#ifdef DEF_FN
    int i;
    for (i=0; i<MAX_VARNUM; ++i) {fn_place[i]=0;}
#endif
	tok[3]='\0'; 
	statement[128]='\0';
	statelen=0;
	nextptr=ptr+1;
    tokenizer_begin();
	ptr_stack[0]=reader_ptr;
    lin_stack[0]=1;
	hlptr=strfind(program,"@debug");
	if (hlptr>ptr) { debug=atoi(hlptr+7);
		if (debug<0) debug=1; }
	hlptr=strfind(program,"@subtitle");
    if (hlptr>=program) {ptr=hlptr+10;
       tokenizer_string(subtitle,123);} 
    else strcpy(subtitle,":...,....");
    hlptr=strfind(program,"@title");
    if (hlptr>=program) {ptr=hlptr+7;
       tokenizer_string(statement,123);} 
    else strcpy(statement,"unk @title");
    hlptr=0;
  }
  current_line=first_line;
  current_token=TOKENIZER_CR;
  nextptr=first_ptr;
  ptr=first_ptr;
  if (hlptr) tokenizer_next();
}
/*---------------------------------------------------------------------------*/
short tokenizer_variable_num(void)// varfactor only &63 CAREful
{
short r=-999; //, str=current_token==TOKENIZER_STRINGVARIABLE;
 
// string additions
 if (*ptr=='#' || *ptr=='$') { ptr++;
	 if (*ptr=='$') ptr++;
// end of string additions
	 if (*(ptr)=='(') //  let #(i+x)=... |data #(i+x)..., 
		{ return -TOKENIZER_LEFTPAREN;}
	 //ptr++; 
	 if (current_token==TOKENIZER_VARIABLE) nextptr=ptr+1;
	 if (*ptr=='#') {     return MAX_VARNUM-2;}
	 if (*ptr=='@') {
		ptr++; nextptr++; return MAX_VARNUM-1;} // == loop 
	 if (*ptr=='+') {
		ptr++; nextptr++; return MAX_VARNUM;} // 0x7fffffff
	 if (isdigit(*ptr)) {
		return *ptr-48+(MAX_VARNUM-12); }
	};
 while (*ptr=='%') { //  %0..%9 ,Args like in Batchfiles; for data and shift
	ptr++; nextptr=ptr+1;
    if (!isdigit(*ptr)) {r=-1; break;}
    r = *ptr-48;
    if (r>9) {r=-1; break;}
    return r+MAX_VARNUM-12; 
    };
 // else default:current_token=TOKENIZER_ERROR;
 if (r==-999) r = *ptr - (*ptr>='a'?'a':('A'-26));
 if (r>=MAX_VARNUM||r<0) { r=-1;
    current_token=singlechar();};
// string additions
 if (*(ptr+1)=='$') {current_token=TOKENIZER_STRINGVARIABLE;
    nextptr=ptr+2;  }
// end of string additions
return r;
}
/*---------------------------------------------------------------------------*/
void tokenizer_string(char *dest, int len)
{
  char *string_end;
  int s,string_len;
  
  s=*ptr=='"' || *ptr=='\'';
  if (len==123) string_end = strpbrk(ptr + 1,"\r\n\0"); else
  if (*ptr=='"') {string_end = strpbrk(ptr + 1,"\"\0"); // \r\n
	   string_end--;}
  else {string_end = strpbrk(ptr + 1,"'\0"); // \r\n
	   string_end--;}
  if (string_end == NULL) {return;}
  string_len = string_end - ptr;
  if (len < string_len) string_len = len;
  memcpy(dest, ptr + s, string_len);
  dest[string_len&255] = '\0';
  
  if (len!=123) {
   //  nextptr=string_end+1;
	 if (debug>1 && string_len+statelen<123)
		{strcat(statement,dest); statelen+=string_len;}
	 else bugstr('$');
  }
}
/*---------------------------------------------------------------------------*/
int tokenizer_stringlookahead(){ //short token
// return 1 (true) if next 'defining' token is string not integer
 int token = current_token;
 if (token>TOKENIZER_CHR$) return -token;

static char const *saveptr, *savenextptr;
static int savetoken, saveline;
int si=-1;
  saveptr = ptr; savenextptr = nextptr;
  savetoken = current_token; saveline = current_line;
  
  while (si <0) {
  switch(token)
    {
    case TOKENIZER_CR: if (si>-2) si=0;  break;
    case TOKENIZER_MOD: tokenizer_inc(); break;
    case TOKENIZER_LEFTPAREN:
    case TOKENIZER_PLUS:          si=-2; break;
    case TOKENIZER_STRING:        si=1;  break;
    default:
        if (token >= TOKENIZER_STRINGVARIABLE
         && token <= TOKENIZER_CHR$) {si = 1;}
        else si = 0; // numeric or logic function
    }
    if (si < 0) {tokenizer_next();
       token = current_token;}
  }
  ptr = saveptr; nextptr = savenextptr;
  current_token = savetoken; current_line = saveline;
  tokenizer_curptr();
  return si==1 ? 1:-token; 
}
/*---------------------------------------------------------------------------*/
void tokenizer_statement(char *dest, int len)
{
  statelen=strlen(statement)&127;
  if (statelen<len) len=statelen; 
  memcpy(dest, statement, len);
  dest[len] = '\0';
}
/*---------------------------------------------------------------------------*/
void tokenizer_label(char *dest, int len)
{
  char *string_end;
  int string_len;
  
  if (tokenizer_token() != TOKENIZER_LABEL) {
     return; }
  // allow string \r, \n, tabulation and space to end labels
  string_end = strpbrk(ptr + 1, " \t\r\n\0");

  if(string_end == NULL) {return;}
  string_len = string_end - ptr - 1;
  if (len < string_len) {string_len = len;}
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
  tokenizer_eol();
}
/*---------------------------------------------------------------------------*/

unsigned short tokenizer_reader_pos(int lines)
{ short x=0;
  short y=(lines>0) ? 1:-1; 
  lines=lines*y;
  short l=lines;
  if (ptr+y>reader_ptr) { 
  while (lines>0) { x++; ptr+=y;
    if (*ptr=='\n' || x==44) { l--; lines--;
       if (*ptr=='\n') {current_line+=y; x=0;}
       if (x==44) {l--; x=0;}
    } 
    if (!l || ptr==reader_ptr || !*ptr) break;
    } // while
  if (*ptr=='\n') {ptr++; if (y<0) current_line++;}
  nextptr=ptr;
  }
  current_token = (*ptr==0) ? TOKENIZER_ENDOFINPUT : TOKENIZER_UBASIC_READER;
  return (*ptr==0) ? 0:lines;
}

/*---------------------------------------------------------------------------*/
void tokenizer_error_print(void)
{
  DEBUG_PRINTF("tokenizer_error_print: '%s'\n", ptr);
}
/*---------------------------------------------------------------------------*/
void tokenizer_end(void)
{
  first_line=1;
  current_line=1;
  nextptr=0;
  debug=0;
}

/* Ubasic.bat
if %~x1?==? (cmd /T:4E/Kdir/b/ad/s/on %1) else (cmd /T:1F/K type %0 %*|find /N "statement"|find /V "DEBUG")
explorer D:\hostluaPortable\CHDK-Shell\trunk\trunk-6087\lib\ubasic
/ *----------- 818 ----- 5905 ----* /
/*------------1061 ----- 6087 ----*/