/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 /*
 * Modified to support simple string variables and functions by David Mitchell
 * November 2008.
					http://zenoshrdlu.com/kapstuff/zsubasic.html
 * Changes and additions are marked 'string additions' throughout
 */
 
#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#include "camera_info.h"
#include "shot_histogram.h"
#include "levent.h"
#include "script.h"
#include "console.h"
#include "conf.h"
#include "lang.h"
#include "gui_lang.h"
#include "action_stack.h"
#include "shooting.h"
#include "sd_card.h"
#include "backlight.h"
#include "battery.h"
#include "temperature.h"
#include "clock.h"
#include "file_counter.h"
#include "lens.h"
#include "debug_led.h"
#include "keyboard.h"
#include "usb_remote.h"
#include "shutdown.h"
#include "sound.h"
#include "motion_detector.h"

#include "string.h"
#include "time.h"
#include "math.h"
#include "gui_fselect.h"
#include "gui_tbox.h"

#ifdef UBASIC_TEST
#include "camera_functions.h"
Conf conf;                      // TODO: initialisation
_cam_info camera_info;          // TOTO: initialisation
#else
#include "script_key_funcs.h"
#include "properties.h"
#endif

#include "ubasic.h"
#include "tokenizer.h"


static char const *program_ptr;
// Forward references
//static long ubasic_get_variable(short varnum);
// Variable number to store return values from Action Stack functions (e.g. motion_detect, shoot)
static unsigned short ubasic_as_ret_var_num;
static short ubasic_get_key_arg();

#define INCLUDE_OLD_GET__SYNTAX

#define MAX_VARNUM 64 // ++ #0 - #9 + ## + #@=loop 
static int variables[MAX_VARNUM+1];
#define MAX_FIELD 257 // @(x+y...)+=...
static int field[MAX_FIELD+1];

// string additions      // 8kb
#define MAX_STRINGVARLEN 0x00ff
#define MAX_BUFFERLEN    0x3000
#define GBGCHECK         0x2e00
#define MAX_SVARNUM  MAX_VARNUM //
static char*stringvariables[MAX_SVARNUM+1];
static char stringbuffer[MAX_BUFFERLEN+1];
static char sbuf[MAX_STRINGVARLEN+1]; // for spc$ & read$
static int  freebufptr = 0;
static char const abcdef[MAX_SVARNUM+1]="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789#@\0";
//                                      :0b...,....1....,....2....,....3....,....4....,....5....,....6..@ 
static void *addr;
FILE  *readfile=NULL;
// end of string additions

#define MAX_STRINGLEN 128// Caefix, :...,....1....,....2....,....3....,....4..
static char string[MAX_STRINGVARLEN+1];
static char buf[MAX_STRINGVARLEN+1]; //127 printbuffer + spare to truncate
static char writefile[MAX_VARNUM+1]; // write appends to "A/CHDK/LOGS/Log_%04d.Log" || write [=["A/Romlog.log"],]....
static char scripttitle[65]=" :) ";
       char subtitle[MAX_VARNUM+1];

#define MAX_IF_STACK_DEPTH 12
static short if_stack[MAX_IF_STACK_DEPTH];
static int if_stack_ptr;

struct select_state {
  int select_value;
  short case_run;
};
#define MAX_SELECT_STACK_DEPTH 5
static struct select_state select_stack[MAX_SELECT_STACK_DEPTH];
static short select_stack_ptr;

#define MAX_GOSUB_STACK_DEPTH 10
static unsigned short gosub_stack[MAX_GOSUB_STACK_DEPTH];
static short gosub_stack_ptr;

#define MAX_BREAK_STACK_DEPTH 36 // Caefix, :...,....1....,....2....,....3....,....4..
static unsigned short break_stack[MAX_BREAK_STACK_DEPTH];
static short break_stack_ptr;

#define MAX_DO_STACK_DEPTH 8
static unsigned short do_stack[MAX_DO_STACK_DEPTH];
static short do_stack_ptr;

#define MAX_WHILE_STACK_DEPTH 8
static unsigned short while_stack[MAX_WHILE_STACK_DEPTH];
static short while_stack_ptr;

struct for_state {
  unsigned short line_after_for;
  unsigned short for_variable;
  int to;
  int step;
};
#define MAX_FOR_STACK_DEPTH 5
static struct for_state for_stack[MAX_FOR_STACK_DEPTH];
static short for_stack_ptr;

static short ended, end_wait, loop, trace, traconly, debug, ifkey=0, on_go, go_fwd, jogdial, logfile=0;
static int lsec, sec=0;
static short str_len, pause, mem_key=0;
static int stm_ctr=0, stm_ptr=0, subj_dist=0;
static unsigned start_sd;
static unsigned first_tick, start_tick, program_size;
static char const on=1, off=0; // ++ on | off 
/*---- 164 ------------------------------------------------------------------*/

// Forward references
#define PUSH_POP  // && #def in tokenizer.c ~74 too
static int factor(void);
static int expr(void);
static int relation(void);
static int varfactor(unsigned short n);
static unsigned get_time(int tmode);
static unsigned wait_click_statement(void);
static void line_statement(void);
static void statement(void);
static void do_statement(void);
static void data_statement(void);
static void print_statement(short token);
static void break_statement(void);
static void moment_statement(unsigned d);
static short  set_mf_statement(void);

static unsigned stopwatch(int v); 
static int beep(void) { play_sound(4); stopwatch(stm_ptr);return tokenizer_curptr();}

//unsigned unsign(unsigned unsig);
//inline unsigned unsign(unsigned unsig) {return unsig;}

// string additions
static const char nullstring[] = "\0"; 
static void  svar_init(void);
static char* store(const char *);
static char* sexpr(int);
///elf2flt unknown segment 1 for name: 'store'
///gmake[1]: *** [.o2/ubasic.flt] Error 6
///gmake: *** [all-recursive] Error 1
//static char* sconcat(char *, const char *);
//static char* sleft(char *, int); 
//static char* sright(char *,int);
//static char* smid(char *, int, int);
//static char* memstr(const char *, int);
//static char* sstr(int);
//static char* schr(int);
//static int sinstr(int, const char*, const char*);
static int garbage_collect();
// end of string additions
static char* file_browser();
static char* textbox();
static int   svar_selected=0;
static char  string_selected[] = "A/camnotes.txt";

/*
ubasic scheduling controls
ubasic returns control to kbd_task as soon as either of the following is is exceeded
defaults is 1 ubasic line per kbd_task cycle, mimicing old behavior
*/
#define YIELD_MAX_LINES_DEFAULT 1
#define YIELD_MAX_MS_DEFAULT 10
// maximum number lines before yielding (labels, blank lines and REMs excluded)
static unsigned short yield_max_lines;
// maximum number of ticks before yielding
static unsigned short yield_max_ms;
// set flag to immediately transfer execution flow to main CHDK to process external action
// TODO a better approach to this might be to check if AS_RUN is on top of the action_stack
static short flag_yield;

short ubasic_error;
const char *ubasic_errstrings[UBASIC_E_ENDMARK] =
{
    "No err",
    "Parse err",
    "Unk stmt",
    "Unk key",
    "Unk label",
    "gosub: Stack ful",
    "bad return",
    "if: Stack ful",
    "bad endif",
    "select: Stack ful",
    "bad end_select",
    "for: Stack ful",
    "bad next",
    "do: Stack ful",
    "bad until",
    "while: Stack ful",
    "bad wend",
    "Unk err" 
};

/*---------------------------------------------------------------------------* /
void extern_void(void (*func)(void))
{
    func();
}
/*----- 206 -----------------------------------------------------------------*/
unsigned short ubasic_linenumber();
inline unsigned short ubasic_linenumber()
{  return tokenizer_line_number();}

/*---------------------------------------------------------------------------*/
static void reset(void)
{
stopwatch(program_size);
  ended = 0;
  trace= off; // [tron | troff ]
  traconly=off; // tronly
  jogdial=0;
  flag_yield = 0;
  for_stack_ptr = gosub_stack_ptr = while_stack_ptr = do_stack_ptr = if_stack_ptr = select_stack_ptr = 0;
  break_stack_ptr = 0;
  tokenizer_pop(0x1000);
  ubasic_error = UBASIC_E_NONE;
  tokenizer_init(program_ptr);
  if (readfile!=NULL) {fclose(readfile);
      readfile =NULL;}
  if (!stm_ctr) {
     tokenizer_statement(scripttitle,45);}
  variables[MAX_VARNUM-1]=loop;
  string[MAX_STRINGLEN]='\0';
  buf[MAX_STRINGLEN]='\0';
  writefile[44]='\0';
  string[0]='\0';
  sbuf[0]='\0';
  buf[0]='\0';
  str_len=0;
  go_fwd=0;
  on_go=0;
  pause=0;
  lsec=0;
}

/*----- 233 -----------------------------------------------------------------*/
int ubasic_init(const char *program, __attribute__ ((unused))int is_ptp)
{ int i; 
  program_ptr = program;
  program_size = strlen(program);
//  msleep(3000);play_sound(5);
  first_tick = get_tick_count();
  start_sd = GetFreeCardSpaceKb();
  yield_max_ms = YIELD_MAX_MS_DEFAULT;
  yield_max_lines = YIELD_MAX_LINES_DEFAULT;
  for (i=0; i<=MAX_FIELD; i++) field[i]=0;
  tokenizer_end(); // clean some stuff evtl. left from last runned script. Without Vars A..#9 !!
  loop = off; // reset if error or eof : "loop 1|x" // if loop ...
  debug= off;
  end_wait=0;
  reset();
  svar_init(); // string addition
  stm_ctr=0;
  beep();
 // moment_statement(11);
  // Keyboard init
  kbd_get_autoclicked_key();
  camera_info.state.kbd_last_clicked = 0;
  camera_info.state.kbd_last_checked_time = first_tick;
  return 1;
}
/*---------------------------------------------------------------------------*/
void reader_init()  // TOKENIZER_UBASIC_READER
{	// limits: linelen 127, 65432 lines, memory
	ended=0; pause=-1; traconly=100;
    play_sound(0);
    console_set_layout(0,0,45,12);
    tokenizer_reader_pos(-1);
    if (stm_ctr==0)  // not ubasic script
	   {go_fwd=-2; pause=5;} 
}
/*---------------------------------------------------------------------------*/
static void svar_init() // string addition
{
   int i;
   freebufptr = 0;
 //  for (i=0; i<MAX_VARNUM; i++) Nice to give to next .bas script
  //    variables[i] = 0;
   for (i=0; i<=MAX_SVARNUM; i++) 
       stringvariables[i] = store(nullstring);
   memset(stringbuffer,'\0',MAX_BUFFERLEN);
   memset(sbuf,'\0',MAX_STRINGVARLEN);
}
/*---------------------------------------------------------------------------*/
static void accept(short token)
{
  if (is_eof() || ended) {
      trace=off; traconly=off; debug=1;
      token=TOKENIZER_IS_INFINITY; ended=1;
      ubasic_error = UBASIC_E_PARSE; return;
  }
  if (token) {//  0 = Joker
  short t=tokenizer_token();
  if (token !=t) {
     if (token<0) {  if (token == -t) {tokenizer_next();}
         return;   } // == if (is_token(-token)) tokenizer_next();
    DEBUG_PRINTF("Token not what was expected (Expected %d, got %d)\n", token, t);
    tokenizer_error_print();
	sprintf(buf, "%d->Expected %d,got %d",
            tokenizer_curptr(),token, t);
	script_console_add_error((long)buf); buf[0]='\0';
	moment_statement(99);
	if (!trace||tokenizer_finished()) {
		ended = 1;
		ubasic_error = UBASIC_E_PARSE;}
//	if (t==TOKENIZER_CR) tokenizer_sep();
    return;  }
  DEBUG_PRINTF("Expected %d, got it\n", t);
  }//  0 
  tokenizer_next();
 // if (is_eof()) {if (!trace) {ended = 1; traconly=off;} return;}
}
/*---------------------------------------------------------------------------*/
static void accept_cr()
{
	if (ended) return;
	short e=0;
	short t=tokenizer_token();
  while (t != TOKENIZER_CR) {
	switch(t) { // 
	  case TOKENIZER_SEP:   if (!e) return;  // Caefix, 6.4.19)
	  case TOKENIZER_COMMA:
		   tokenizer_next(); 
		   while (is_cr()) {accept(0);} break;
	  case TOKENIZER_ELSE:
	  case TOKENIZER_ELSEIF:  e=9;  break;
	  case TOKENIZER_END: tokenizer_eol(); // Compat Lua.
           if (!e) {tokenizer_sep();} break;
	  case TOKENIZER_DO:
		 {tokenizer_eol(); do_statement(); break_statement(); tokenizer_sep();}
	  case TOKENIZER_ENDOFINPUT:
		 {ended = 1; traconly=off;  return;}
	  case TOKENIZER_LET:	if (!e) return;  // experimental let
	} // 
	tokenizer_next(); 
	t=tokenizer_token();
  } // while
}
/*---------------------------------------------------------------------------*/
static char eos()
{
	short t = tokenizer_token();
	if (t == TOKENIZER_CR
	 || t == TOKENIZER_SEP
	 || t == TOKENIZER_THEN
	 || t == TOKENIZER_ELSE
	 || t == TOKENIZER_ELSEIF
	 || t == TOKENIZER_END) //  (compat Lua: if o==0 then O=0 end) 
	 { return 1; }
	return 0;
}
/*---------------------------------------------------------------------------* /
static void accept_cr()
{
    while(tokenizer_token() != TOKENIZER_CR &&
	    tokenizer_token() != TOKENIZER_ENDOFINPUT) tokenizer_next();
    accept(TOKENIZER_CR);
}
/ *---------------------------------------------------------------------------*/
static int shooting_get_near_limit_of_acceptable_sharpness()
{
  shooting_update_dof_values();
  return camera_info.dof_values.near_limit;
}
static int shooting_get_far_limit_of_acceptable_sharpness()
{
  shooting_update_dof_values();
  return camera_info.dof_values.far_limit;
}
static int shooting_get_depth_of_field()
{
  shooting_update_dof_values();
  return camera_info.dof_values.depth_of_field;
}
static int shooting_get_min_stack_distance()
{
  shooting_update_dof_values();
  return camera_info.dof_values.min_stack_distance;
}

/*---------------------------------------------------------------------------*/
static void writeln(char *buf) // =~ print to file
{
  if (logfile) { 
    FILE *fh = fopen(writefile,"ab");
    if (!fh) {
		sprintf(writefile, "A/uBerror.Log");
		FILE *fh = fopen(writefile,"ab");
		if(!fh) return;}
	fwrite((char *)buf,strlen(buf),1,fh);
	fwrite((char *)"\n",1,1,fh);
	fclose(fh); }
}
/*---------------------------------------------------------------------------*/
static char* readln(FILE *fh, int len)
{
  int eof=FALSE, cut=len<0;
  static char* s=sbuf; *s=''; s[1]='\0';
  char* c=addr;
  register int l=0;
  if (fh) { 
    if (cut) len=-len;
    len%=MAX_STRINGVARLEN;
    while (l<len && !eof) {
      fread(c,1,1,fh);
      if (*c == '\n') {
          if (cut) {*s='\0'; cut=2; break;}
          if (s[l-1]=='\r') l--;
          if (l<2 || (l>2 && // read table for arg$ /args
          ((s[l-1]!=',' && s[l-2]!=',')))) break;
      }
      s[l]=*c;
      if (*c && !feof(fh)) l++; else eof=TRUE; /* eof? binary? */
    }
    s[l]='\0';
    if (eof) {fclose(fh); readfile=NULL;} // 
    if (cut==2 && fh) readln(fh,MAX_STRINGVARLEN-1);
    if (*s) {while (l && s[l]<33) {s[l]='\0'; l--;} l++;}
    else l=0;
  }
  s[l]='\0';
  return s;
}

/*---------------------------------------------------------------------------*/
static short varlist(void)  // or varlist$
{
short i=0, n=0, s=0;
int   val;
static char buf[MAX_STRINGLEN];
    sprintf(buf, "%s--> %d/%d", scripttitle, beep(), program_size);
    script_console_add_error((long)buf);

int str = is_token(TOKENIZER_STRINGVARIABLE);
if (str) tokenizer_next();
  for (i=0; i<MAX_VARNUM; ++i) {
    buf[0]=0;//  string additions
	if (*(stringvariables[i])) s++;
	if (variables[i]) n++;
    if (str) val = strlen(stringvariables[i]); else val = variables[i];
    if (val) {
        if (str) {strncpy(string,stringvariables[i],MAX_STRINGLEN-5);// truncate
             sprintf(buf, "#%c$='%s'", abcdef[i],string);}
        else sprintf(buf, "#%c=%d",    abcdef[i], val);
        script_console_add_error((long)buf);
    }
  }
    //i = conf.platformid;
    sprintf(buf, "@line:%d :#=%d:$=%d,%d",tokenizer_line_number(),n, s,-garbage_collect());//,"  PID_",i);
    script_console_add_error((long)buf);
    moment_statement(96);
    get_time(7);
    get_time(8);
    accept_cr();
return n; // count vars !=0
}
/*---------------------------------------------------------------------------*/
void        ubasic_set_variable(int varnum, int value);
inline void ubasic_set_variable(int varnum, int value) {
//  if(varnum >= 0)// && varnum < MAX_VARNUM) 
    variables[varnum&63] = value; //else ended=1;
}

/*/ string additions here
/ *---------------------------------------------------------------------------*/
void        ubasic_set_stringvariable(int svarnum, char *svalue);
inline void ubasic_set_stringvariable(int svarnum, char *svalue) {
//  if(svarnum >=0 && svarnum <MAX_SVARNUM)
    stringvariables[svarnum&63] = svalue;
}

/*---------------------------------------------------------------------------* /
char* ubasic_get_stringvariable(int varnum) {
  if(varnum>=0 && varnum< MAX_SVARNUM) {
      return stringvariables[varnum];  }
  return store(nullstring); }
/ *---------------------------------------------------------------------------*/
static int string_space_check(int l)
{
   // returns true if not enough room for new string
   int i;
   i = ((MAX_BUFFERLEN - freebufptr) <= (l + 2)); // +2 to play it safe
   if (i) {
      ended= MAX_BUFFERLEN-garbage_collect()<MAX_STRINGVARLEN; 
   }
   return i&ended;
}
/*---------------------------------------------------------------------------*/
static char* store(const char *s1) // scpy return a copy of s1
{
   int bp = freebufptr;
   int l;
   l = strlen(s1);
   if (l && string_space_check(l))  return "\0";
   strcpy(stringbuffer+bp, s1);
   freebufptr = bp + l + 1;
   return stringbuffer+bp;
}
/*---------------------------------------------------------------------------*/
int garbage_collect() 
{
  register int i, totused = 0;
  char *temp;
  char *tp;
//  stringvariables[MAX_SVARNUM] = store('\0');
  if (freebufptr < GBGCHECK)
     return freebufptr - GBGCHECK;
  for (i=0; i< MAX_SVARNUM; i++) { // calculate used space
     totused += strlen(stringvariables[i]) + 1;
  }
  DEBUG_PRINTF("Garbage collector called - reclaiming %d bytes\n", (freebufptr - totused));
  temp = malloc(totused);	// alloc temporary space to store vars
  tp = temp;
  for (i=0; i< MAX_SVARNUM; i++) { // copy used strings to temporary store
     strcpy(tp, stringvariables[i]);
	 tp += strlen(tp) + 1;
  }
  freebufptr = 0;
  tp = temp;
  for (i=0; i< MAX_SVARNUM; i++) { //copy back to buffer
     stringvariables[i] = store(tp);
	 tp+= strlen(tp) + 1;
  }
  free(temp); // free temp space
  return totused;
}
/*---------------------------------------------------------------------------*/
static char* memstr(const void *mem, int l) // return the l chars of mem at offset l1
{
  int bp = freebufptr;
  register int len = 0;
     while (*((const char*)((unsigned)(mem)+len)) && len<l) len++;
     if (len && string_space_check(len)) return "\0"; //stringbuffer+bp;

     for (l=0; l<len; ++l)
     {*(stringbuffer+bp+l)=*((const char*)((unsigned)(mem)+l));}//*((char*)(mem+l));

      *(stringbuffer+bp+len)=0;
 // strncpy((stringbuffer+bp), mem,len);
  freebufptr = bp + len + 1;
  return stringbuffer+bp;
}

/*---------------------------------------------------------------------------* /
static char* sconcat(char *s1, const char*s2) // return the concatenation of s1 and s2
{
   int bp = freebufptr;   
   int rp = bp;
   register int l1, l2;
   l1 = strlen(s1);
   l2 = strlen(s2);
   if (string_space_check(l1+l2))  return '\0';
   strcpy((stringbuffer+bp), s1);
   bp += l1;
   if (l1 == MAX_STRINGVARLEN) {
      freebufptr = bp + 1;
	  return (stringbuffer + rp); 
   }
   l2 = strlen(s2);
   strcpy((stringbuffer+bp), s2);
   if (l1 + l2 > MAX_STRINGVARLEN) {
      l2 = MAX_STRINGVARLEN - l1;
	  // truncate
	  *(stringbuffer + bp + l2) = '\0';
   }   
   freebufptr = bp + l2 + 1;
   return (stringbuffer+rp);   
}
/ *---------------------------------------------------------------------------*/
static void* sshl(char *s, int x) // shift to left (cut) x chars of s
{
  if (x<=0)        return s;
  int l=strlen(s);
  if (l<=x) {*s=0; return s;}
  register int v;
  for (v=x; v<=l; v++) s[v-x]=*(s+v);
  s[l-x+1]=0;      return s;
}
/*---------------------------------------------------------------------------*/
static char* sleft(const char *s1, int l) // return the left l chars of s1
{
   int bp = freebufptr;
   int rp = bp;
   register int i, len=strlen(s1);
   if (l<0) l=len+l;
   if (string_space_check(l)) return "\0";
   for (i=0; i<l; i++) {
      *(stringbuffer +bp) = *s1;
	  bp++; s1++;}
   *(stringbuffer + bp) = '\0';
   freebufptr = bp+1;
   return stringbuffer + rp;
}
/*---------------------------------------------------------------------------*/
static char* sright(const char *s1, int l) // return the right l chars of s1
{
   int bp = freebufptr;
   int rp = bp;
   register int i, j;
   j = strlen(s1);
   if (l<0) //    cut 1..-l
      if (j>=-l) {s1+=-l; l=j-l; }
   if (string_space_check(l)) return "\0";
   for (i=0; i<l; i++) {
      *(stringbuffer + bp) = *(s1 + j-l);
      bp++; s1++;}
   *(stringbuffer + bp) = '\0';
   freebufptr = bp+1;
   return stringbuffer + rp;
}
/*---------------------------------------------------------------------------*/
static char* smid(const char *s1, int l1, int l2) // return the l2 chars of s1 starting at offset l1
{
   int bp = freebufptr;
   int rp = bp;
   int i, j;
   j = strlen(s1);
   if (l1<0) //    cut 1..-l1
	  if (j>=-l1) {s1+=-l1; j+=l1; l1=1;}
   if (l2<0) l2=j+l2; // cut right -l2
   if (l2<1 || l1>j) return store(nullstring);
   if (string_space_check(l2-l1)) return "\0";
   if (l2 > j-l1) l2 = j-l1;
   for (i=l1; i<l1+l2; i++) {
      *(stringbuffer + bp) = *(s1 + l1 -1);
      bp++; s1++;
   }
   *(stringbuffer + bp) = '\0';
   freebufptr = bp+1;
   return stringbuffer + rp;
}
/*---------------------------------------------------------------------------*/
static char* sstr(int j) // return the integer j as a string
{
   int bp = freebufptr;
   int rp = bp;
   if (string_space_check(10)) return "\0";
   sprintf((stringbuffer+bp),"%d",j);
   freebufptr = bp + strlen(stringbuffer+bp) + 1;
   return stringbuffer + rp;
}
/*---------------------------------------------------------------------------*/
static char* schr(int j) // return the character whose ASCII code is j
{
   int bp = freebufptr;
   int rp = bp;
   //if (string_space_check(1)) return "\0";
   sprintf((stringbuffer+bp),"%c",j);
   freebufptr = bp + 2;
   return stringbuffer + rp;
}
/*---------------------------------------------------------------------------*/
static int sinstr(int j, const char *s, const char *s1) // return the position of s1 in s (or 0) 
{
   char *p;
   p = strstr(s+j, s1);
   return p==NULL ? 0:(p - s + 1);
}
/*---------------------------------------------------------------------------*/
static char* sfactor(int svar, int token)  // string form of factor
{
   static char *r;
   int i, v, par, a, a1=0, a2=0;
   register int j=0;
   if (svar) {r=stringvariables[svar-1];}
   else {r=stringvariables[MAX_SVARNUM-1];}

   if (!token) token=tokenizer_token();
 //  if (trace) { // stopwatch(-token);
   //    stopwatch(tokenizer_curptr());}

 if (!svar)
 switch(token) {
    case TOKENIZER_ABC$:
         accept(TOKENIZER_ABC$); return abcdef; // memstr(abcdef,MAX_SVARNUM); //
    case TOKENIZER_TEXTBOX:
    case TOKENIZER_STR$: accept(-token);
    case TOKENIZER_ASC: 
    case TOKENIZER_VAL:
    case TOKENIZER_LEFTPAREN:{  par=is_paren();
         if (par) {accept(TOKENIZER_LEFTPAREN);
            if (is_token(TOKENIZER_RIGHTPAREN)) return "\0";
         }
         j = tokenizer_stringlookahead()==1;
         if (j) r = sexpr(0); else r = sstr(expr());
         if (ended) return "-00($..Error)";
		 if (token==TOKENIZER_ASC || token==TOKENIZER_VAL
          || token==TOKENIZER_TEXTBOX)          return r;
         if (par && !is_comma()) accept(TOKENIZER_RIGHTPAREN);
		 return r;} // r = store(r); 
	case TOKENIZER_MOD:{ //  %0$..%9$ ,Args like in Batchfiles; for data and shift
		 j = tokenizer_variable_num();
		 if (j<0) {accept(TOKENIZER_STRINGVARIABLE); return "<<%Error%";}
		 accept(-TOKENIZER_STRINGVARIABLE);
         r = stringvariables[j];    return r;} //r = store(
	case TOKENIZER_STRINGVARIABLE: {
		 j = tokenizer_variable_num();
		 if (j==-TOKENIZER_LEFTPAREN)
		    {j = varfactor(0x1000);}
		 accept(-TOKENIZER_STRINGVARIABLE);
		 if (j<0) {accept(TOKENIZER_ABC$); return "<<$Error$";}
		 r = stringvariables[j%63]; return r; // store(
    }
    case TOKENIZER_STRING: {
         tokenizer_string(string, sizeof(string));
         r = string; // memstr(string,MAX_STRINGVARLEN)
         accept(TOKENIZER_STRING); return r;
	}
    case TOKENIZER_READ$:{ accept(-token);
         j=MAX_STRINGVARLEN-1;
         par=is_paren(); accept(-TOKENIZER_LEFTPAREN); 
         a1=tokenizer_token(); *string=0;
         switch(a1) {
           case TOKENIZER_NUMBER:
           case TOKENIZER_VARIABLE: j=expr(); break;
           case TOKENIZER_STRING: {
             tokenizer_string(string, sizeof(string));
             accept(TOKENIZER_STRING);
           case TOKENIZER_STRINGVARIABLE:
             if(*string==0)
                strcpy(string,sfactor(0,TOKENIZER_STRINGVARIABLE));
			if (string[1] != '/') return "<<A/FileError";
			string[0]='A';
            if (!readfile) { readfile = fopen(string,"r");
		    if (!readfile) return *string=='A' ? string:"><";} // "<<(FileError)"
            break;}
         } // switch
         if (ended) break;
         if (par) accept(TOKENIZER_RIGHTPAREN);
         readln(readfile,j); return sbuf;}  //store(
    case TOKENIZER_COMMA: accept(0); return "\0";
    case TOKENIZER_SKIP:  sexpr(0);  return "\0";
    default:{
      j=tokenizer_stringlookahead();
      if (j!=1) {stopwatch(30000000-token); beep();
      ubasic_error=UBASIC_E_PARSE;
      ended=1; break;} 
    break;}
  } // switch

  if (ended) return "\0";

  char *s, *s0;
  s=malloc(MAX_STRINGVARLEN+MAX_STRINGLEN); 
  if  (!s) {accept(TOKENIZER_MEMORY); return "\0";}
  *s=0; s0=s; //

  if (svar) {strcpy(s,r);}
  if (is_token(token)) accept(token); // or eaten by let ...?
  par=is_paren(); if (par) accept(TOKENIZER_LEFTPAREN);

  switch(token) {
    case TOKENIZER_CHR$:{ j=1;
         i = expr();  if (ended) break;
         if (i<0 || i>255) i = 0;
         if (par && is_comma()) {accept(0); j=expr()&127;}
         memset(s,i,j);   s[j]=0;
         r=store(s); break;}
    case TOKENIZER_SPC$:{
         j = expr();  if (ended) break;
         memset(s,' ',j); s[j]=0;
         r=store(s); break;}
    case TOKENIZER_LEFT$:{
         sprintf(s,"%s",sexpr(0)); // if (ended) return 0;
         accept(TOKENIZER_COMMA);
         i = expr();   if (ended) break;
         if (*s) r = sleft(s,i);  else *r='\0';
         break;}
    case TOKENIZER_RIGHT$:{
         sprintf(s,"%s",sexpr(0));// if (ended) return 0;
         accept(TOKENIZER_COMMA);
         i = expr();   if (ended) break;
         if (*s) r = sright(s,i); else *r='\0';
         break;}
    case TOKENIZER_MID$:
         sprintf(s,"%s",sexpr(0));  if (ended) break;
         accept(TOKENIZER_COMMA);
    case TOKENIZER_MEM$:{
         i = expr();   if (ended) break;
         if (par && !is_token(TOKENIZER_RIGHTPAREN))
         {   accept(TOKENIZER_COMMA);
             j = expr();   if (ended) break;
         } else j = MAX_STRINGVARLEN; // ensure we get all of it
         if (token==TOKENIZER_MID$) {
            if (*s) r = smid(s,i,j);  else *r='\0';
            break;}
         else {
           if (j>=MAX_STRINGVARLEN) j=MAX_STRINGVARLEN-1;
          // addr=(void *)((char const *)i);
         //  addr=(void *)(unsigned)i;
           addr=(char*)((unsigned)(i));  *r='\0';
           if (addr<=(void*)camera_info.maxramaddr || // gui_debug.c
	          (addr>=(void*)camera_info.rombaseaddr &&
              (unsigned)addr <= 0xfffffffc)) //    r=memstr(*((void**)addr),j); else
             r=memstr(addr,j);
         break;}
         }
    case TOKENIZER_ARG$:{ v=1; a=44; //=','
         sprintf(s,"%s,,",sexpr(0)); if (ended) break;
         if (par && !is_token(TOKENIZER_RIGHTPAREN)) {
             accept(TOKENIZER_COMMA); v = expr()+1;}
         if (par && !is_token(TOKENIZER_RIGHTPAREN)) {
             accept(TOKENIZER_COMMA); r = sexpr(0);
             a=*r;} // opt. delim
         i=strlen(s); s[i]=a; s[i-1]=a; if (ended) break;
         if (v<1 || v>=i) *s='\0'; else
         for (j=0; j<i; j++) {
             if (*(s+j)==a) {a1=a2; a2=j+1; v--;}
             if (v==0) break;}
         if (a2) a2--;
         s[a2]=0;
         if (ended || a2-a1<=1) break;
         while (a2 && *(s+a2)<33) {s[a2]=0; a2--;}
         while (*(s+a1) && *(s+a1)<33) a1++;
         r = store(s+a1);
         break;}
    case TOKENIZER_TRIM$:{
         if (!svar) r=sexpr(0);
         if (ended || !*r) break;
         strcpy(s,r);
         a2=strlen(r);
         while (a2  && *(s+a2)<33) {s[a2]=0; a2--;} //|| strchr("'\"", *(r+j))
         while (*(s+a1) && *(s+a1)<33) a1++; // || *r=='"' || *r==39
         r = store(s+a1);
         break;}

     case TOKENIZER_UP$:{
          if (!svar) r=sexpr(0); 
		  if (ended | !*r) break;
          for (j=0; j<strlen(r); j++)
            if (*(r+j)>='a' && *(r+j)<='z') r[j] -= 32;
          break;}
     case TOKENIZER_LOW$:{
          if (!svar) r=sexpr(0); 
          if (ended | !*r) break;
          for (j=0; j<strlen(r); j++)
            if (*(r+j)>='A' && *(r+j)<='Z') r[j] |= 32;
          break;}
     case TOKENIZER_REV$:{
          if (!svar) r=sexpr(0);
          if (ended | !*r) break;
          i=strlen(r);
          for (j=0; j<i>>1; j++) {
          a=r[i-j-1]; r[i-j-1]=r[j]; r[j]=a;}
          break;}
     case TOKENIZER_REP$:{ i=999; // all
          if (!svar) r=sexpr(0);
          if (ended) break;
		  j=strlen(r);
          accept(-TOKENIZER_COMMA);
          tokenizer_string(s,2); 
          a1=*s; a2=*(s+1); // allow cut.
          accept(TOKENIZER_STRING);
          if (is_comma()) {accept(0); i=expr();}
		  if (*r && *s)
		  for (a=0; a<j; ++a)
              if (r[a]==a1) {r[a]=a2; i--;
                 if (i<1) break;}
          break;}
     case TOKENIZER_STROL$:
     case TOKENIZER_STROR$:{ i=1; a1=0;
          if (!svar) r=sexpr(0);
          if (ended) break;
          j=strlen(r);
          if (is_comma()) {accept(0); i=expr()%j;}
          if (!*r || i==0 || ended) break;
		  if (i<0) i=j+i;
		  if (token==TOKENIZER_STROR$) i=j-i;
          strncpy(s,r,j-i);
          sshl(r,i);
          for (a=j-i; a<j; a++) {r[a]=s[a1]; a1++;}
          break;}
     default:{
		accept(-TOKENIZER_STRINGVARIABLE);
	//	r=store(stringvariables[svar-1]);
		r=stringvariables[svar-1];
        break;}
  } //  switch
  s=s0; free(s);
  if (ended) {stopwatch(30000000+tokenizer_curptr());
               return "<< sfactor-Error";}

  if (par) accept(TOKENIZER_RIGHTPAREN);
  while (tokenizer_char()==TOKENIZER_PRINT) { // C style:	I?x:y == wenn I, rechne mit x sonst y
     if (*r) {accept(TOKENIZER_PRINT); r=sexpr(0);
		accept(TOKENIZER_LABEL); sexpr(0);} 
     else {tokenizer_skip(TOKENIZER_LABEL); 
		accept(TOKENIZER_LABEL); r=sexpr(0);}
  }
  if (is_token(TOKENIZER_PLUS) || is_token(TOKENIZER_TO)) {
      stringvariables[MAX_SVARNUM-1]=r;
      r=sexpr(0x1000+MAX_SVARNUM-1);}
  return r;
}
/*---------------------------------------------------------------------------*/
static char* sexpr(int svar) // string form of expr
{
  int op=0;
  static char *s1;
  char *s2='\0';
  char *s; s=malloc(MAX_STRINGVARLEN+MAX_STRINGLEN);
  if  (!s) {ended=1; return "<<NoMemory";}

  if (svar> MAX_SVARNUM) {svar&=63; // let A$+=b$...
       s1 = strcpy(s,stringvariables[svar]); svar++;
       if (is_token(TOKENIZER_EQ)) op=TOKENIZER_TO;}
  else s1 = strcpy(s,sfactor(0,0));
  
  while (!ended && tokenizer_char()==TOKENIZER_PRINT) { // C style:	I?x:y == wenn I, rechne mit x sonst y
        accept(TOKENIZER_PRINT);
     if (*s1) { s1=strcpy(s,sexpr(0));
        accept(TOKENIZER_LABEL); s2 = sfactor(0,TOKENIZER_SKIP);} 
     else {tokenizer_skip(TOKENIZER_LABEL);
        accept(TOKENIZER_LABEL); s1 = strcpy(s,sexpr(0));}
  }
  if(!op) op = tokenizer_token();
  DEBUG_PRINTF("sexpr s1= '%s' op= %d\n", s1, op);

  while (!ended &&
  (op==TOKENIZER_PLUS || op==TOKENIZER_TO || op==TOKENIZER_PRINT))
  { // concat +string or ..expr()
    if (op==TOKENIZER_PRINT && tokenizer_char()!=TOKENIZER_PRINT) break;
    do accept(0); while (is_cr());
    if (op!=TOKENIZER_TO) s2 = sfactor(0,0);
    if (ended) break;

    switch (op) {
    case TOKENIZER_TO:
       if (tokenizer_stringlookahead() !=1)
            s2 = sstr(expr());
	   else s2 = sfactor(0,0);
    case TOKENIZER_PLUS: break;
    case TOKENIZER_PRINT: { // C style:	I?x:y == wenn I, rechne mit x sonst y
       if (*s1) {accept(TOKENIZER_LABEL);
            sfactor(0,TOKENIZER_SKIP); *s2='\0';} 
       else {accept(TOKENIZER_LABEL); s2=sfactor(0,0);}
    break;}
    DEBUG_PRINTF("sexpr s2= '%s' op= %d\n", s2, op);
	if (ended) break;}
	if (*s2) s1 = strcat(s,s2); 
    op = tokenizer_token();
  } // while
   DEBUG_PRINTF("sexpr returning s1= '%s'\n", s1);  

  s[MAX_STRINGVARLEN-1]='\0';
  if (!svar) {s1=store(s);}
  else {ubasic_set_stringvariable(svar-1, store(s));
        s1=stringvariables[svar-1];}
  if (ended) stopwatch(40000000+tokenizer_line_number());

  free(s);
  return s1;
}
/*---------------------------------------------------------------------------*/
static int slogexpr(int svar) // string logical expression
{
    const char *s1, *s2;
    int p, op, r=0;
    if (is_paren()) return factor(); //accept(TOKENIZER_LEFTPAREN);
    if (svar) s1=stringvariables[svar&63]; else s1 = sexpr(0);
    if (eos()) return *s1!='\0';
    if (is_comma() || is_token(TOKENIZER_SEMICOLON)) {
        if (svar_selected!=0) {
           stopwatch(-tokenizer_line_number()); return -1;}
        svar_selected=-1;  // only for  "print" (">>",a$,...)
        strcpy(string_selected,s1); return 0;}
    op = tokenizer_token();
    if (op==TOKENIZER_IN)  { accept(TOKENIZER_IN);
        p=is_paren(); if (p) accept(TOKENIZER_LEFTPAREN);
		s2 = sexpr(0); 
		if (is_token(TOKENIZER_TO)) {accept(TOKENIZER_TO); // in "a".."z"
		    r = strcmp(s1,s2)>=0;  s2 = sexpr(0);
		    r &=strcmp(s1,s2)<=0; }
        else do {
            r |=strcmp(s1,s2) == 0;
			if (is_comma()) {accept(TOKENIZER_COMMA); // ... in 2,4,6,x...
			s2 =sexpr(0);}
			} while (is_comma());
        accept(-TOKENIZER_SEMICOLON);
        if (p) accept(TOKENIZER_RIGHTPAREN);
     return r;
    }
    if (op>=TOKENIZER_LT && op<=TOKENIZER_GT) accept(op);
    else { accept(TOKENIZER_OUT_OF_FOCUS); return 0;} // ~150
    s2 = sexpr(0); if (ended) return 0;

    r = strcmp(s1,s2);
    switch(op) {
      case TOKENIZER_LT: r = r< 0; break;
      case TOKENIZER_LE: r = r<=0; break;
      case TOKENIZER_EQ: r = r==0; break;
      case TOKENIZER_NE: r = r!=0; break;
      case TOKENIZER_GE: r = r>=0; break;
      case TOKENIZER_GT: r = r> 0; break;
   }
  if (ended) stopwatch(50000000+tokenizer_curptr());
  return r;
}

/*---------------------------------------------------------------------------*/
static int str2u(const char *s) // val([,])
{ int t, c=0, com=0, neg=0;
  register unsigned u=0;
  while (*s && *s<'!') s++;
  if (!*s)  return 0;
  if (*s=='+')  {s++;}
  if (*s=='~')  {s++; com=1;}
  if (*s=='-')  {s++; neg=1;}
  while (*s && *s<'!') s++;
  if (*s=='0' && (s[1]|32)=='x') { s+=2;
     while (( *s>'/'    &&  *s<':')
        || (((*s|32)>96 && (*s|32)<'g'))) {
        if (c==9) break;
        u<<=4;
        u |= (*s<64) ? (*s-48):(*s|32)-87;
        c++; s++;}
 }else
     while (*s>'/' && *s<':') {
        if (c==11) break;
        t=u<<3; u<<=1;
        u += t+*s-48;
        c++; s++;}
  if (neg) u=-u;
  if (com) u=~u;
  return u;
}
// end of string additions
/*
:...;....1....;....2....;....3....;....4....:....5....;....6....;....7....;....*/
/*---------------------------------------------------------------------------*/
static int factor(void)
{
  int r=0, j=0, v=0, par=0;
  unsigned u;
  tConfigVal configVal;
// string function additions
  const char *s, *s1;
// end of string additions
  register short token = tokenizer_token();
  
  if (token == TOKENIZER_STRING) { 
      r=ubasic_get_key_arg();
      if (!ubasic_error) return r;
      ubasic_error=UBASIC_E_NONE; // ignore
   token=tokenizer_token();
  }
  DEBUG_PRINTF("factor: token %d\n", token);
  
  switch(token) {
  case TOKENIZER_VARIABLE:       return varfactor(MAX_VARNUM);
  case TOKENIZER_FIELD: {
    tokenizer_next();  r = relation();
    accept(TOKENIZER_RIGHTPAREN);return field[r%MAX_FIELD];
  }
  case TOKENIZER_NUMBER:         r = tokenizer_value();
    DEBUG_PRINTF("factor: number %d\n", r);
    accept(TOKENIZER_NUMBER);    return r;
// string function additions
  case TOKENIZER_STRINGVARIABLE:{ 
    r = tokenizer_variable_num();
    if (r==-TOKENIZER_LEFTPAREN) r = varfactor(0x1000);
    if (r<0) {accept(TOKENIZER_VARIABLE); break;}
    accept(-TOKENIZER_STRINGVARIABLE);
    return slogexpr(0x1000|r);}
// end of string additions
  case TOKENIZER_MOD:{ //  %0..%9 ,Args like in Batchfiles; for data and shift
    r = tokenizer_variable_num();
    if (r<0) {accept(TOKENIZER_VARIABLE); break;}
    if (is_token(TOKENIZER_STRINGVARIABLE)) return slogexpr(0x1000|r);
    r = variables[r];
    break;}
  case TOKENIZER_CR:// "Expected 3, got 1"
  case TOKENIZER_ENDOFINPUT:
  case TOKENIZER_MD_DETECT_MOTION:
  case TOKENIZER_PRINT:
  case TOKENIZER_ERROR:
    accept(TOKENIZER_VARIABLE); return 0;
    break;
  //  default: break;
  }
  
  accept(token);
  switch(token) {
  case TOKENIZER_PLUS:      r = factor();
    break;
  case TOKENIZER_MINUS:     r = -factor();
    break;
  case TOKENIZER_NOT:       r = ! factor();
    break;
  case TOKENIZER_LNOT:      r = ! relation();
    break;
  case TOKENIZER_LEFTPAREN: r = relation();
    if (!is_comma() && !is_token(TOKENIZER_SEMICOLON))
       accept(TOKENIZER_RIGHTPAREN);
    break;

  case TOKENIZER_ABS:  r = factor();
    if (r<0) {r=-r;};
    break;
  case TOKENIZER_SGN:  r = factor();
    if (r<0) {r=-1;};  if (r>0) {r=1;};
    break;
  case TOKENIZER_UNS:  r = (unsigned short)factor();
    break;
  case TOKENIZER_ODD:  r = factor()&1;
    break;
  case TOKENIZER_EVEN: r = 1-(factor()&1);
    break;
  case TOKENIZER_CAP:  r = factor();
    while (abs(r)>9)   r = r/10;
    break;
  case TOKENIZER_SQRT: v=factor();
     r = (int)(sqrt((double)v)/*+0.5?*/); 
    break;
  case TOKENIZER_ON:
  case TOKENIZER_TRUE: r=1;     break;
  case TOKENIZER_TRON: r=trace; break;
  case TOKENIZER_OFF:
  case TOKENIZER_FALSE:         break;
  
  case TOKENIZER_GET_SHOOTING:
    r = shooting_in_progress();
    break;
  case TOKENIZER_GET_TICK_COUNT:
    r = get_tick_count();
	flag_yield=1;
    break;

  case TOKENIZER_IS_KEY:{
//	moment_statement(20);  // Turns Cam OFF // Why?
	if (ifkey<0) r = script_key_is_pressed(ubasic_get_key_arg());
    else r = script_key_is_clicked(ubasic_get_key_arg());
//	moment_statement(20); ... OFF // no need, but strange.
	flag_yield=1;
    break;}
  case TOKENIZER_IS_PRESSED:{
    if (ifkey<1) r = script_key_is_pressed(ubasic_get_key_arg());
    else r = script_key_is_clicked(ubasic_get_key_arg());
	flag_yield=1;
    break;}
  case TOKENIZER_GET_WHEEL:{
      r=get_jogdial_direction();
	  jogdial=0;
	  if (r==0) break;
      r = r==JOGDIAL_LEFT ? -1:1; // left=-1 | right=1
	flag_yield=1;
    break;}
  case TOKENIZER_WHEEL_LEFT:{
	  if (jogdial==-1) {jogdial=0; r=1; break;} 
	  if (jogdial== 1) {break;} 
      r = get_jogdial_direction();
	  jogdial=r==JOGDIAL_RIGHT; // 
	  if (!r) break;
	  r = r==JOGDIAL_LEFT ? 1:0;
	flag_yield=1;
    break;}
  case TOKENIZER_WHEEL_RIGHT:{
	  if (jogdial== 1) {jogdial=0; r=1; break;}
	  if (jogdial==-1) {break;}  
      r = get_jogdial_direction();
	  jogdial=r==JOGDIAL_LEFT; // 
	  if (!r) break;
	  r = r==JOGDIAL_RIGHT;
	flag_yield=1;
    break;}
  case TOKENIZER_GET_AFLOCK:
	r = shooting_get_prop(camera_info.props.af_lock);
    break;
  case TOKENIZER_GET_AF_FRAME:
    r = camera_info.props.propset==1 ? shooting_get_prop(64) : shooting_get_prop(8);
//	r = shooting_get_prop(camera_info.props.af_frame);
    break;
  case TOKENIZER_GET_FOCUS_MODE:
    r = shooting_get_real_focus_mode();
    break;
  case TOKENIZER_GET_FOCUS_STATE:
    r = shooting_get_focus_state();
    break;
  case TOKENIZER_GET_FOCUS_OK:
    r = shooting_get_focus_ok();
    break;
  case TOKENIZER_GET_ZOOM_STEPS:
    r = zoom_points; // ==1  on EOS-M
    break;
  case TOKENIZER_GET_ZOOM:
    r = shooting_get_zoom();
    break;
  case TOKENIZER_GET_FOCAL_LENGTH:
    r = get_focal_length(shooting_get_zoom()); //--(lens_get_zoom_point()) wg.  EOS-M
    break;
  case TOKENIZER_GET_EFF_FOCAL_LENGTH:
    r = get_effective_focal_length(shooting_get_zoom());
    r = r * conf.zoom_scale / 100; // from gui_osd.c
	break;
//ChangeFocusDistanceToPosition_FW
  case TOKENIZER_GET_FOCUS:
    subj_dist = shooting_get_subject_distance();
	if (camera_info.dof_values.distance_valid!=0)
    r = subj_dist;
    break;
  case TOKENIZER_GET_SD_OVER_MODES:
    r = sd_over_modes() ; //  0..7
// CAM_SD_OVER_IN_AF           & 1
// CAM_SD_OVER_IN_AFL          & 2 
// CAM_SD_OVER_IN_MF           & 4
    break;
  case TOKENIZER_GET_DRIVE_MODE:
    r = shooting_get_drive_mode();
    break;
  case TOKENIZER_SET_MF:
    r = set_mf_statement();
    break;
  case TOKENIZER_GET_MIN_STACK_DIST:
    r = shooting_get_min_stack_distance();
    break;
  case TOKENIZER_GET_NEAR_LIMIT:
    r = shooting_get_near_limit_of_acceptable_sharpness();
    break;
  case TOKENIZER_GET_FAR_LIMIT:
    r = shooting_get_far_limit_of_acceptable_sharpness();
    break;  
   case TOKENIZER_GET_DOF:
    r = shooting_get_depth_of_field();
    break;
  case TOKENIZER_GET_HYPERFOCAL_DIST:
    r = shooting_get_hyperfocal_distance();
    break;
  case TOKENIZER_IS_INFINITY:
	r = (unsigned)subj_dist==INFINITY_DIST;
  //  r = shooting_get_far_limit_of_acceptable_sharpness()==-1;
    break;
  case TOKENIZER_OUT_OF_FOCUS:{
	if (!eos()) r=expr();
	if (r==-1) {r=subj_dist>0; break;}
	r = subj_dist;
	if (r<=shooting_get_near_limit_of_acceptable_sharpness()) {r=1; break;}
    if (r>=shooting_get_far_limit_of_acceptable_sharpness()) r=-1; else r=0;
	break;}
// return true if sensor is enabled (live view on), false if not (playback, rec with display off, display off power save)
  case TOKENIZER_GET_IMAGER_ACTIVE:
	r = shooting_get_imager_active() ? 1:0;
	break;
  case TOKENIZER_GET_EV:
    r = shooting_get_ev_correction1();
    break;
  case TOKENIZER_GET_TV96: //  Tv + Av = Bv + Sv 
    r = shooting_get_tv96();
    break;
  case TOKENIZER_GET_CURRENT_TV96:
    r = shooting_get_current_tv96();
    break;
  case TOKENIZER_GET_USER_TV96:
    r = shooting_get_user_tv96();
    break;   
  case TOKENIZER_GET_USER_TV_ID:
    r = shooting_get_user_tv_id();
    break;
  case TOKENIZER_GET_AV96:
    r = shooting_get_av96();
    break;  
  case TOKENIZER_GET_CURRENT_AV96:
    r = shooting_get_current_av96();
    break;
  case TOKENIZER_GET_USER_AV96:
    r = shooting_get_user_av96();
    break;    
  case TOKENIZER_GET_USER_AV_ID:
    r = shooting_get_user_av_id();
    break;
  case TOKENIZER_GET_MAX_AV96:
    r = shooting_get_max_av96();
    break;
  case TOKENIZER_GET_MIN_AV96:
    r = shooting_get_min_av96();
    break;
  case TOKENIZER_GET_BV96: // Bv = Tv + Av - Sv
    r = (int)shooting_get_bv96();
    break;
  case TOKENIZER_GET_SV96:
    r = (int)shooting_get_sv96_real();
    break;
  case TOKENIZER_GET_CURRENT_DELTA_SV96:
    r = shooting_get_current_delta_sv96();
    break;
  case TOKENIZER_GET_CURRENT_BASE_SV96:
    r = shooting_get_current_base_sv96();
    break;
// from shooting_get_luminance() *1000
  case TOKENIZER_GET_LUMINANCE:
//    static const double K = 12.5;               // K is the reflected-light meter calibration constant
    r = (int)shooting_get_bv96();
    r = (int)(100 * 12.5 * pow(2.0,((double)(r-168)/96.0)) *1000);
    break;
  case TOKENIZER_GET_ISO_MODE:
    r = shooting_get_iso_mode();
    break;
  case TOKENIZER_GET_ISO_MARKET:
    r = (int)shooting_get_iso_market();
    break;
  case TOKENIZER_GET_ISO_REAL:
    r = (int)shooting_get_iso_real();
    break;
  case TOKENIZER_GET_EXP_COUNT:
    r = get_exposure_counter();
    break;
  case TOKENIZER_GET_JPG_COUNT:
    r = GetJpgCount();
    break;
  case TOKENIZER_GET_RAW_COUNT:
    r = GetRawCount();
    break;
  case TOKENIZER_GET_RESOLUTION:
    r = shooting_get_resolution();
    break;
  case TOKENIZER_GET_QUALITY:
    r = shooting_get_prop(camera_info.props.quality);
    break;
  case TOKENIZER_GET_RAW_NR:
    r = conf.raw_nr;     //NoiseReduction set 0 if CamError E26 (Ixus300hs)
    break;
  case TOKENIZER_GET_FREE_DISK_SPACE:
    r = GetFreeCardSpaceKb();
    break;
  case TOKENIZER_GET_PROPSET:
    r = camera_info.props.propset;
    break;
  case TOKENIZER_GET_PROP:
    v = expr();
    r = shooting_get_prop(v);
    break;
  case TOKENIZER_GET_HISTO_RANGE:
    par=is_paren();
    r = expr();
	accept(-TOKENIZER_COMMA);
    v = expr();
    if (par) accept(-TOKENIZER_RIGHTPAREN);
#ifdef UBASIC_TEST
    (void)r; (void)v;
    r = 0;
#else
    r = (unsigned short)libshothisto->shot_histogram_get_range(r, v);
#endif
    break;
  case TOKENIZER_GET_VBATT:
    r = (unsigned short) stat_get_vbatt();
    break;
  case TOKENIZER_GET_TEMPERATURE:{ r = expr();
  switch (r) {
    case 0: r = get_optical_temp(); 
      break;
    case 1: r = get_ccd_temp(); 
      break;
    case 2: r = get_battery_temp();
      break;
    case 3:
    default: // do something sane if given a bad index
      r = get_battery_temp() + get_ccd_temp() + get_optical_temp(); // Dont panic!
      break;
  }
  break;}
  case TOKENIZER_GET_FLASH_MODE:
    r = shooting_get_flash_mode();
    break;
  case TOKENIZER_GET_FLASH_READY:
    r = shooting_is_flash();
    break;
  case TOKENIZER_GET_ORIENTATION_SENSOR:
    r = shooting_get_prop(camera_info.props.orientation_sensor);
    break;
  case TOKENIZER_GET_IS_MODE:
    r = shooting_get_is_mode();
    break;
  case TOKENIZER_GET_ND_CURRENT_EV96:
    r = shooting_get_nd_current_ev96();
    break;
  case TOKENIZER_GET_ND_VALUE_EV96:
    r = shooting_get_nd_value_ev96();
    break;
  case TOKENIZER_GET_ND_PRESENT:
    if (camera_info.cam_has_nd_filter)
    r = camera_info.cam_has_iris_diaphragm ? 2:1;
    break;
  case TOKENIZER_GET_RAW_SUPPORT:
    r = (is_raw_possible() && !camera_info.state.mode_play);
    break;    
 case TOKENIZER_GET_RAW:
#ifdef UBASIC_TEST
    r = 1;
#else
    r = conf.save_raw;
#endif    
    break;
  case TOKENIZER_GET_MODE:
    if (camera_info.state.mode_play) r = 1;
    if (!camera_info.state.mode_play && camera_info.state.mode_video) r = 2;
  break;
  case TOKENIZER_GET_METERING_MODE:
	r= shooting_get_prop(camera_info.props.metering_mode);
  break;
  // get CHDK capture mode value, or 0 if in playback or unknown (broken modemap)
  // NOTE: different from get_mode, since this returns the actual value
  case TOKENIZER_GET_CAPTURE_MODE:
    if (camera_info.state.mode_rec) 
      r = camera_info.state.mode_shooting;
    break;
  // check if CHDK capture mode exists in modemap of this camera
  case TOKENIZER_IS_CAPTURE_MODE_VALID:
    v = expr();
    r = (shooting_mode_chdk2canon(v) == -1) ? 0 : 1;
    break;
  case TOKENIZER_GET_VIDEO_BUTTON:
    r = (camera_info.cam_has_video_button) ? 1 : 0;
    break;
  case TOKENIZER_GET_VIDEO_RECORDING:
    r = is_video_recording();
    break;
  case TOKENIZER_GET_MOVIE_STATUS:
    r = get_movie_status();
    break;
  case TOKENIZER_GET_CONFIG_VALUE:
    r = expr();
	accept(-TOKENIZER_COMMA);
    v = expr();
    if( conf_getValue(r, &configVal) == CONF_VALUE) r = configVal.numb; else r = v;
    break;
  case TOKENIZER_RANDOM:{  par=is_paren();
    r = expr();
	accept(-TOKENIZER_COMMA);
    v = expr();
    // shouldn't srand every time...
    srand((int)shooting_get_bv96()+(unsigned short)stat_get_vbatt()+get_tick_count());
    // wtf
    //action_push_delay(rand()%10);
    flag_yield=1;
    r += rand()%(v-r+1);
    if (par) accept(-TOKENIZER_RIGHTPAREN);
  break;}

  // APEX functions
  case TOKENIZER_ISO_TO_SV96:
    r = shooting_get_sv96_from_iso(expr());
    break;
  case TOKENIZER_SV96_TO_ISO:
    r = shooting_get_iso_from_sv96(expr());
    break;
  case TOKENIZER_ISO_REAL_TO_MARKET:
    r = shooting_iso_real_to_market(expr());
    break;
  case TOKENIZER_ISO_MARKET_TO_REAL:
    r = shooting_iso_market_to_real(expr());
    break;
  case TOKENIZER_SV96_REAL_TO_MARKET:
    r = shooting_sv96_real_to_market(expr());
    break;
  case TOKENIZER_SV96_MARKET_TO_REAL:
    r = shooting_sv96_market_to_real(expr());
    break;
  case TOKENIZER_APERTURE_TO_AV96:
    r = shooting_get_av96_from_aperture(expr());
    break;
  case TOKENIZER_AV96_TO_APERTURE:
    r = shooting_get_aperture_from_av96(expr());
    break;
  case TOKENIZER_USEC_TO_TV96:
    r = shooting_get_tv96_from_shutter_speed((float)expr()/1000000.0);
    break;
  case TOKENIZER_TV96_TO_USEC:
    r = (int)(shooting_get_shutter_speed_from_tv96(expr()) * 1000000.0 + 0.5);
    break;
  case TOKENIZER_SECONDS_TO_TV96:
    par=is_paren();
    v = expr();
	accept(-TOKENIZER_COMMA);
    r = expr();
    if (par) accept(-TOKENIZER_RIGHTPAREN);
    r = shooting_get_tv96_from_shutter_speed((float)v/(float)r);
    break;
  case TOKENIZER_FORCE_ANALOG_AV:
    r=kbd_force_analog_av(expr());
    break;
  case TOKENIZER_FORCE_USB_PRESENT:
    r=force_usb_state(expr());
    break;
  case TOKENIZER_GET_USB_POWER:
    if (!eos()) { v = expr(); }
    r = get_usb_power(v);
    break;
  case TOKENIZER_GET_ALT_MODE:
    r = (camera_info.state.gui_mode != 0);
    break;    
  case TOKENIZER_GET_DISPLAY_MODE:
    r = shooting_get_display_mode();
    break;
  case TOKENIZER_GET_DRAW_TITLE_LINE:
    r = camera_info.state.osd_title_line ;
    break;
  case TOKENIZER_SCRIPT_AUTOSTARTED:
    r = camera_info.state.auto_started;
    break;
  case TOKENIZER_GET_SCRIPT_AUTOSTART:
#ifdef UBASIC_TEST
    r = 0;
#else
    r = conf.script_startup;
#endif
    break;
  case TOKENIZER_GET_PLATFORM_ID: r = conf.platformid;
    break;
  case TOKENIZER_GET_DIGIC:       r = camera_info.cam_digic;
    break;
  case TOKENIZER_GET_CANON_RAW_SUPPORT:  r = camera_info.cam_canon_raw;
    break;
  case TOKENIZER_GET_CANON_IMAGE_FORMAT: r = shooting_get_canon_image_format();
    break;
  case TOKENIZER_GET_DISK_SIZE:   r = GetTotalCardSpaceKb();
    break;
  case TOKENIZER_SWAP_PARTITIONS:
    v = expr();
    if (ended) {return 0;}
    r = swap_partitions(v);
    break;
  case TOKENIZER_GET_DAY_SECONDS:
    {
        struct tm *ttm;
        ttm = get_localtime();
        r = ttm->tm_hour * 3600 + ttm->tm_min * 60 + ttm->tm_sec;
    }
    break;
  case TOKENIZER_WAIT_CLICK:
	  r=wait_click_statement(); // r= ms since last click
    break;
  case TOKENIZER_MD_DETECT_TICK:
    r = camera_info.perf.md_detect_tick;
    flag_yield=1;
    break;
  case TOKENIZER_GET_TIME:{
    par = (is_paren());
    if (par) tokenizer_next();
    if (is_token(TOKENIZER_STRING)) {
        tokenizer_string(string, sizeof(string));
        r=strlen(strpbrk(" SrJTRpYMDhms",string))-1;
        accept(TOKENIZER_STRING); }
    else r=expr();
    r=get_time(r);
    if (par) accept(TOKENIZER_RIGHTPAREN);
    break;}
  case TOKENIZER_MSEC:    r = get_tick_count()-first_tick; // whole line same 
    break;
  case TOKENIZER_LSEC:    r = get_tick_count()/1000-lsec;
    break;
  case TOKENIZER_SEKUNDE: r = sec;
    break;
  case TOKENIZER_MINUTE:  r = sec/60;
    break;
  case TOKENIZER_STUNDE:  r = sec/3600;
    break;
  case TOKENIZER_PEEK: {
    par=is_paren();
    u=(unsigned)expr(); v=4;
    if (is_comma()) {
		accept(TOKENIZER_COMMA); v=expr()%5; }
    if (u & (v-1)) {accept(TOKENIZER_PEEK); break;}
    if (v==4) r=(unsigned int) *(unsigned int *)(u); else
    if (v==2) r=(unsigned short) *(unsigned short *)(u); else
    if (v==1) r=(unsigned char) *(unsigned char *)(u);
    if (par) accept(-TOKENIZER_RIGHTPAREN);
    break; }
  case TOKENIZER_CRC: {
    par=is_paren();
    u=(unsigned)expr(); v=4;
    if (is_comma()) {
		accept(TOKENIZER_COMMA); v=expr(); }
	for (;j<v;j++) r+=(unsigned char) *(unsigned char *)(u+j);
    if (par) accept(-TOKENIZER_RIGHTPAREN);
    break; }
  case TOKENIZER_MEMORY: {r=2049;   char* s2;
    do {r--; s2=umalloc(r*1024);} while (!s2);
    if (s2) ufree(s2);
    break;}
  case TOKENIZER_BEEP:   {r = beep();}
	break;
  case TOKENIZER_LOOP:
	  loop=variables[MAX_VARNUM-1];
      r=loop;	break;
  case TOKENIZER_DEBUG:   r = debug;	break;
  case TOKENIZER_BUG:     r = tokenizer_line_number();
    break;

  case TOKENIZER_VARLIST: r = varlist(); // or varlist$
	break;
  case TOKENIZER_STRLEN:  r = str_len;// (last print,echo,...)
	break;
 // string function additions
  case TOKENIZER_LEN:{
    if (tokenizer_char()==TOKENIZER_STRINGVARIABLE) {accept(0);
         for (; j<MAX_SVARNUM; ++j) r+= strlen(stringvariables[j]);}
    else r = strlen(sfactor(0,0));
    break;}
  case TOKENIZER_ASC: { par=is_paren();
	s = sfactor(0,token); v=strlen(s); if (ended) return 0;
    if (par && is_comma()) { accept(TOKENIZER_COMMA);
        r=expr()-1;
		if (ended) return 0;
    }
    if (par) accept(-TOKENIZER_RIGHTPAREN);
    if (r<0) r=v+r;
    if (r>v) r=-1;
    r = (r>=0) ? s[r]:0;
    break;}
  case TOKENIZER_VAL: { par=is_paren();// '(' & ')'? eaten in --> ?
    s = sfactor(0,token); v=strlen(s); if (ended) return 0;
    if (par && is_comma()) { accept(TOKENIZER_COMMA);
       if (tokenizer_stringlookahead()==1) {
          r = sinstr(0, s, sfactor(0,0));} //val("img_1234","_")
       else {r=expr();} // =val("img_1234",5) [|-4]
       if (ended) return 0;
    }
    if (par) accept(-TOKENIZER_RIGHTPAREN);
    if (r<0) r=v+r;
    if (r>v) r=-1;
    r = (r>=0) ? str2u(s+r) : 0;
    break;}
  case TOKENIZER_ARGS:{ v=44; par=is_paren(); // ','
    if (par) accept(TOKENIZER_LEFTPAREN);
	s = sexpr(0); if (ended) break;
    if (par && is_comma()) { accept(TOKENIZER_COMMA);
	   s1=sfactor(0,0); v=*s1; // not default separator
       if (ended) return 0;}
	for (j=0; j<strlen(s); ++j) if (s[j]==v) r++;
    if (par) accept(-TOKENIZER_RIGHTPAREN);
	r++;
  break;}
  case TOKENIZER_INSTR: {r=tokenizer_token();
    accept(TOKENIZER_LEFTPAREN);     j = 1; // offset to find
    if (tokenizer_stringlookahead()!=1)
	if ( r == TOKENIZER_NUMBER
	  || r == TOKENIZER_VARIABLE
	  || r == TOKENIZER_LEFTPAREN) { j = expr();
      accept(TOKENIZER_COMMA);
	} 
	s=sexpr(0); if (ended) break;
    if (j<0) j=strlen(s)+j;
	accept(TOKENIZER_COMMA);
	s1 = sexpr(0); if (ended) break;
    accept(TOKENIZER_RIGHTPAREN);
	if ( *s && !*s1) return strlen(s);
	if (!*s || j <1) return 0;
	r = sinstr(j, s, s1);
    break;}
  case TOKENIZER_OPEN:{ r = readfile!=NULL;
    par=is_paren();
    if (par) {accept(TOKENIZER_LEFTPAREN);
	    if (r) {tokenizer_skip(TOKENIZER_RIGHTPAREN);
			accept(0); return 1;} //
        token=tokenizer_token(); *string='\0';
        switch(token) {
          case TOKENIZER_STRING: {
            tokenizer_string(string, sizeof(string));
            accept(TOKENIZER_STRING);
          case TOKENIZER_STRINGVARIABLE:
            if(*string==0)
                strcpy(string,sfactor(0,TOKENIZER_STRINGVARIABLE));
			if (string[1] != '/') {accept(TOKENIZER_SLASH); break;}
			string[0]='A';
            if (!readfile) { readfile = fopen(string,"r");} // "<<(FileError)"
          break;}
		} // switch
	}
    if (par) accept(TOKENIZER_RIGHTPAREN);
    r = readfile!=NULL;
	break;}
 // end of string additions 
 
  case TOKENIZER_MAX:{ r =-0x7ffffff;
    par=is_paren();
    if (par) accept(TOKENIZER_LEFTPAREN);
    do {accept(-TOKENIZER_COMMA);
        v=expr(); if (v>r) r=v;} // 
    while (is_comma());
    accept(-TOKENIZER_SEMICOLON);
    if (par) accept(TOKENIZER_RIGHTPAREN);
    break;}
  case TOKENIZER_MIN:{ r=0x7ffffff;
    par=is_paren();
    if (par) accept(TOKENIZER_LEFTPAREN);
    do {accept(-TOKENIZER_COMMA);
        v=expr(); if (v<r) r=v;} // 
    while (is_comma());
    accept(-TOKENIZER_SEMICOLON);
    if (par) accept(TOKENIZER_RIGHTPAREN);
    break;}
  case TOKENIZER_AVERAGE:{
    par=is_paren();
    if (par) accept(TOKENIZER_LEFTPAREN);
    do {if (is_comma()) {accept(0); accept(-TOKENIZER_CR);}
        r+=expr(); v++;
		if (par && ended) tokenizer_skip(TOKENIZER_RIGHTPAREN);} // 
       // ended=0; ubasic_error=UBASIC_E_NONE;
    while (is_comma());
    accept(-TOKENIZER_SEMICOLON);
    if (par) accept(TOKENIZER_RIGHTPAREN);
    if (v) r/=v;
	break;}
  
  // catch some Errors
  case TOKENIZER_EXCEPTION: // Inf.Loop, Romlog on demand?  20..60! secs later
   ended=0; while (ended<1); {ended++;}
  case TOKENIZER_COMMA:  break; // ignore or error later
  case TOKENIZER_STRING:  r = factor(); // ignore
	break;
  case TOKENIZER_DEF_FN: {
    par=is_paren(); 
    if (par) accept(TOKENIZER_LEFTPAREN);
    r = tokenizer_token(); accept(r);// if def (abc$) then...
    r = 3 ? 0:r;  // unk_stm fails as variable (got 3)
    if (par) tokenizer_skip(TOKENIZER_RIGHTPAREN);
    ended=0; ubasic_error=UBASIC_E_NONE;
	if (par) accept(TOKENIZER_RIGHTPAREN);
	break;}
  case TOKENIZER_TEST: {  // r=expr();// sandbox for developing
    par=is_paren(); 
    if (par) accept(TOKENIZER_LEFTPAREN);
    if (!is_token(TOKENIZER_RIGHTPAREN)) r=relation(); 
   //r=shooting_get_exif_subject_dist();
     //  r=tokenizer_curptr();
    if (par && is_comma()) {
        accept(TOKENIZER_COMMA); v=expr(); }
    if (par) accept(TOKENIZER_RIGHTPAREN);
    ended=0; ubasic_error=UBASIC_E_NONE; // ignore
    if (ended) accept(TOKENIZER_TEST);
    return r;
  }

  //ARM Begin
      
  default: {
 // string function additions
    if (tokenizer_stringlookahead()==1) return slogexpr(0);
 // end of string additions 
    r=tokenizer_curptr();
    stopwatch(2000000+r);
    r = 999999999;//0x7fffffff+1;
  break;}
  } // upper switch
  if (is_paren()) {// more compat(). to some Lualines.
	 tokenizer_skip(TOKENIZER_RIGHTPAREN);
     accept(TOKENIZER_RIGHTPAREN);} // Caefix
  return r;
}
/*---------------------------------------------------------------------------*/
static int term(void)
{
  int f1, f2, f3=0;
  register short op, par=0;
      f1 = factor();
   //if (ended) return 0;
  op = tokenizer_token();
  if (op == TOKENIZER_STRING) {accept(0); // ignore
    op = tokenizer_token();}
  DEBUG_PRINTF("term: token %d\n", op);
  if  (is_cr()) return f1;
  while(
   (op >= TOKENIZER_LT && op <= TOKENIZER_GT) ||
	op == TOKENIZER_IN ||
    op == TOKENIZER_ASTR ||
	op == TOKENIZER_SLASH ||
	op == TOKENIZER_XOR||
    op == TOKENIZER_OR ||
	op == TOKENIZER_SHL||
	op == TOKENIZER_SHR||
	op == TOKENIZER_MOD ) { do tokenizer_next(); while (is_cr());
	if ( is_token(TOKENIZER_TRUE))
		{f1 = f1!=0;   accept(TOKENIZER_TRUE); return f1;}
	if ( op == TOKENIZER_IN) {
		par=is_paren();
		if (par) accept(TOKENIZER_LEFTPAREN);}
	if ( op == TOKENIZER_EQ) {
		if ( is_token(TOKENIZER_SEP))
			{tokenizer_next(); op=TOKENIZER_SEP;} else
		if ( is_token(TOKENIZER_EQ)) tokenizer_next();
	}
    f2 = factor();  if (ended) {return 0;}
    DEBUG_PRINTF("term: %d %d %d\n", f1, op, f2);
    switch(op) {
    case TOKENIZER_LT:     f1 = f1 < f2; break;
    case TOKENIZER_LE:     f1 = f1 <= f2; break;
    case TOKENIZER_EQ:     f1 = f1 == f2; break;
    case TOKENIZER_NE:     f1 = f1 != f2; break;
    case TOKENIZER_GE:     f1 = f1 >= f2; break;
    case TOKENIZER_GT:     f1 = f1 > f2; break;
    case TOKENIZER_ASTR:   f1 *= f2; break;
    case TOKENIZER_OR:     f1 |= f2; break;
    case TOKENIZER_XOR:    f1 ^= f2; break;
    case TOKENIZER_SLASH:  if (!f2) {f1=0xffffffff; break;}
      f1 /= f2; break; // avoid div by zero
    case TOKENIZER_MOD:    if (!f2) {accept(op); break;}
      f1 %= f2; break;
    case TOKENIZER_SHL:    f1 <<=(f2 & 31); break;
    case TOKENIZER_SHR:    f1 >>=(f2 & 31); break;
    case TOKENIZER_IN: { op = f1==f2; 
	  if (is_comma()) do {
         accept(TOKENIZER_COMMA); // ... in 2,4,6,x...
         if (f1 == expr()) op=1; } while (is_comma());
      else { accept(TOKENIZER_TO); f3 = expr();
		 op = (f1>=f2) && (f1<=f3);} // in 2..x
         accept(-TOKENIZER_SEMICOLON);
	  if (par) accept(TOKENIZER_RIGHTPAREN);
	  f1 = op; break; }
    case TOKENIZER_SEP:	accept(TOKENIZER_SEMICOLON);
	  f3 = expr(); 	//	 if x =~ 100;4 then ? "x=[96..104] == true"
      f1 = (f1>=f2-f3) && (f1<=f2+f3); break;
    }
    op = tokenizer_token();
  }
  DEBUG_PRINTF("term: %d\n", f1);
  if (op==TOKENIZER_ENDOFINPUT) ended=1;
  return f1;
}
/*---------------------------------------------------------------------------*/
static int expr(void)
{
  int t1, t2;
  register short op;
  
  t1 = term();
 // if (ended) {return 0;}
  op = tokenizer_token();
  DEBUG_PRINTF("expr: token %d\n", op);
  while(op == TOKENIZER_PLUS ||
	op == TOKENIZER_MINUS ||
	op == TOKENIZER_AND ||
    op == TOKENIZER_LOR ||
    op == TOKENIZER_NOR ||
	op == TOKENIZER_NAND||
	op == TOKENIZER_LXOR||
   (op == TOKENIZER_PRINT && tokenizer_char()==TOKENIZER_PRINT ))
  
  { do tokenizer_next(); while (is_cr());
    t2 = term();
//    if (ended) {return 0;}
    DEBUG_PRINTF("expr: %d %d %d\n", t1, op, t2);
   switch(op) {
    case TOKENIZER_PLUS:  t1 += t2; break;
    case TOKENIZER_MINUS: t1 -= t2; break;
    case TOKENIZER_AND:   t1 &= t2; break;
    case TOKENIZER_LOR:   t1 = t1 || t2; break;
    case TOKENIZER_NAND:  t1 = t1 && !t2; break;
    case TOKENIZER_NOR:   t1 = !t1 && !t2; break;// weder noch   
    case TOKENIZER_LXOR:  t1 = !t1 + !t2 ==1; break;
    case TOKENIZER_PRINT: // C style:	I?x:y == wenn I, rechne mit x sonst y
      if (t1) {t1 = t2; accept(TOKENIZER_LABEL); expr();} 
	  else { accept(TOKENIZER_LABEL); t1=expr();}
      break;
    }
    op = tokenizer_token();
  }
 // if (ended) {return 0;}
  if (op==TOKENIZER_LAND) {tokenizer_next(); t1 = (t1 && relation());}
  DEBUG_PRINTF("expr: %d\n", t1);
  return t1;
}
/*---------------------------------------------------------------------------*/
static int relation(void)
{
  int r1, r2;
  short op;
// string additions
  if (tokenizer_stringlookahead()==1)  { return slogexpr(0); }
// end of string additions

  r1 = expr();  //  if (ended) {return 0;}
  op = tokenizer_token();
  DEBUG_PRINTF("relation: token %d\n", op);
  while(op == TOKENIZER_LAND || op == TOKENIZER_PRINT) {
    if (op == TOKENIZER_PRINT && tokenizer_char() != TOKENIZER_PRINT) break;
    accept(0);
    r2 = expr();
 //   if (ended) {return 0;}
    DEBUG_PRINTF("relation: %d %d %d\n", r1, op, r2);
    switch(op) {
    case TOKENIZER_LAND:
      r1 = r1 && r2;
      break;
    case TOKENIZER_PRINT: // C style:	I?x:y == wenn I, rechne mit x sonst y
      if (r1) {r1 = r2;	accept(TOKENIZER_LABEL); expr();} 
	  else {accept(TOKENIZER_LABEL); r1=expr();}
    }
    if (ended) {return 0;}
    op = tokenizer_token();
  }
  return r1;
}
/*---------------------------------------------------------------------------*/
static int varfactor(unsigned short n)
{
  short p=n, s=is_token(TOKENIZER_STRINGVARIABLE);
  if (s||n==MAX_VARNUM) n=tokenizer_variable_num();
  if (!s) accept(TOKENIZER_VARIABLE); else accept(TOKENIZER_STRINGVARIABLE); 
  if (is_paren()) { tokenizer_next();
      n=(expr()); //if (ended) return 0;
      accept(TOKENIZER_RIGHTPAREN);
  }
  DEBUG_PRINTF("varfactor: obtaining %d from variable %d\n", variables[n], n);
  if (n == MAX_VARNUM) return 0x7fffffff; // #+
  if (n >  MAX_VARNUM) accept(TOKENIZER_VARIABLE);
  if (s || p>=0x1000)  return n;
  return variables[n&63];
}

/*---------------------------------------------------------------------------*/

// Process a sleep function from the stack
static int action_stack_AS_UBASIC_SLEEP()
{
    if (get_tick_count() >= action_top(2))
    {
        action_pop_func(1);
        return 1;
    }
    return 0;
}
 
/*---------------------------------------------------------------------------*/
static int sleep_delay(int delay)
{
    /* delay of -1 signals indefinite (actually 1 day) delay */
    if (delay == -1)
        delay = 86400000;

    if (delay > 0) {
        return delay + get_tick_count();
    }
    return 0;
}
 
static void sleep_statement(void)
{
    accept(TOKENIZER_SLEEP);
    
    int d=expr();
    if (ended) {return;}

	int delay = sleep_delay(d);
 
    if (delay > 0)
    {
		if (d==500) {play_sound(5); d=400;} // 
        action_push(delay);
        action_push_func(action_stack_AS_UBASIC_SLEEP);
	//	if ( abs(delay-get_tick_count()) > 10) beep(); // strange tune.
    } // => get_tick_count updated once per cycle, not per (?).
 
    flag_yield=1;
    DEBUG_PRINTF("End of sleep\n");
    accept_cr();
}
 
static void moment_statement(unsigned d) //:...i....1....;....2....;....3....;....4..
{
	d=d+get_tick_count();
	if (is_token(TOKENIZER_MOMENT)) // '.'
		{accept(0); d=d+50;}
	if (is_token(TOKENIZER_TO)) // '..'
		{accept(0); d=d+200;}
	if (is_token(TOKENIZER_SEKUNDE)) // ''
		{accept(0); d=d+990;}
	action_push(d);
    action_push_func(action_stack_AS_UBASIC_SLEEP);
  flag_yield=1;
}
/*---------------------------------------------------------------------------*/ 
static unsigned stopwatch(int v)  // u=tmode(8) "R"
{
  static char buf[44];
  unsigned u=(get_tick_count()-first_tick);
  buf[0]=0;
  sprintf(buf, "--%d : %d:%02d.%02d",v,  u/60000,(u%60000)/1000,u%1000/10);
  script_console_add_error((long)buf);
  return u; 
}
/*---------------------------------------------------------------------------*/
static unsigned get_time(int tmode)
{  int r=42;
   short s,m,h,Y,M,D;
   static struct tm *ttm;
   ttm = get_localtime();
   s=ttm->tm_sec; m=ttm->tm_min; h=ttm->tm_hour; // --~>1000  "SrJTRpYMDhms"
   D=ttm->tm_mday; M=ttm->tm_mon+1; Y=1900+ttm->tm_year;   // "smhDMYpRTJrS"
      if      (tmode==0) r = s;
      else if (tmode==1) r = m;
      else if (tmode==2) r = h;
      else if (tmode==3) r = D;
      else if (tmode==4) r = M;
      else if (tmode==5) r = Y;
	  else if (tmode==6) r = get_tick_count()%1000;
      else {			 r =(get_tick_count()-first_tick); // Runtime
		   if (tmode==7) // buf needs own line
              sprintf(buf, "@ %d:%d : %02d.%02d",sec/3600,sec%3600/60,sec%60,r%1000); else
           if (tmode==8) sprintf(buf, "@ %d/%d/%d # %d:%02d:%02d", Y,M,D,h,m,s); else // Timestamp
           if (tmode==9)  return 1721013+367*Y-7*(Y+(M+9)/12)/4+275*M/9+D; else // Julian day
           if (tmode==10) return r/1000; else
           if (tmode==11) return h * 3600 + m * 60 + s; // GET_DAY_SECONDS

		else {return stopwatch(tmode);}
           // Show timeline, Caefix
           script_console_add_error((long)buf); buf[0]='\0';
		   flag_yield=1;}
    return r;
}

/*---------------------------------------------------------------------------*/
static void print_statement(short token)
{
static char buf[MAX_STRINGVARLEN+1];
int i,x=0;
unsigned short t; 
unsigned u=0;
short echo=token==TOKENIZER_ECHO;
short pscr=token==TOKENIZER_WRITE;
short scol=token==TOKENIZER_STRLEN;
  buf[0]=0;
  do { accept(-token); t=tokenizer_token();} while (t == TOKENIZER_CR);
  if (pscr && t==TOKENIZER_EQ) { // write [=["A/Romlog.log"],]....
    tokenizer_next();   t=tokenizer_token();
	if (tokenizer_stringlookahead()==1) {strncpy(writefile,sfactor(0,0),63);
		if (writefile[1]=='/') *writefile='A';
		else sprintf(writefile, "A/CHDK/LOGS/Log_%04d.Log", abs(logfile));
	}
	accept(TOKENIZER_COMMA);
  }
  if (is_token(TOKENIZER_PRINT)) { accept(0); // ?? ...
      strcpy(buf, stringvariables[MAX_SVARNUM-2]);}
  if (ended) pscr=0; else	t=tokenizer_token();

  do { string[0]=0; x=0;
    DEBUG_PRINTF("Print loop\n");
	int blen=strlen(buf); if (ended || blen>123) break;
    if(t == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      sprintf(buf+blen, "%s", string);
      tokenizer_next();
    } else if (t == TOKENIZER_EQ) { tokenizer_inc();
		t=tokenizer_token();
		switch(t) {
		case TOKENIZER_PLUS: { struct tm *ttm = get_localtime();
			sprintf(string, "%08d :%02u:%02u:%02u", get_tick_count(),
			ttm->tm_hour, ttm->tm_min, ttm->tm_sec); break;}
		case TOKENIZER_SLASH:	get_target_dir_name(string); break;
		case TOKENIZER_ASTR: 	sprintf(string, "%s",scripttitle); break;
		case TOKENIZER_SEMICOLON:sprintf(string, "%s", subtitle); break;
		case TOKENIZER_MOMENT:	sprintf(string, "Img_%04d",get_exposure_counter()); break;
		case TOKENIZER_PRINT:	sprintf(string, "%s-%s-%s-%s-%s",
			 camera_info.platform, camera_info.platformsub, camera_info.chdk_ver,
			 camera_info.build_number, camera_info.build_svnrev);
			break; 
		case TOKENIZER_NOT: sprintf(string, "M%d SD%d Dm%d Mm%d F%d AF%dL%d",
			 camera_info.state.mode_shooting, sd_over_modes(), shooting_get_drive_mode(),
			 shooting_get_prop(camera_info.props.metering_mode),
			 shooting_get_real_focus_mode(), shooting_get_prop(8), // 8=AF_FRAME : 1 = FlexiZone, 2 = Face AiAF / Tracking AF
			 shooting_get_prop(camera_info.props.af_lock));
			break;
		case TOKENIZER_EQ: i=subj_dist;
			 if (i<1) i=shooting_get_exif_subject_dist();
				sprintf(string, "%d <%d> %d", 
				shooting_get_near_limit_of_acceptable_sharpness(),
				i, shooting_get_far_limit_of_acceptable_sharpness());
			break;
		case TOKENIZER_GT: sprintf(string, "%d => %d", shooting_get_zoom(),
		get_effective_focal_length(shooting_get_zoom())* conf.zoom_scale / 100);
			break;
		case TOKENIZER_OR:
			sprintf(string, "B%d D%d F%d W%d G%d I%d S%d",break_stack_ptr,do_stack_ptr,
			for_stack_ptr,while_stack_ptr,gosub_stack_ptr,if_stack_ptr,select_stack_ptr);
			break;
		case TOKENIZER_MOD: {tokenizer_next();
			u=(unsigned)(expr()); x=1;
			if (u>0xFFFF) {sprintf(string, "0x%08x",u); break;}
			if (u>0xFF)    sprintf(string, "0x%04x",u);
			else		   sprintf(string, "0x%02x",u);
            break;}
		case TOKENIZER_LEFTPAREN: { x=1;
			 u=(unsigned)(factor()); addr=(char*)(u);
             sprintf(string, "*0x%x=(%s)", addr, memstr(addr,40));
             break;}
		// use tok_singlechar to continue...
	//	case TOKENIZER_AND: break;
	//	case TOKENIZER_MINUS: {break;}
	//	case TOKENIZER_LT:  { break;}
	//	case TOKENIZER_XOR: { break;}
		// use tok_singlechar to continue...
		  default:{	strcat(buf," ( ");
			tokenizer_string(string,123); tokenizer_eol();
			strcat(buf,string);
			strcat(buf," )");}
		} // switch
		if (ended) return;

		if ((x==0) && !eos()) tokenizer_next();
		strcat(buf,string); // ?=?

    } else if (t == TOKENIZER_SEMICOLON) {
		do {tokenizer_next();
		    if (is_comma()) {strcat(buf, ","); tokenizer_next();}  // ?a;,b == ?a;",";b --> a,b 
		   } while (is_cr());
    } else if (t == TOKENIZER_COMMA) {
		do {tokenizer_next();
		    if (is_comma()) {strcat(buf, ","); tokenizer_next();}
			strcat(buf, " ");
		} while (is_cr());
    } else if(t == TOKENIZER_LABEL) {tokenizer_next();
		//for (u=0; u<=strlen(buf); ++u) {if (buf[u]=='^' && buf[u+1]!='^') buf[u]='"';}
		if (scol) strcat(buf,"\n"); else
		if (pscr) writeln(buf); else {
		if (echo) console_add_line(buf); else
		script_console_add_line((long)buf); };
		if (!scol) buf[0]='\0';
		if (eos() || is_token(TOKENIZER_RIGHTPAREN))
           {strcat(buf,":---=----=--+-=----=----"); break;}
    } else if (eos()
		|| t == TOKENIZER_MOMENT
		|| t == TOKENIZER_ENDOFINPUT) { break;
    // string additions
    } else if (t == TOKENIZER_READ$) {
           sprintf(buf+blen, "%s", sfactor(0,t));
    } else if (t == TOKENIZER_LEFTPAREN
            && tokenizer_stringlookahead()==1) { t=expr();
           if ( svar_selected==0)
                sprintf(buf+blen, "%d", t);
           else { if (t==0) {
                svar_selected=0;
                sprintf(buf+blen, "%s", string_selected);}
                  else 
                sprintf(buf+blen, "%s", "<<Need sleep.");
           }
    } else if (tokenizer_stringlookahead()==1) {
           sprintf(buf+blen, "%s", sexpr(0));
    } else if (t == TOKENIZER_RIGHTPAREN) { break; // scol
	  // end of string additions
    } else {sprintf(buf+blen, "%d", expr());}

	if (ended | eos()) break;
    t=tokenizer_token();
  } while ( t != TOKENIZER_LET
		&&  t != TOKENIZER_DATA);
  str_len=strlen(buf);
 // for (u=0; u<=strlen(buf); ++u)  {if (buf[u]=='^' && buf[u+1]!='^') buf[u]='"';} // "^up^" --> "up"
  
// string additions
  stringvariables[MAX_SVARNUM-2]=store(buf); // ##$
  if (t == TOKENIZER_RIGHTPAREN && scol) return; // str$(A$,...)
// string additions
	if (pscr) writeln(buf);
	if (!scol) {
		buf[MAX_STRINGLEN-1]=0; // truncate
		if (echo) console_add_line(buf); else
			script_console_add_line((long)buf);	
    }
  DEBUG_PRINTF("End of print\n");
  if ( t == TOKENIZER_MOMENT) accept(0);
  t=tokenizer_token();
  if ( !eos()
	&& t != TOKENIZER_DATA
	&& t != TOKENIZER_LET) accept_cr();
  flag_yield=1; //:...i....1....;....2....;....3....;....4..
}
/*---------------------------------------------------------------------------*/

static void jump_linenum(int linenum)
{
  unsigned short li=ubasic_linenumber();
  if (linenum==li) {return;} // ignore
  if (linenum<li) { linenum=abs(linenum);
	tokenizer_init(program_ptr); }
  while (!is_token(TOKENIZER_NUMBER) || expr() != linenum) {
   /*  Original originell, Speicherjogging?
		//tokenizer_next();
    } //while(tokenizer_token() != TOKENIZER_CR &&
		//	tokenizer_token() != TOKENIZER_ENDOFINPUT);*/
    tokenizer_eol();
	if (is_cr()) {tokenizer_next();}
	else{
	//if (tokenizer_token() == TOKENIZER_ENDOFINPUT)
		if (camera_info.state.state_kbd_script_run == 1) {  
			sprintf(buf, " @%d --> %d",li, linenum);
			script_console_add_error((long)buf);
			DEBUG_PRINTF("Label %d not found", linenum);
			ubasic_error = UBASIC_E_UNK_LABEL;
		ended=1; return;}
	}
  DEBUG_PRINTF("jump_linenum: Found line %d\n", tokenizer_num());
  }
}
/*---------------------------------------------------------------------------*/
static void jump_line(unsigned short linenum)
{
  //<< to visited lines only // next, until, wend, return, (reader<<)
  if (linenum<ubasic_linenumber()) 
     tokenizer_init(program_ptr);
  while(tokenizer_line_number() < linenum) 
  {tokenizer_eol(); tokenizer_next();}
  //* swallow the CR that would be read next
  while (is_cr()) tokenizer_next();
///
  if (traconly>9)
  stopwatch(ubasic_linenumber());
}
/*---------------------------------------------------------------------------*/
// TODO: error handling?
int jump_label(char * label)
{
  unsigned short li=ubasic_linenumber();
  char currLabel[MAX_STRINGLEN];
if (go_fwd>=0) {
  if (!go_fwd) tokenizer_init(program_ptr);
  currLabel[0] = 0;
  do  {
	tokenizer_eol(); tokenizer_next();
    if (is_token(TOKENIZER_LABEL)) {
        tokenizer_label(currLabel, sizeof(currLabel));
      if(strcmp(label, currLabel) == 0) {
        DEBUG_PRINTF("jump_linenum: Found line %d\n", tokenizer_line_number());
		tokenizer_eol();
       //= accept(TOKENIZER_CR);
        break;
      }
    }
  } while(!is_eof());
  }
  if (is_eof() || go_fwd<0) {
    if (camera_info.state.state_kbd_script_run == 1) {  
		sprintf(buf, "@%d :-> '!%s'",li, label);
		script_console_add_error((long)buf);
      DEBUG_PRINTF("Label %s not found", label);
      ubasic_error = UBASIC_E_UNK_LABEL;
	  go_fwd--; ended=1;}
      return 0;
  } else {
      return 1;
  }
}
/*---------------------------------------------------------------------------*/
static void goto_statement(short token)
{ int li, ln=0;
  short re=token==TOKENIZER_UBASIC_READER;
  if (re) {pause=10; traconly=100;}
  accept(token);
  while (on_go && !is_cr()) { on_go--;
     if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();
     if (is_comma()) tokenizer_next();}
  if (is_cr()) { if (re) return;
		DEBUG_PRINTF("ubasic.c: goto_statement(): no label specified\n");
		ended = 1;
		ubasic_error = UBASIC_E_UNK_LABEL;}
  if (ended) return;

  if (is_token(TOKENIZER_STRING)) { ln=0xBA5EBA11;
      tokenizer_string(string, sizeof(string));
      go_fwd = string[0]=='>'; 
	  accept(TOKENIZER_STRING);}
  else {ln=expr();}
  if (re && is_comma()) {tokenizer_next(); pause=expr();}
  while (is_comma()) { tokenizer_next(); 
        if (is_token(TOKENIZER_STRING)) tokenizer_next(); else expr();}
///
  data_statement();
  if (ln==0) return; // continue / ignore
  
  if (token == TOKENIZER_EXIT) {
      for_stack_ptr = gosub_stack_ptr = while_stack_ptr = do_stack_ptr = 0;
      if_stack_ptr = select_stack_ptr = break_stack_ptr =0;
      tokenizer_pop(0x1000);}

  if (token == TOKENIZER_GOSUB) { li=tokenizer_line_number();
  if (gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
      gosub_stack[gosub_stack_ptr] = li+1;
      gosub_stack_ptr++;
      if (gosub_stack_ptr == MAX_GOSUB_STACK_DEPTH) varlist();  // :)
      break_stack[break_stack_ptr] = TOKENIZER_RETURN;
      break_stack_ptr++;
#ifdef PUSH_POP
        tokenizer_push();
#endif
  } else {
      DEBUG_PRINTF("gosub_statement: gosub stack exhausted\n");
      ended = 1;
      ubasic_error = UBASIC_E_GOSUB_STACK_EXHAUSTED;}
  if (re && is_cr()) console_set_layout(0,0,45,12);
  } // gosub

  if (ln==0xBA5EBA11) {jump_label(string);} else {jump_linenum(ln);}  // :)
}
/*---------------------------------------------------------------------------*/
static void return_statement(void) // breaks loops before, eats endif...
{ short t;
  accept(TOKENIZER_RETURN);
  data_statement();
  if (gosub_stack_ptr > 0) { 
    do { break_stack_ptr--;
	   t=break_stack[break_stack_ptr];
	   if (t == TOKENIZER_NEXT) {for_stack_ptr--;	tokenizer_pop(1);}
	   if (t == TOKENIZER_UNTIL) {do_stack_ptr--;	tokenizer_pop(1);}
	   if (t == TOKENIZER_WEND) {while_stack_ptr--;	tokenizer_pop(1);}
	   if (t == TOKENIZER_ENDIF) if_stack_ptr--;
       if (t == TOKENIZER_END_SELECT) select_stack_ptr--;
	} while (t != TOKENIZER_RETURN);
    gosub_stack_ptr--;
    tokenizer_pop(-1);
#ifndef PUSH_POP
    jump_line(gosub_stack[gosub_stack_ptr]);
#endif
  } else {
    DEBUG_PRINTF("return_statement: non-matching return\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_RETURN;
  }
}
/*---------------------------------------------------------------------------*/
static void on_statement(void) // on x=0,1,2,3 goto 0,100,20,"3xxx" [,"default"]
{
  int e,v;
  on_go=0;
  accept(TOKENIZER_ON);
  short var = tokenizer_variable_num();
  if (var==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);
       v = variables[var];}
  else v = varfactor(var);
  accept(TOKENIZER_EQ);
  do {e=expr();
     if (!is_comma()) break;
     if (e==v) break;
	 on_go++; tokenizer_next();
	 } while (1);
  while (is_comma()) {tokenizer_next(); expr();}
  if (e!=v) on_go++;
}
/*---------------------------------------------------------------------------*/
static void next_statement(void)
{
  short var, value;
  
  accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_NEXT
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 &&
    var == for_stack[for_stack_ptr - 1].for_variable) {
    value = variables[var] + for_stack[for_stack_ptr - 1].step;
    ubasic_set_variable(var, value);
    
    if(((for_stack[for_stack_ptr - 1].step > 0) && (value <= for_stack[for_stack_ptr - 1].to)) ||
       ((for_stack[for_stack_ptr - 1].step < 0) && (value >= for_stack[for_stack_ptr - 1].to))) {
        data_statement();
        tokenizer_pop(0);
#ifndef PUSH_POP
    	jump_line(for_stack[for_stack_ptr - 1].line_after_for);
#endif
    } else { 
      for_stack_ptr--;
      break_stack_ptr--;
      tokenizer_pop(1);
  //    tokenizer_eol();
    }
  } else {
    DEBUG_PRINTF("next_statement: non-matching next (expected %d, found %d)\n", for_stack[for_stack_ptr - 1].for_variable, var);
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_NEXT;
  }

}
/*---------------------------------------------------------------------------*/
static void for_statement(void)
{
  unsigned short for_variable;
  int fo, to, step=1;
  unsigned short li=tokenizer_line_number();
  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  fo=expr();
  ubasic_set_variable(for_variable, fo);
  if (is_comma()) {accept(TOKENIZER_COMMA);}  else {accept(TOKENIZER_TO);}
  to = expr();
  if (!is_cr() && !is_token(TOKENIZER_DATA)) {
      if (is_comma()) accept(TOKENIZER_COMMA); else accept(TOKENIZER_STEP); ///  :)
	  step = expr();
  }

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = li+1;
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    for_stack[for_stack_ptr].step = step;
    DEBUG_PRINTF("for_statement: new for, var %d to %d\n",
		 for_stack[for_stack_ptr].for_variable,
		 for_stack[for_stack_ptr].to);
	
    break_stack[break_stack_ptr] = TOKENIZER_NEXT;
    for_stack_ptr++;
    break_stack_ptr++;
#ifdef PUSH_POP
    tokenizer_push();
#endif
	// BUG!?  for i=1 to 0  etc. WILL RUN ONCE in old uBasic! not in LUA 
  if ((fo>to && step>0) || (fo<to && step<0)) {break_statement(); return;}
  // :...i....1....;....2....;....3....;....4..
  data_statement();
  } else {
    DEBUG_PRINTF("for_statement: for stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_FOR_STACK_EXHAUSTED;
  }
}
/*---------------------------------------------------------------------------*/
static void do_statement(void)
{
  if(do_stack_ptr < MAX_DO_STACK_DEPTH) {
     do_stack[do_stack_ptr] = tokenizer_line_number()+1;
     break_stack[break_stack_ptr] = TOKENIZER_UNTIL;
     do_stack_ptr++;
     break_stack_ptr++;
     accept(-TOKENIZER_DO);
     data_statement();
#ifdef PUSH_POP
     tokenizer_push();
#endif
  } else {
    DEBUG_PRINTF("do_statement: do stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_DO_STACK_EXHAUSTED;
  }
}
/*---------------------------------------------------------------------------*/
static void until_statement(void) // do=repeat
{
  int r;
  
  if(do_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_UNTIL
    r = relation();
    if(!r) {	data_statement();
	  tokenizer_pop(0);
#ifndef PUSH_POP
      jump_line(do_stack[do_stack_ptr-1]);
#endif
    } else {
      do_stack_ptr--;
      break_stack_ptr--;
	  tokenizer_pop(1);
      tokenizer_eol(); // --> next line
    }
  } else {
    DEBUG_PRINTF("until_statement: unmatched until\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_UNTIL;
  }
}
/*---------------------------------------------------------------------------*/
static void while_statement(void)
{
  int r;
#ifdef PUSH_POP
  tokenizer_push();
#endif
  accept(-TOKENIZER_WHILE);
  if(while_stack_ptr < MAX_WHILE_STACK_DEPTH) {
    if ((while_stack_ptr == 0)
	 ||((while_stack_ptr > 0) && (while_stack[while_stack_ptr-1] != tokenizer_line_number()))){
      while_stack[while_stack_ptr] = tokenizer_line_number();
      break_stack[break_stack_ptr] = TOKENIZER_WEND;
      while_stack_ptr++;
      break_stack_ptr++;
    }
  } else {
    DEBUG_PRINTF("while_statement: while stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_WHILE_STACK_EXHAUSTED;
    return;
  }

  r = relation();
  if(while_stack_ptr > 0) {
    if (r) data_statement();
    else {tokenizer_eol(); break_statement();} // --> line after wend
  } else {
    DEBUG_PRINTF("while_statement: unmatched wend\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_WEND;
  }
}
/*---------------------------------------------------------------------------*/
static void wend_statement(void)
{
  if(while_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_WEND
	data_statement();
	tokenizer_pop(-1);
#ifndef PUSH_POP
    jump_line(while_stack[while_stack_ptr-1]);
#endif
	while_statement();
  } else {
    DEBUG_PRINTF("wend_statement: unmatched wend\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_WEND;
  }
}
/*---------------------------------------------------------------------------*/ 
static void break_statement(void)
{
  register short t,tt,token, i=1;
  short tok = tokenizer_token();
  if (tok ==TOKENIZER_END_SELECT )  {return;}
  short cont= tok==TOKENIZER_CONT;
  if (break_stack_ptr > 0) {
	 while (break_stack[break_stack_ptr-1]==TOKENIZER_ENDIF)
	      { break_stack_ptr--; if_stack_ptr--; }
  }
  if (break_stack_ptr > 0) {
	token=(break_stack[break_stack_ptr-1]);
	 if ( cont || tok==TOKENIZER_BREAK )
		{ accept(0); data_statement(); }
	 if ( token == TOKENIZER_UNTIL) { t=TOKENIZER_DO;}
	 if ( token == TOKENIZER_NEXT ) { t=TOKENIZER_FOR;}
	 if ( token == TOKENIZER_WEND ) { t=TOKENIZER_WHILE;}
	 if ( token == TOKENIZER_END_SELECT ) {t=TOKENIZER_SELECT;}
	 
	 do {	tokenizer_next();
		tt=tokenizer_token();
		if ( tt == t ) { i++; } else 
		if ( tt == token) { i--; } else 
		if ( tt == TOKENIZER_CASE) accept_cr(); else  //	 eat exec {...}
		if ( tt == TOKENIZER_DATA 
		  || tt == TOKENIZER_LET
		  || tt == TOKENIZER_DEF_FN
		  || tt == TOKENIZER_MD_DETECT_MOTION) tokenizer_eol(); else 
		if ( tt == TOKENIZER_ENDOFINPUT) { ended=1; return; }
	} while (i>0);
	
	while (cont) { cont=0;
	  if ( token == TOKENIZER_WEND && while_stack_ptr==0) {token=0; break;} //
	  if ( token == TOKENIZER_NEXT
		|| token == TOKENIZER_WEND 
		|| token == TOKENIZER_UNTIL) { statement();	return;}
	}
	 if ( token == TOKENIZER_END_SELECT ) { return;}
    break_stack_ptr--;
	 if ( token == TOKENIZER_NEXT ) { for_stack_ptr--; tokenizer_pop(1);}
	 if ( token == TOKENIZER_UNTIL) { do_stack_ptr--; tokenizer_pop(1);}
	 if ( token == TOKENIZER_WEND ) { while_stack_ptr--; tokenizer_pop(1);}
	 accept_cr();
  } else {
      DEBUG_PRINTF("break_statement: NO do for while \n");
      ended = 1;
      ubasic_error = UBASIC_E_UNK_LABEL;}
}
/*---------------------------------------------------------------------------*/
/* IF-STATEMENT                                                              */

static void
endif_statement(void)
{
  if(if_stack_ptr > 0) {
    accept(break_stack[break_stack_ptr-1]);  // expected: TOKENIZER_ENDIF
    if_stack_ptr--;
    break_stack_ptr--;
  } else {
    DEBUG_PRINTF("ubasic.c: endif_statement(): endif without if-statement\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_IF;
  }
}
/*---------------------------------------------------------------------------*/
static void else_statement(void)
{
  short r=0, endif_cntr, f_nt, d0;
  
  if(if_stack_ptr > 0) {
    r = if_stack[if_stack_ptr-1];
  }
  else{
    DEBUG_PRINTF("ubasic.c: else_statement(): else without if-statement\n");
	play_sound(0);
    ended = 1;
    accept(TOKENIZER_IF);
    ubasic_error = UBASIC_E_PARSE;
    return;
  }
  accept(TOKENIZER_ELSE);
  DEBUG_PRINTF("else_statement: relation %d\n", r);
  
  d0=(is_token(TOKENIZER_DO));
  if (d0) {if (r) tokenizer_eol();
	  do_statement();}
  if (is_cr()) {
	  accept(TOKENIZER_CR);
	  if(!r) {
      DEBUG_PRINTF("else_statement: result true\n");
      return;
    } else { // eat the rest...
      DEBUG_PRINTF("else_statement: result false\n");
	if (d0) {break_statement(); return;}
      endif_cntr=0;
      while(((!is_token(TOKENIZER_ENDIF ))
           || endif_cntr) && tokenizer_token() != TOKENIZER_ENDOFINPUT){
        f_nt=0;
        if( is_token(TOKENIZER_IF)) { //!tokenizer_skip(TOKENIZER_THEN);
          endif_cntr+=1;
          if (debug)
			{accept(TOKENIZER_IF); relation(); ended=0;} // div by unset var?
        } // short IF
		if( is_token(TOKENIZER_THEN)) { tokenizer_next();
          if (is_token(TOKENIZER_DO))
			 {tokenizer_eol(); do_statement();
			  break_statement(); tokenizer_sep();}
		// then followed by CR -> multi line
          if (is_cr()) { f_nt=1;
          } else { // single line
            endif_cntr--;
            while(!is_token(TOKENIZER_ENDIF) && !is_cr() && !is_eof()){
               tokenizer_next();
            }
            if (is_token(TOKENIZER_ENDIF)){
              DEBUG_PRINTF("ubasic.c: else_statement(): endif in singleline if-statement\n");
              ended = 1;
			  ubasic_error = UBASIC_E_UNMATCHED_IF;
              return;
            }
          }
        }
        if( debug && is_token(TOKENIZER_ELSEIF)) { accept(TOKENIZER_ELSEIF); relation();}
        if( is_token(TOKENIZER_ENDIF))  { endif_cntr--; }
        if (!f_nt) { tokenizer_next(); }
      }
    }
    endif_statement();
  }else{ //statement();

  //  DEBUG_PRINTF("ubasic.c: else_statement(): CR after ELSE expected\n");
  //  ended = 1;
  //  ubasic_error = UBASIC_E_PARSE;
  }
}

/*---------------------------------------------------------------------------*/
static void if_statement(short token)
{
  int r,s=0;
  short t, else_cntr,endif_cntr,f_nt,f_sl;
  
  accept(token); 
	  
  DEBUG_PRINTF("if_statement: get_relation\n");
  r = relation();
  DEBUG_PRINTF("if_statement: relation %d\n", r);
  accept(-TOKENIZER_THEN);
  if (ended) {return;}

  if (is_cr()) { f_nt=1;
    // CR after then -> multiline IF-Statement
    if(if_stack_ptr < MAX_IF_STACK_DEPTH) {
	  if (is_token(TOKENIZER_ELSEIF)) {
		if_stack_ptr--;
        s=if_stack[if_stack_ptr];
      } else {
        break_stack[break_stack_ptr]=TOKENIZER_ENDIF;
        break_stack_ptr++; }
      if (s) r=0; else if_stack[if_stack_ptr] = r;
      if_stack_ptr++;
    } else {
      DEBUG_PRINTF("if_statement: IF-stack depth exceeded\n");
      ended = 1;
      ubasic_error = UBASIC_E_IF_STACK_EXHAUSTED;
      return;
    }
    DEBUG_PRINTF("if_statement: stack_ptr %d\n", if_stack_ptr);
    accept(TOKENIZER_CR);
    if(r) {
      DEBUG_PRINTF("if_statement: result true\n");
      return;

    }else {
      DEBUG_PRINTF("if_statement: result false\n");

      else_cntr=endif_cntr=0; // number of else/endif possible in current nesting
      f_nt=f_sl=0; // f_nt flag for additional next token, f_fs flag single line

	  t=tokenizer_token();
      while( ((t != TOKENIZER_ELSE && t != TOKENIZER_ELSEIF && t != TOKENIZER_ENDIF)
                 || else_cntr || endif_cntr)
			&& t != TOKENIZER_ENDOFINPUT) {
        f_nt=0;
        // nested if
        if( t == TOKENIZER_IF) {  /// if (debug)
			{accept(t); relation(); ended=0;} // div by unset var?
            //!tokenizer_skip(TOKENIZER_THEN);
      //    else_cntr+=1;//;....4..
          endif_cntr+=1;
          f_sl=0;
          DEBUG_PRINTF("IF: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          f_nt=1;
          accept( -TOKENIZER_THEN);
          if (tokenizer_token() == TOKENIZER_DO)
			 {tokenizer_eol(); do_statement();
			  break_statement(); tokenizer_sep();}
          DEBUG_PRINTF("THEN: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          if (!is_cr()) { f_sl=1; tokenizer_eol();} 
          else {else_cntr+=1;}  //;....4..
          DEBUG_PRINTF("THEN_SL: line %d, token %d, else %d, end %d\n",
                tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
        } // long IF
        if (is_token(TOKENIZER_ELSE)) {
          else_cntr--;
          DEBUG_PRINTF("ELSE: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),tokenizer_token(),else_cntr,endif_cntr);
          if (else_cntr<0) { 
            DEBUG_PRINTF("ubasic.c: if_statement(): else without if-statement\n");
            ended = 1;
            ubasic_error = UBASIC_E_UNMATCHED_IF;
            return;
          }
        }
        if(!f_sl && (is_token(TOKENIZER_ENDIF))) {tokenizer_eol();
          endif_cntr--;
          if (endif_cntr != else_cntr) else_cntr--;
          DEBUG_PRINTF("ENDIF: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),t,else_cntr,endif_cntr);
        } else {
          if (f_sl && (is_cr()))  {
            f_sl=0;
            endif_cntr--;
            if (endif_cntr != else_cntr) else_cntr--;
            DEBUG_PRINTF("ENDIF_SL: line %d, token %d, else %d, end %d\n", tokenizer_line_number(),t,else_cntr,endif_cntr);
          }else{
            if (is_token(TOKENIZER_ENDIF)) { accept(TOKENIZER_CR)
              DEBUG_PRINTF("ubasic.c: if_statement(): endif in singleline if-statement\n");
              ended = 3;
              ubasic_error = UBASIC_E_PARSE;
              return;}
          }
        }
        if (!f_nt) { tokenizer_next(); }
	  t=tokenizer_token();
      } // while (t !=)...
      if (t == TOKENIZER_ELSE) {else_statement(); return;}
      if (t == TOKENIZER_ELSEIF) {
         if_statement(TOKENIZER_ELSEIF); return;}
    }  // !r
    endif_statement();
  }else {
  // Singleline IF-Statement
  // Now possible: if a=3 then a=33~b=3~c=3 else a=111 ~ b=222~ c=333
	if(r) { //statement(); 
		return;}
//  } else {	// !r
	do  {tokenizer_next();
	   t=tokenizer_token();
		if  (t == TOKENIZER_ELSEIF) { 
             if_statement(TOKENIZER_ELSEIF); return;}
		if  (t == TOKENIZER_ELSE) {accept(TOKENIZER_ELSE); return;}
		if  (t == TOKENIZER_DO)
			{tokenizer_eol(); do_statement();
			break_statement(); tokenizer_sep();} //  select, for & while could follow next line}
	//	if  (t == TOKENIZER_EXEC) {tokenizer_next(); tokenizer_sep();} // eaten by tok.wskip();
	}while (!is_cr() && !is_eof());	// !r
  }
}

/*---------------------------------------------------------------------------*/
/* SELECT-STATEMENT                                                          */
/*---------------------------------------------------------------------------*/
static void
dec_select_stack(void)
{
  if(select_stack_ptr > 0) {
      select_stack_ptr--;
      break_stack_ptr--;
  } else {
    DEBUG_PRINTF("select_statement: SELECT-Stack fail\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
  }
}
/*---------------------------------------------------------------------------*/
static void end_select_statement(void)
{
  if(select_stack_ptr > 0) {
	accept(break_stack[break_stack_ptr-1]); //TOKENIZER_END_SELECT
	tokenizer_eol();
    dec_select_stack();
  } else {
    DEBUG_PRINTF("ubasic.c: end_select_statement(): end_select without select-statement\n");
    ended = 1;
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
  }
}
/*---------------------------------------------------------------------------*/
static void case_statement(void)
{
  int select_value, case_value_1, case_value_2, case_value_eq=1;
  short t, case_run, case_gosub = 0;
  unsigned short cur_ln, gosub_ln = 0;
  
  if(select_stack_ptr > 0) {
    select_value = select_stack[select_stack_ptr - 1].select_value;
    case_run = select_stack[select_stack_ptr - 1].case_run;

    while (!case_run) { 
	t=tokenizer_token();
	if (t == TOKENIZER_CASE_ELSE)
		{case_value_eq=1; accept(TOKENIZER_CASE_ELSE);}
	else {
	  if (t == TOKENIZER_END_SELECT) { break;}
	  accept(TOKENIZER_CASE);
      if (ended) return;
      case_value_1 = expr();
      case_value_eq = (select_value == case_value_1);
      if (case_value_eq) { DEBUG_PRINTF("case_statement: case_value_eq %d, case_value %d\n", case_value_eq, case_value_1); }  

      if(is_token(TOKENIZER_TO)) {
        accept(TOKENIZER_TO);
        case_value_2 = expr();
        if (case_value_1 < case_value_2) {
          case_value_eq = ((select_value >= case_value_1) && (select_value <= case_value_2));
          DEBUG_PRINTF("case_statement: case_value %d to %d\n", case_value_1, case_value_2);
        } else {
          case_value_eq = ((select_value >= case_value_2) && (select_value <= case_value_1));
          DEBUG_PRINTF("case_statement: case_value %d to %d\n", case_value_2, case_value_1);
        }
      } else if (is_comma()) {
        do {
          accept(TOKENIZER_COMMA);
          if (case_value_eq) {
            case_value_2 = expr();
          } else {
            case_value_1 = expr();
            case_value_eq = (select_value == case_value_1);
          }
        } while (is_comma());
        DEBUG_PRINTF("case_statement: case_value_eq %d, case_value_comma %d\n", case_value_eq, case_value_1);
      }
      if (is_token(TOKENIZER_LABEL)) accept (TOKENIZER_LABEL); // :
      else accept(TOKENIZER_SEMICOLON);
	}  
      if (case_value_eq) {
        //case_gosub = (tokenizer_token() == TOKENIZER_GOSUB);
//GOSUB - save curr linenumber
        cur_ln = tokenizer_line_number();
//GOSUB
		do { t=tokenizer_token();
        switch(t) {
			case TOKENIZER_EXEC: {break;}; // breakstm OK {...here}
			case TOKENIZER_GOSUB:
				{case_gosub = 1;     goto_statement(t); break;};
			case TOKENIZER_EXIT:
			case TOKENIZER_GOTO:
				{dec_select_stack(); goto_statement(t); return;};
            default: statement();
            }  // switch
		}
		while (is_token(TOKENIZER_SEP));
//GOSUB  - save new linenumber, reset to curr linenumber
      if (case_gosub) {
        gosub_ln = ubasic_linenumber();
        tokenizer_pop(-1); 
#ifndef PUSH_POP
        jump_line(cur_ln);
#endif
        DEBUG_PRINTF("case_statement: GOSUB: toLN=%d, nextLN=%d\n", gosub_ln, cur_ln+1);
      }
//GOSUB
        DEBUG_PRINTF("case_statement: case execute\n");
        case_run = 1;
        select_stack[select_stack_ptr - 1].case_run = case_run;
      } else { // !eq
        DEBUG_PRINTF("case_statement: case jump; case_run: %d\n", case_run);
		if  (is_token(TOKENIZER_DO)) {tokenizer_eol();
			do_statement();
			break_statement();} //  select, for & while could follow next line}
		if  (is_token(TOKENIZER_EXEC)) accept_cr(); // eaten by tok.wskip();
        tokenizer_eol();
		tokenizer_next();
      }
//REM
    while (is_cr())	{tokenizer_next();}
//REM
	} // case_run
	  t=tokenizer_token();
      if ((t != TOKENIZER_CASE)
	   && (t != TOKENIZER_EXEC)
	   && (t != TOKENIZER_CASE_ELSE)
       && (t != TOKENIZER_END_SELECT)) {
         DEBUG_PRINTF("ubasic.c: select_statement(): don't found case, case_else or end_select\n");
         ended = 1;
        // ubasic_error = UBASIC_E_PARSE;
		ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
      } else { 
//GOSUB test for end_select and set to gosub-
        if (case_gosub) break_stack_ptr--;
        break_statement(); // dont 'break|cont' in script.bas!
        end_select_statement();
        if (case_gosub) {
          gosub_stack[gosub_stack_ptr-1] = tokenizer_line_number()+1; //+1;
#ifdef PUSH_POP
          tokenizer_push();
#endif
          break_stack[break_stack_ptr] = TOKENIZER_RETURN;
          break_stack_ptr++;
          jump_line(gosub_ln);
          DEBUG_PRINTF("end_select_statement: GOSUB: returnLN=%d\n", gosub_stack[gosub_stack_ptr-1]);
        }
//GOSUB        
      }  
  } else { ended = 1;
    DEBUG_PRINTF("case_statement: SELECT-Stack fail\n");
    ubasic_error = UBASIC_E_UNMATCHED_END_SELECT; }
}
/*---------------------------------------------------------------------------*/
static void select_statement(void)
{
  accept(TOKENIZER_SELECT);
  while (on_go) { on_go--;
     expr(); if (is_comma()) tokenizer_next();}
  int select_value = expr(); // I+x
  accept(0);
  if (ended) return;
  
//REM:...,....1....,.
  //  while (tokenizer_token() == TOKENIZER_REM)
	//	{tokenizer_eol(); tokenizer_next();}
//REM
  
  if(select_stack_ptr < MAX_SELECT_STACK_DEPTH) {
    select_stack[select_stack_ptr].select_value = select_value;
    select_stack[select_stack_ptr].case_run = 0;
    DEBUG_PRINTF("select_statement: new select, value %d\n",select_stack[select_stack_ptr].select_value);
	break_stack[break_stack_ptr]=TOKENIZER_END_SELECT;
    select_stack_ptr++;
    break_stack_ptr++;
	
	tokenizer_skip(TOKENIZER_CASE); //:...,....1....,.
    if (!is_token(TOKENIZER_CASE)) {
      DEBUG_PRINTF("ubasic.c: select_statement(): don't found case-statement\n");
      ended = 1;
      //ubasic_error = UBASIC_E_PARSE;
      ubasic_error = UBASIC_E_UNMATCHED_END_SELECT;
    }
    else { }///case_statement(); }
  } else {
    DEBUG_PRINTF("select_statement: SELECT-stack depth exceeded\n");
    ended = 1;
    ubasic_error = UBASIC_E_SELECT_STACK_EXHAUSTED;
  }
}
/* SELECT-STATEMENT END                                                      */
/*---------------------------------------------------------------------------*/

static void let_statement(void)
{
  int   t=0, v=-1, val=0, x=0;
  short var, sep, str, far, token;
  
if (is_token(TOKENIZER_LET)) accept(0);
do {
	token=tokenizer_token();
    str = (token>TOKENIZER_STRINGVARIABLE && token<=TOKENIZER_CHR$);
    if (str) {
       accept(token);
	   accept(TOKENIZER_LEFTPAREN);
    }
	far = token==TOKENIZER_FIELD;
	if (far) { accept(token);       var = relation()%MAX_FIELD;
    accept(TOKENIZER_RIGHTPAREN);   v = field[var];}
    else {var = tokenizer_variable_num();
      if (var ==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);  // RIGHTPAREN--
            if (!str) v = variables[var];}
      else  if (!str) v = varfactor(var);
    }
// string additions here
    str |= token==TOKENIZER_STRINGVARIABLE;
    accept(-TOKENIZER_STRINGVARIABLE);

if (str) {
  static char *r, *s;
  s=stringvariables[var];
  r=stringvariables[MAX_SVARNUM-1];
  int co=0, len=strlen(s);
  if (token > TOKENIZER_STRINGVARIABLE)
	 { co=is_comma(); if (co) accept(TOKENIZER_COMMA); }
  switch (token) {
    case TOKENIZER_CHR$:
    case TOKENIZER_SPC$: x=expr()&255; 
    case TOKENIZER_STR$:
    case TOKENIZER_READ$: {
      if (ended) return;
	  switch (token) {
        case TOKENIZER_STR$: { 
             accept(-TOKENIZER_EQ);
             print_statement(TOKENIZER_STRLEN);
             r=store(stringvariables[MAX_SVARNUM-2]);
			 break;}
        case TOKENIZER_CHR$: r=schr(x); break;
        case TOKENIZER_SPC$:
             memset(sbuf,' ',x); r[x]=0;
             r=store(sbuf); break;
        case TOKENIZER_READ$:
             sfactor(0,TOKENIZER_READ$);
             r=store(sbuf); break;
	  } // switch
      if (ended) return;
      ubasic_set_stringvariable(var,r);
      accept(-TOKENIZER_RIGHTPAREN);
	  break;}
    case TOKENIZER_LEFT$: { if (co) x=expr();
      if (x<0) x=len-x;
      if (ended) return;
      if (x>len) x=len;
      accept(TOKENIZER_RIGHTPAREN);
      if (is_token(TOKENIZER_EQ)) {
		  tokenizer_next(); r = sexpr(0);
         if (ended) return;
         if (!co) x=strlen(r);
         if (x>len) x=len;
         t=strlen(r);
         if (x>t)  x=t;
         for (v=0; v<x; v++) s[v]=*(r+v);
	  }
      else { if (len>x) s[x]=0; }
      break;}
    case TOKENIZER_RIGHT$: { if (co) x=expr();
	  t=x;
      if (x>0) x=len-x; else x=len+x;
      if (ended) return;
      accept(TOKENIZER_RIGHTPAREN);
      if (x>len) x=len;
      if (is_token(TOKENIZER_EQ)) { x=t;
		  tokenizer_next(); r = sexpr(0);
         if (ended) return;
         sep=strlen(r);
         if (co) x=len-x; else x=len-sep;
         if (x>len-sep) x=len-sep;
         if (x<0) x=0;
	     t=0; for (v=x; v<x+sep; v++)
			  if (*(r+t)) {s[v]=*(r+t); t++;}
      } else s=sshl(s,x);
      break;}
    case TOKENIZER_MID$: { // mid$(a%,t,x)=b$ // replace x chars @ pos t with left$(b$,x)
      x=expr(); if (ended) return;
      if (x<0) x=len-x;
      sep=len-x; len-=x;
      if (is_comma()) { accept(TOKENIZER_COMMA);
          sep=expr(); if (sep<0) sep=len-sep;}
	  if (ended) return;
      accept(TOKENIZER_RIGHTPAREN);
      if (is_token(TOKENIZER_EQ)) {
		  tokenizer_next(); r = sexpr(0);
         if (ended) return;
         t=strlen(r); if (sep>t) sep=t;
     //    if (x+t>len) {/*accept(TOKENIZER_LEN);*/ return;}
         if (len<=x+sep) break;
         t=0; for (v=x; v<x+sep; v++) {s[v-1]=*(r+t); t++;}
      } else { s[x+sep]=0; s=sshl(s,x);}
      break;}
    case TOKENIZER_UP$:
    case TOKENIZER_LOW$:
    case TOKENIZER_REP$:
    case TOKENIZER_REV$:
    case TOKENIZER_TRIM$:
    case TOKENIZER_STROL$:
    case TOKENIZER_STROR$: if (len) sfactor(var+1,token);
      if (ended) return;
      accept(TOKENIZER_RIGHTPAREN);
      break;

   default: { accept(-TOKENIZER_STRINGVARIABLE);
   sep=tokenizer_token();
   switch (sep) {
    case TOKENIZER_EQ: { accept(TOKENIZER_EQ);
       if (is_token(TOKENIZER_FILE_BROWSER)) {
           svar_selected=var+1; file_browser(); break;}   // string_selected;
       if (is_token(TOKENIZER_TEXTBOX)) {
           svar_selected=var+1; textbox();      break;}   // string_selected;
       if (is_comma()) ubasic_set_stringvariable(var,"\0");
       else sexpr(var+1);
       break;}
    case TOKENIZER_PLUS: { tokenizer_inc();
       if (tokenizer_char()==TOKENIZER_EQ) sexpr(var+0x1000);
 //      ubasic_set_stringvariable(var,sconcat(stringvariables[var], sexpr(0)));
       break; }
    case TOKENIZER_PRINT: {
       if (tokenizer_char()!=TOKENIZER_PRINT) return; else accept(0);
       if (*(stringvariables[var])!=0) { //  	I$ ? y$:n$ == wenn I<>'', I=x sonst I=y
          sexpr(var+1); //==ubasic_set_stringvariable(var, sexpr(0));
          accept(TOKENIZER_LABEL); sfactor(0,TOKENIZER_SKIP); break;}
       tokenizer_skip(TOKENIZER_LABEL); accept(TOKENIZER_LABEL);
          sexpr(var+1); //==ubasic_set_stringvariable(var, sexpr(0));
       break;}
    DEBUG_PRINTF("let_statement: string assign '%s' to %d\n", stringvariables[var], var);
   } // switch
  } // default
 } // switch
  if (ended) return;

} else { // !str 
  sep=tokenizer_token();
  switch (sep) {
    case TOKENIZER_EQ: {accept(0);
		 val=expr(); break;}
    case TOKENIZER_PRINT:  accept(0);
      if (v) { val = expr(); accept(TOKENIZER_LABEL); expr();} // C style:	I?x:y == wenn I<>0, I=x sonst I=y
	  else { expr(); accept(TOKENIZER_LABEL); val = expr();}
      break;
    case TOKENIZER_PLUS: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v+expr(); break;}
         accept(TOKENIZER_PLUS);
		 val=v+1; break;}
    case TOKENIZER_MINUS: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v-expr(); break;}
         accept(TOKENIZER_MINUS);
		 val=v-1; break;}
    case TOKENIZER_ASTR: { tokenizer_inc();
	 if (is_token(TOKENIZER_EQ)) { tokenizer_next();
		 val=v*expr(); break;}
         accept(TOKENIZER_ASTR);
		 val=v*v; break;}
    case TOKENIZER_SLASH: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 x=expr();
		 if (x==0) {ended=1; return;}
		 val=v/x; break;}
    case TOKENIZER_MOD: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 x=expr();
		 if (x==0) {ended=1; return;}
		 val=v%x; break;}
    case TOKENIZER_AND: {
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v&expr(); break;}
    case TOKENIZER_OR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v|expr(); break;}
    case TOKENIZER_XOR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v^expr(); break;}
    case TOKENIZER_SHL: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v<<(expr() & 31); break;}
    case TOKENIZER_SHR: { 
		 tokenizer_inc(); accept(TOKENIZER_EQ);
		 val=v>>(expr() & 31); break;}
	default: {
    if (stm_ctr<4 && !is_eof()) { reader_init();} // TOKENIZER_UBASIC_READER
    if (traconly>9) return; 
	
      ended=3;
	  DEBUG_PRINTF("let_statement: Wrong token: '%s'\n",tokenizer_token());
      accept(TOKENIZER_EQ); // "Expected 15, got 3"
      ubasic_error = UBASIC_E_UNK_STATEMENT;
    }
  }  // switch
  if (far) {field[var]=val;}
  else {variables[var]=val;
  DEBUG_PRINTF("let_statement: assign %d to %d\n", variables[var], var);}
}
  if (ended) return;
  t=tokenizer_token();
  sep= t==TOKENIZER_SEP;
  if (sep // print compatibel:
	||t == TOKENIZER_COMMA
	||t == TOKENIZER_SEMICOLON
	||t == TOKENIZER_STRING)	{tokenizer_next();}
  t=tokenizer_token();
  if (t == TOKENIZER_MOMENT
	||t == TOKENIZER_SEKUNDE)	{moment_statement(0);
  t=tokenizer_token();}
  } // endif str
  while (t==TOKENIZER_VARIABLE || (t>=TOKENIZER_STRINGVARIABLE && t<=TOKENIZER_CHR$));
  
  sep=sep & (t != TOKENIZER_CR
		  && t != TOKENIZER_ELSE
		  && t != TOKENIZER_ELSEIF);
  if (sep) {tokenizer_sep(); return;}
  if (t == TOKENIZER_CR) return;
  accept_cr();
}
/*---------------------------------------------------------------------------*/
// 10+1 extra vars : #0 - #9 + ##
static void data_statement(void) // to use with goto,gosub,do,for,while,return,break,cont too
{ 
  short var=MAX_VARNUM-13;
  short x=0, far=0;
  int val=0;
  while (is_token(TOKENIZER_SEP)) tokenizer_next();
  if (eos()) {tokenizer_eol(); return; }

  short t=tokenizer_token();
  switch (t) {
    case TOKENIZER_PRINT:
    case TOKENIZER_ECHO:
    case TOKENIZER_WRITE:
		 print_statement(t);
	break; }
	if (is_token(TOKENIZER_DATA)) accept(0);
	if (is_token(TOKENIZER_LET)) {let_statement(); return;}
	
	if (is_token(TOKENIZER_LABEL)) {// data:@(x)=1,2,3,a*b,,....
		accept(0); far = 1;
		accept(TOKENIZER_FIELD);  var = relation()%MAX_FIELD-1;
        accept(TOKENIZER_RIGHTPAREN);} 
	else if (is_token(TOKENIZER_EQ)){
		accept(0); var=var+expr()%10;} 
	else if (is_token(TOKENIZER_PLUS)){
		accept(0); x=-expr()%10; var = var+x;}
		
	do { var++; x++;
		if (is_comma()) {accept(0);
		   while (is_cr()) accept(0);}
		t=tokenizer_token();
		if (t == TOKENIZER_EQ
		  ||t == TOKENIZER_SEMICOLON	// skip 1st
		  ||t == TOKENIZER_STRING) {accept(0);}
		if (eos()) {tokenizer_eol(); return;}
		if (is_comma()) val=0; else val=expr();
		if (far) {field[var]=val;
		   if (var>=MAX_FIELD) tokenizer_eol();}
		else {
		   if (x>0) {variables[var]=val;
		      if (var>=MAX_VARNUM-3) tokenizer_eol();}
		}
	}while (is_comma());
  variables[MAX_VARNUM-2]=val; // ##= last set var or field
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void shift_statement(void)
{
  short var=MAX_VARNUM-13;
  short i=0, s=1, z=10;
  int val;
  accept(TOKENIZER_SHIFT);
  if (!eos()) s=abs(expr())%10;
  if (is_comma()) {accept(0); z=abs(expr()%10);}
  if (s*z>0)
  for (i=0; i<=z; i++) {
      val=var+i+s; 
      val= (val>=MAX_VARNUM-2) ? 0 : variables[val];
      variables[var+i]=val; 
  } // for 
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void swap_statement(void)
{
  int v1,v2;
  short n1,n2,token;
  accept(TOKENIZER_SWAP);
do {
  accept(-TOKENIZER_SEMICOLON);
  token=tokenizer_token();
  int str=token==TOKENIZER_STRINGVARIABLE;
  switch(token) {
    case TOKENIZER_FIELD: {
      accept(token);                  n1 = relation()%MAX_FIELD;
      accept(TOKENIZER_RIGHTPAREN);   v1 = field[n1];
      accept(TOKENIZER_COMMA);
      accept(token);                  n2 = relation()%MAX_FIELD;
      accept(TOKENIZER_RIGHTPAREN);   v2 = field[n2];
      field[n1]=v2; field[n2]=v1;
    break; }
    case TOKENIZER_VARIABLE:
    case TOKENIZER_STRINGVARIABLE: {
           n1=tokenizer_variable_num();
      accept(-TOKENIZER_STRINGVARIABLE);
      if ( n1==-TOKENIZER_LEFTPAREN) { n1 = varfactor(0x1000);
           v1 = variables[n1];} else if (!str) v1 = varfactor(n1);
      accept(TOKENIZER_COMMA);
           n2=tokenizer_variable_num();
      if ( n2==-TOKENIZER_LEFTPAREN) { n2 = varfactor(0x1000);
           v2 = variables[n2];} else if (!str) v2 = varfactor(n2);
		   
      if (str) {
		  accept(TOKENIZER_STRINGVARIABLE);
		  void *v; v=stringvariables[n1];
		  ubasic_set_stringvariable(n1, stringvariables[n2]);
		  ubasic_set_stringvariable(n2, v);
	  } else {variables[n1]=v2; variables[n2]=v1;}
    break;}
  }  // switch
} while (is_token(TOKENIZER_SEMICOLON));
accept_cr();
}
/*---------------------------------------------------------------------------* /
static void
rem_statement(void)
{
 tokenizer_eol();
 // accept(TOKENIZER_REM);
 // DEBUG_PRINTF("rem_statement\n");
 // accept(TOKENIZER_CR);
}
/ *---------------------------------------------------------------------------*/

static void shoot_statement(void)
{
	unsigned cs;
	short i,w=0;
  accept(TOKENIZER_SHOOT);
  ubasic_as_ret_var_num = -1;
  cs = GetFreeCardSpaceKb();
  if (is_token(TOKENIZER_VARIABLE)) // shoot else failed here
  {
    ubasic_as_ret_var_num = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE); // how to use?
  }
  while (shooting_in_progress()) {moment_statement(42);}  //4..
  action_push_shoot(1);
  //:...i....1....;....2....;....3....;....4..  Caefix
  if ( is_token(TOKENIZER_SEP)) { // rawproof??
	unsigned delay=get_tick_count()+9;
	for (i=0; (i<=conf.save_raw); ++i) {
	  do { 
		delay+=50; w+=50;
		moment_statement(delay);
		if (w>7000) { play_sound(6); break; }
		} while (cs >= GetFreeCardSpaceKb());
	  cs = GetFreeCardSpaceKb();
	  play_sound(4);
	};
  }
  if (cs<42000) {ended=1; loop=off;} // keep some sparespace
  flag_yield=1;
  DEBUG_PRINTF("End of shoot\n");
  accept_cr();
}

/*---------------------------------------------------------------------------*/
short set_mf_statement()
{
   short x;
   accept(-TOKENIZER_EQ);
   x=(expr() > 0) ? DoMFLock() : UnlockMF();
   accept_cr();
   return x;
}


static void set_focus_statement()
{
    accept(TOKENIZER_SET_FOCUS);
    int sd = expr();// %65536; "some" cams know higher values.
	if (sd==-1) {sd=INFINITY_DIST;}
	if (sd<0) {sd=65536-sd;}
	if (sd==0) {return;} //4....,....
    // if sd override not available now, fail immediately without calling set_focus
    // to avoid unexpected results with SET_LATER
    if(shooting_can_focus()) {
        //// NOTE duplicated in modules/luascript.c and lib/ubasic/ubasic.c
        // in AF lock or MF (canon or set by MF functions), set focus now
        if ( shooting_get_prop(camera_info.props.af_lock) 
          || shooting_get_focus_mode()
          || camera_info.state.mode_video)  // TODO video needs to be investigated, carried over from old code
        {
          shooting_set_focus(sd, SET_NOW);
        }
        else
        {
          // in an AF mode, set later
          shooting_set_focus(sd, SET_LATER);
        }
    }
    accept_cr();
}
/*---------------------------------------------------------------------------*/
static void md_detect_motion_statement()
{
    short columns;
    short rows;
    short pixel_measure_mode;
    int detection_timeout;
    int measure_interval;
    short threshold;
    short draw_grid;
	// var
    short clipping_region_mode=0;
    short clipping_region_row1=0;
    short clipping_region_column1=0;
    short clipping_region_row2=0;
    short clipping_region_column2=0;
    short parameters=0;
    short pixels_step=1;
    short msecs_before_trigger=16;

    accept(TOKENIZER_MD_DETECT_MOTION);

  //  int par=is_paren();
    columns=expr();
	tokenizer_next();

    rows=expr(); tokenizer_next();

    pixel_measure_mode=expr(); tokenizer_next();

    detection_timeout=expr(); tokenizer_next();
    measure_interval=expr(); tokenizer_next();

    threshold=expr(); tokenizer_next();

    draw_grid=expr(); tokenizer_next();

    ubasic_as_ret_var_num = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
/*
    if (tokenizer_token() != TOKENIZER_CR && tokenizer_token() != TOKENIZER_ELSE) {
        // eat COMA	
        //			tokenizer_next();
    }
*/
	 if (is_comma()) { tokenizer_next();
        clipping_region_mode = expr()%3;
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_column1 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_row1 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_column2 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        clipping_region_row2 = expr();
    }
	 if (is_comma()) { tokenizer_next();
        parameters = expr();
    }
	 if (is_comma()) { tokenizer_next();
        pixels_step = expr();
    }
	 if (is_comma()) { tokenizer_next();
        msecs_before_trigger = expr();
    }

  //  if (par) accept(-TOKENIZER_RIGHTPAREN);
    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    libmotiondetect->md_init_motion_detector(
        columns, rows, pixel_measure_mode, detection_timeout, 
        measure_interval, threshold, draw_grid,
        clipping_region_mode,
        clipping_region_column1, clipping_region_row1,
        clipping_region_column2, clipping_region_row2,
        parameters, pixels_step, msecs_before_trigger);
#endif

    flag_yield=1;
}

/*---------------------------------------------------------------------------*/
static void click_statement(void)
{
  short k;
  accept(TOKENIZER_CLICK);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_click(k);
    flag_yield=1;
  }

  DEBUG_PRINTF("End of click\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void press_statement(void)
{
  short k;
  accept(TOKENIZER_PRESS);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_press(k);
    flag_yield=1;
  }
  DEBUG_PRINTF("End of press\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void release_statement(void)
{
  short k;
  accept(TOKENIZER_RELEASE);
  k = ubasic_get_key_arg();
  if (k > 0) {
    action_push_release(k);
    flag_yield=1;
  }
  DEBUG_PRINTF("End of release\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
// do
// Wai. 1000
//  K=0 // !! How about that ?  Caefix
//  is_key K "up"   200 ~|goto|data|let..
//  is_key K "set"  500 ~|data|let..
//  is_key K "down" 800 ~|data|let..
// ...   
// until K>0
// goto K*10
//
static void is_key_statement(void)
{ // ++ is_pressed_statement
    short var,val;
	short k=ifkey>-1 && is_token(TOKENIZER_IS_KEY);
	accept(0); flag_yield=1;
    var = tokenizer_variable_num();
  if (var==-TOKENIZER_LEFTPAREN) {var = varfactor(0x1000);
       val = variables[var];}
  else val = varfactor(var);// !!
	if (val) {accept_cr(); tokenizer_eol();	return;} // allready set, so ignore
  //  accept(TOKENIZER_VARIABLE);
  
	if(k) val=script_key_is_clicked(ubasic_get_key_arg()); // 
	else  val=script_key_is_pressed(ubasic_get_key_arg()); // ++
	  
	if (!val) {accept_cr(); tokenizer_eol();} else {
	if (is_token(TOKENIZER_NUMBER)) {val=expr();}
    } 
    ubasic_set_variable(var,val);
    DEBUG_PRINTF("End of is_key\n");
}

/*---------------------------------------------------------------------------*/

static void set_exit_key_statement()
{
    int k;
    accept(TOKENIZER_SET_EXIT_KEY);
    if(is_token(TOKENIZER_STRING)) 
    {
        tokenizer_string(string, sizeof(string));
        k = script_keyid_by_name(string);
        if (k <= 0)
        {
          ubasic_error = UBASIC_E_UNK_KEY;
          ended = 1 ;
        }
        else
        {
            script_set_terminate_key(k,string);
        }
    } 
    else 
    {
       DEBUG_PRINTF("ubasic.c: set_exit_key no key specified\n");
       ended = 1;
       ubasic_error = UBASIC_E_UNK_KEY;
    }
    accept_cr() ;
}

/*---------------------------------------------------------------------------*/
// Wait for a button to be pressed and released (or the timeout to expire)
static int action_stack_AS_UBASIC_WAIT_CLICK()
{
    // Check key pressed or timeout
    if ((get_tick_count() >= action_top(2)) || camera_info.state.kbd_last_clicked)
    {
        // If timed out set key state to "no_key", otherwise key pressed so set last checked time
        if (!camera_info.state.kbd_last_clicked)
            camera_info.state.kbd_last_clicked=0xFFFF;
        else
            camera_info.state.kbd_last_checked_time = camera_info.state.kbd_last_clicked_time;

        action_pop_func(1);
        return 1;
    }
    return 0;
}

static unsigned  wait_click_statement()
{
unsigned ti=get_tick_count();
    int delay = -1;
	if (!eos()) {
	if (is_token(TOKENIZER_SEKUNDE)) {accept(0); delay=1000;}
    else { delay = expr();
	  if ( delay == 0) delay = -1;}
    }
    delay = sleep_delay(delay);

    accept_cr();

    // Reset 'clicked' key if it has not changed since last time
    if (camera_info.state.kbd_last_clicked_time <= camera_info.state.kbd_last_checked_time)
    {
        camera_info.state.kbd_last_clicked = 0;
    }

    // Set up for wait or click testing
    action_push(delay);
    action_push_func(action_stack_AS_UBASIC_WAIT_CLICK);

    // Check for short delay or key already pressed by calling action stack routine once now
    if (action_stack_AS_UBASIC_WAIT_CLICK() == 0)
    {
        flag_yield=1;
    }
//  ti=get_tick_count()-ti; // allways 0 !?
  ti=camera_info.state.kbd_last_clicked_time-ti; // nice val ?
  return ti;
}

/*
:...;....1....;....2....;....3....;....4....:....5....;....6....;....7....;....*/
/*---------------------------------------------------------------------------*/
/* adapted from luascript.c :*/

static void return_string_selected(const char *str) {
    // Reconnect button input to script - will also signal action stack
    // that file browser / textbox is finished and return last selected file
    // to script caller
    camera_info.state.state_kbd_script_run = SCRIPT_STATE_RAN;

    // Push selected file as script return value
    strcpy(string_selected, (str && str[0]) ? str : nullstring );
}

static int action_stack_AS_WAIT_MODULE()
{
    // state_kbd_script_run is set to 0 when the file browser is started from a Lua script
    // it is reset back to 1 when the file browser exits and control is returned back to
    // the script
    if (camera_info.state.state_kbd_script_run)
    {
        action_pop_func(0);
    }
    return 0;
}

static char* file_browser()   // ~statement
{ 
  char *dir;
  accept(TOKENIZER_FILE_BROWSER);
  dir=sexpr(0);
  if (ended) strcpy(string_selected,"Error:file_browser");
  if (*dir=='\0' || dir[1] != '/') dir[1]='\0';
  *dir='A';
     // Disconnect button input from script so buttons will work in file browser
  camera_info.state.state_kbd_script_run = SCRIPT_STATE_INACTIVE;
     // Push file browser action onto stack - will loop doing nothing until file browser exits
  action_push_func(action_stack_AS_WAIT_MODULE);
     // Switch to file browser gui mode. Path can be supplied in call or defaults to "A" (root directory).
  libfselect->file_select(LANG_STR_FILE_BROWSER,
     dir, "A", return_string_selected);
    // Yield the script so that the action stack will process the AS_FILE_BROWSER action

  flag_yield=1;
  return string_selected;
}
/*
:...;....1....;....2....;....3....;....4....:....5....;....6....;....7....;....*/
static char* textbox()    // ~statement
{
    int  size=MAX_STRINGLEN-1;
    char *box='\0', *txt='\0', *def='\0';
 //   accept(TOKENIZER_TEXTBOX);
    
    if (!is_comma()) {box=sfactor(0,TOKENIZER_TEXTBOX);
        if (ended) strcpy(string_selected,"Error.");
        if ( is_comma()) { accept(TOKENIZER_COMMA);
           if (!is_token(TOKENIZER_RIGHTPAREN)) txt=store(sfactor(0,0));
           if (ended) strcpy(string_selected,"Error..");
           if (is_comma()) accept(TOKENIZER_COMMA);
           if (!is_token(TOKENIZER_RIGHTPAREN))
               def=strcpy(string_selected,sexpr(0));
           if (ended) strcpy(string_selected,"Error...");
           if (!is_token(TOKENIZER_RIGHTPAREN))
               {accept(TOKENIZER_COMMA); size=expr();}
        }
	}
 //   if (ended) {return string_selected;}
    accept(-TOKENIZER_RIGHTPAREN);

    if (*box=='\0') box=store("Text Box");
    if (*txt=='\0') txt=store("Enter Text:");
    if (*def=='\0') def=strcpy(string_selected,nullstring);

    // Disconnect button input from script so buttons will work in the textbox
    camera_info.state.state_kbd_script_run = SCRIPT_STATE_INACTIVE;
    // Switch to textbox gui mode. Text box prompt should be passed as param.
	
    int rv = libtextbox->textbox_init(
	    (int)box,   //title
        (int)txt, //message
        def,                  //default string
        size,                 //max size of a text
        return_string_selected, 0);
    if (rv)
    {
        // Push textbox action onto stack - will loop doing nothing until textbox exits
        action_push_func(action_stack_AS_WAIT_MODULE);
    }
    else
        return_string_selected("\0");

  // Yield the script so that the action stack will process the AS_TEXTBOX action
  flag_yield=1;
  return string_selected;
}

/*---------------------------------------------------------------------------*/
static void cls_statement(void)
{
  accept(TOKENIZER_CLS);
  console_clear();
  DEBUG_PRINTF("cls_statement\n");
  accept_cr();
}
/*---------------------------------------------------------------------------*/
static void set_console_layout(void)
{
  short x1=0;
  short y1=0;
  short x2=25;
  short y2=5;
  accept(TOKENIZER_SET_CONSOLE_LAYOUT); // 
  x1 = expr();
  if (x1<0) {x1=0; y1=0; x2=45; y2=12;}
  if (is_comma()) {accept(TOKENIZER_COMMA); y1 = expr();}
  if (is_comma()) {accept(TOKENIZER_COMMA); x2 = expr();}
  if (is_comma()) {accept(TOKENIZER_COMMA); y2 = expr();}
  console_set_layout(x1,y1,x2,y2);
  accept_cr();  
}
/*---------------------------------------------------------------------------*/
static void set_console_autoredraw(void)
{
  accept(TOKENIZER_SET_CONSOLE_AUTOREDRAW);
  console_set_autoredraw(expr());
  accept_cr();  
}
/*---------------------------------------------------------------------------*/
static void console_redraw_statement(void)
{
  accept(TOKENIZER_CONSOLE_REDRAW);
  console_redraw(0);
    accept_cr();
}

static void set_yield_statement()
{
    accept(TOKENIZER_SET_YIELD);
    short val = expr();
	accept(-TOKENIZER_COMMA);
    yield_max_lines = val?val:YIELD_MAX_LINES_DEFAULT;
    val = expr();
    yield_max_ms = val?val:YIELD_MAX_MS_DEFAULT;
    accept_cr();
}

/*---------------------------------------------------------------------------*/

#ifdef INCLUDE_OLD_GET__SYNTAX

static void get_prop_statement()
{
    short var, var1;
    accept(TOKENIZER_GET_PROP);
    var = expr();
	accept(-TOKENIZER_COMMA);
    var1 = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var1, shooting_get_prop(var));
    accept_cr();
}

// Call 'func' to get a CHDK short value and store in named parameter
static void get_short_var_statement(short token, short (*func)(void))
{
    accept(token);
    short var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var, func());
    accept_cr();
}

// Call 'func' to get a CHDK int value and store in named parameter
static void get_int_var_statement(short token, int (*func)(void))
{
    accept(token);
    short var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    ubasic_set_variable(var, func());
    accept_cr();
}

#endif

/*---------------------------------------------------------------------------*/
static void on_off_statement(short token, void (*on)(void), void (*off)(void))
{
  accept(token);
  if (expr() > 0) on(); else off();
  accept_cr();
}
static void switch_function(short token, void (*func)( void))
{
    accept(token); //...4..exit_alt & enter_alt
    func();
    accept_cr();
}
/*
static void on_off_switch(int val, void (*on)(void), void (*off)(void))
{
  if (val > 0) on(); else off();
  //	on_off_switch(0, DoAFLock, UnlockAF);
  //	on_off_switch(1, DoAFLock, UnlockAF);
}
* /
// Call CHDK function 'func' with one int parameter
static int one_int_param_function(int token, void (*func)(int))
{   int r=0;
    accept(token);
    func(r);
    accept_cr();
	return r;
}
*/
static int one_int_param_function(short token, void (*func)(int))
{   int l,r=0;
    accept(token);
	switch(token) { //
		case TOKENIZER_SET_ZOOM: r=expr();
			if (is_comma()) {
				accept(TOKENIZER_COMMA); l=expr();
				shooting_set_zoom_speed(l); }
		break;
		default: r=expr();
	} //
    func(r);
    accept_cr();
	return r;
}

// Call CHDK function 'func' with one short parameter
static void one_short_param_function(short token, void (*func)(short))
{
    accept(token);
    func(expr());
    accept_cr();
}

/*
// Call CHDK function 'func' with one int parameter plus a second value 'p2'
static void one_int_param_plus_const_function(int token, void (*func)(int,short), short p2)
{
    accept(token);
    func(expr(),p2);
    accept_cr();
}

// Call CHDK function 'func' with one short parameter plus a second value 'p2'
static void one_short_param_plus_const_function(int token, void (*func)(short,short), short p2)
{
    accept(token);
    func(expr(),p2);
    accept_cr();
}

/ *---------------------------------------------------------------------------*/

// Set the property 'prop' value from the next expression
static void set_propcase_statement(short token, short prop)
{
    accept(token);
	shooting_set_prop(prop, expr());
    accept_cr();
}

static void set_prop_statement()
{
    int to, to1;
    accept(TOKENIZER_SET_PROP);
    to = expr();
	accept(-TOKENIZER_COMMA);
    to1 = expr();
	shooting_set_prop(to, to1);
    accept_cr();
}

static void set_ev_statement()
{
     int to;
     accept(TOKENIZER_SET_EV);
     to = expr();
         shooting_set_prop(camera_info.props.ev_correction_1, to);
         shooting_set_prop(camera_info.props.ev_correction_2, to);
     accept_cr();
}

/*---------------------------------------------------------------------------*/


static void set_movie_status_statement()
{
    accept(TOKENIZER_SET_MOVIE_STATUS);
    set_movie_status(expr());
    accept_cr();
}


static void set_led_statement()
{
    short to, to1, to2;
    accept(TOKENIZER_SET_LED);
    to = expr();
	accept(-TOKENIZER_COMMA);
    to1 = expr();
	to2 = 200;
	
	if (is_comma()) {accept(TOKENIZER_COMMA); to2 = expr(); }
	camera_set_led(to, to1, to2);
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void set_capture_mode_canon_statement()
{
    int to;
    accept(TOKENIZER_SET_CAPTURE_MODE_CANON);
    to = expr();
    // if the value as negative, assume it is a mistakenly sign extended PROPCASE_SHOOTING_MODE value
    if( to < 0) 
        to &= 0xFFFF;
    shooting_set_mode_canon(to);
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void set_config_value_statement()
{
    int id, value;
    tConfigVal configVal = {0};
    
    accept(TOKENIZER_SET_CONFIG_VALUE);
    id = expr();
	accept(-TOKENIZER_COMMA);
    value = expr();
    if( conf_getValue(id, &configVal) == CONF_VALUE ) {
        configVal.numb = value;
        configVal.isNumb = 1;
        conf_setValue(id, configVal);
    }
    accept_cr();
}

static void get_config_value_statement()
{
    int var, var1, var2;
    tConfigVal configVal;
    accept(TOKENIZER_GET_CONFIG_VALUE);
    var = expr();
	accept(-TOKENIZER_COMMA);
    var1 = expr();
	accept(-TOKENIZER_COMMA);
    var2 = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    if( conf_getValue(var, &configVal) == CONF_VALUE ) {
        ubasic_set_variable(var2, configVal.numb);
    } else {
        ubasic_set_variable(var2, var1);
    }
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void uB_set_av96(int param){
  shooting_set_av96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_av96_direct(int param){
  shooting_set_av96_direct( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_tv96(int param){
  shooting_set_tv96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_tv96_direct(int param){
  shooting_set_tv96_direct( param, shooting_in_progress()?SET_NOW:SET_LATER);
  return ;
}
static void uB_set_sv96(int param){
  shooting_set_sv96( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_nd_filter_state(int param) {
   shooting_set_nd_filter_state( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
static void uB_set_iso_real(int param) {
  shooting_set_iso_real( param, shooting_in_progress()?SET_NOW:SET_LATER );
  return ;
}
/*---------------------------------------------------------------------------*/

static void md_get_cell_diff_statement()
{
    short var, col, row;
    accept(TOKENIZER_MD_GET_CELL_DIFF);

    col=expr(); tokenizer_next();
    row=expr(); tokenizer_next();

    var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    ubasic_set_variable(var, libmotiondetect->md_get_cell_diff(col,row));
#endif
}

static void md_get_cell_val_statement()
{
    short var, col, row;
    accept(TOKENIZER_MD_GET_CELL_VAL);

    col=expr(); tokenizer_next();
    row=expr(); tokenizer_next();

    var = tokenizer_variable_num();
    accept(TOKENIZER_VARIABLE);
    accept_cr();

#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    ubasic_set_variable(var, libmotiondetect->md_get_cell_val(col,row));
#endif
}

/*---------------------------------------------------------------------------*/
static int _shot_histogram_set(short enable)
{
#ifdef UBASIC_TEST
    return 0;
#else
    return libshothisto->shot_histogram_set(enable);
#endif
}

/*---------------------------------------------------------------------------*/
static void shutdown_statement(void){
  accept(TOKENIZER_SHUT_DOWN);
  camera_shutdown_in_a_second();
  accept_cr();
}

static void reboot_statement() {
    accept(TOKENIZER_REBOOT);
    if(is_token(TOKENIZER_STRING)) {
        tokenizer_string(string, sizeof(string));
        tokenizer_next();
	    reboot(string);
    } else {
	    reboot(NULL);
    }
}

static void set_autostart_statement()
{
    short to;
    accept(TOKENIZER_SET_SCRIPT_AUTOSTART);
    to = expr();
#ifndef UBASIC_TEST
	if (to >= 0 && to <= 3) conf.script_startup=to;
	conf_save();
#endif
    accept_cr();
}

/*---------------------------------------------------------------------------*/
static void poke_statement(void)
{
    accept(TOKENIZER_POKE);
	unsigned a=(unsigned)expr(); accept(TOKENIZER_COMMA);
	unsigned v=(unsigned)expr(); short s=4;
	if (is_comma()) {
		accept(TOKENIZER_COMMA); s=expr()%5; }
	if (a & (s-1)) {accept(TOKENIZER_POKE); return;}
	if (s==4) *(unsigned int *)(a)  = (unsigned int)(v); else
	if (s==2) *(unsigned short *)(a)= (unsigned short)(v); else
	if (s==1) *(unsigned char *)(a) = (unsigned char)(v);
    accept_cr();
}
/*---------------------------------------------------------------------------*/
static void mem_dump_statement(void)
{
    unsigned c, s=0xFF;
    FILE *fh = fopen("A/mem_dump.Log","wb");
    if(!fh) return;
	c=(unsigned)(expr());
    if (is_comma()) {
	    accept(TOKENIZER_COMMA);
	    s=(unsigned)expr();}
//	for (; i<s; i++)
	//	fwrite((const char *)(c+i),1,1,fh);
	fwrite((const char *)c,s,1,fh);
//	fwrite((void*)c,s,1,fh);
	fclose(fh);
	if (ended) return;
	accept_cr();
}
/*---------------------------------------------------------------------------*/
static void clean_statement(void) // #V..#7 // 42[..62]
{ int i;
  accept(TOKENIZER_CLEAN);
  do {
    accept(-TOKENIZER_COMMA);
    accept(-TOKENIZER_SEMICOLON);
    int str=is_token(TOKENIZER_STRINGVARIABLE);
    short a=0;
    short z=MAX_VARNUM-2; // keep #@ = loop = #(63)
    if (tokenizer_char() == TOKENIZER_VARIABLE
     || tokenizer_char() == TOKENIZER_STRINGVARIABLE) {
		a=tokenizer_variable_num();
		z=a; accept(0);
        if (is_token(TOKENIZER_TO)) {accept(TOKENIZER_TO);
		   z=tokenizer_variable_num();
		   accept(0);}
    } else
    if (!eos()) { a=expr()&63;   z=a;
       if (is_token(TOKENIZER_TO))
          {accept(TOKENIZER_TO); z=expr()&63; }
    }
    if (str) {for (i=a; i<=z; ++i) *(stringvariables[i])='\0';}
    else     {for (i=a; i<=z; ++i) variables[i]=0;}
} while (is_comma() || is_token(TOKENIZER_SEMICOLON));
garbage_collect();
accept_cr();
}
/*---------------------------------------------------------------------------*/
static void end_statement(void)
{
  accept(TOKENIZER_END);
  if (!ended & !eos() & !is_token(TOKENIZER_ENDOFINPUT))	end_wait=expr()%10;
  if (readfile!=NULL) fclose(readfile);
  readfile=NULL;
  ended = 1;
  loop=off;
}
/*---------------------------------------------------------------------------*/
static void loop_statement()
{
  accept(TOKENIZER_LOOP);
  accept(-TOKENIZER_EQ);
  loop=expr();
  variables[MAX_VARNUM-1] = loop; // = #@ = #(63)
  accept_cr();
}

/*---------------------------------------------------------------------------* /
static void _statement(void)
{
	// accept(TOKENIZER_);
	//:...i....1....;....2....;....3....;....4..
}
/ *---------------------------------------------------------------------------*/
static void test_statement(void)
{
    extern long physw_status; // (use compiler warning as a hint. :) 
 /*
    strcpy(buf, "PS1: %08x",&physw_status[0]);
	int r=(unsigned int) *(unsigned int *)(0x0001a278); 
	sprintf(buf, "PS1: %08x",r);
    script_console_add_error((long)buf);
  //:...i....1....;....2....;....3....;....4..
 */
}
/*---------------------------------------------------------------------------*/
static void list_statement(void)
{
    short l=is_token(TOKENIZER_LIST);
	short s=is_token(TOKENIZER_STRING);
    short d=traconly>9; // (UBASIC_READER);
	if (l) {tokenizer_inc();}
	if (is_token(TOKENIZER_UBASIC_READER)) {stopwatch(-ubasic_linenumber());}
	if (is_cr()) {return;}
	if (l) {tokenizer_sep();}
	tokenizer_string(string,123);
	if (s||traconly) sprintf(buf, "%s", string); else
	sprintf(buf, "@%d $%d: %d '%s'", stm_ctr, -ubasic_linenumber(), tokenizer_token(), string);
	script_console_add_error((long)buf); buf[0]='\0';
    if (d && pause==0) msleep(traconly*2);
	if (l) {tokenizer_next();}
	if (is_token(TOKENIZER_STRING) || traconly) tokenizer_eol();
 while (is_token(TOKENIZER_SEP)) tokenizer_next();
 //  //:...i....1....;....2....;....3....;....4..
}

/*---------------------------------------------------------------------------*/ 
void ubasic_reader()
{unsigned short i=1;
while (pause==1 || i==1) { i=0;
  if (!camera_info.state.gui_mode) { pause=-1; return;}  // -->> ubasic_run
  switch (kbd_get_autoclicked_key() | get_jogdial_direction()) {
    case KEY_ZOOM_OUT:
    case JOGDIAL_LEFT:  msleep(100);
         tokenizer_reader_pos(-2); 
         stopwatch(ubasic_linenumber());
         console_redraw(0); pause=-1;
        return;
    case KEY_ZOOM_IN:
    case JOGDIAL_RIGHT: msleep(100);
		 if (traconly<100) traconly=100;
         console_redraw(0); pause=0;  break;
    case KEY_LEFT:      pause=0; traconly+=50; break;
    case KEY_RIGHT:     pause=2; traconly-=20; 
         tokenizer_reader_pos(-1);   break;
    case KEY_UP:
         pause=12-tokenizer_reader_pos(-24);   break;
    case KEY_DOWN:
         pause=12-tokenizer_reader_pos(24);    break;
    case KEY_SET:
    case KEY_SHOOT_HALF: pause=(pause==-1) ? 12-tokenizer_reader_pos(12) :1;
         if (is_eof()) { pause=8; tokenizer_reader_pos(-8);}  break;
    case KEY_WIFI:
    case KEY_HELP:
    case KEY_FACE:
    case KEY_DISPLAY:    pause=8;
         tokenizer_reader_pos(65432);
         stopwatch(variables[MAX_VARNUM-2]);
         tokenizer_reader_pos(-8);  break;
    case KEY_VIDEO:
    case KEY_PRINT:
    case KEY_RING_FUNC:
         variables[MAX_VARNUM-2]++; // ##++
         stopwatch(variables[MAX_VARNUM-2]);
         play_sound(4); break;
    case KEY_MENU: { if (go_fwd>-2) // continue script ...
		{traconly=TOKENIZER_UBASIC_READER;} else 
        {console_clear(); stopwatch(ubasic_linenumber());}
         pause=-1;
		 return;}     // -->> ubasic_run
    default: {
        if (pause<5) msleep(abs(traconly*2-10*pause));
        }
    } // switch
    if (traconly<10) traconly=10;
    if (traconly>800) traconly=1000;
	if (pause==1 || is_eof()) { pause=-1;
        if (!is_eof()) {console_redraw(0); }
         tokenizer_reader_pos(-1);
       // stopwatch(ubasic_linenumber());
    }
    msleep(50);
  } // while
}
/*---------------------------------------------------------------------------*/

static void statement(void)
{
  int l;
  ubasic_token token;

	if (is_token(TOKENIZER_MINUS)) tokenizer_next();
 while (is_token(TOKENIZER_SEP)) { tokenizer_next();
	if (is_cr()) tokenizer_next();}
 while (is_token(TOKENIZER_MOMENT) || // . =sleep(50)
		is_token(TOKENIZER_SEKUNDE) ||
		is_token(TOKENIZER_TO)) {moment_statement(0);}
 while (is_cr() || is_token(TOKENIZER_SEP))  {tokenizer_next();}
 
 // Errorology / Reader begin

 if (debug && traconly<9) list_statement(); else
 if (trace) {tokenizer_statement(string,123);
    sprintf(buf, "@%d: '%s'",ubasic_linenumber(), string);
    if (string[0]) script_console_add_error((long)buf);
	}

 if (traconly>9) { if (!pause) list_statement();
	 ubasic_reader(); ended=0;
    if (traconly==TOKENIZER_UBASIC_READER) return; // -->> ubasic_run
    if (pause || is_cr()) return;}
// if (ended) return;
 
 
// Strange bug	///here   
/// if ( pause || ended) {return;}
// -->> In file .o/ubasic.elf:
// elf2flt unknown segment 1 for name: 'accept'
// gmake[1]: *** [.o/ubasic.flt] Error 6
// gmake: *** [all-recursive] Error 1

/// if (pause>2) { pause--; return;} // sometimes
// -->> In file .o/ubasic.elf:
// elf2flt unknown symbol: '__gnu_thumb1_case_shi'
// gmake[1]: *** [.o/ubasic.flt] Error 5
// elf2flt unknown segment 1 for name: 'tokenizer_next'
// gmake[1]: *** [.o/ubasic.flt] Error 6

// Errorology end   

  if (is_token(TOKENIZER_NUMBER)) tokenizer_next();
  if (is_token(TOKENIZER_ON)) {on_statement();} // on i=1,2,3 goto x,y,"Z"
    /*  Fall through. */

  stm_ctr++; stm_ptr=tokenizer_curptr();
  token = tokenizer_token();
  switch(token) { 
  case TOKENIZER_VARIABLE:
  case TOKENIZER_FIELD:  	let_statement();
	if (ended==3) token=TOKENIZER_ERROR;
      break;
  case TOKENIZER_DATA:		data_statement();
      break;
  case TOKENIZER_IF:
  case TOKENIZER_ELSEIF:    if_statement(token);
      break;
  case TOKENIZER_ELSE:		else_statement();
      break;
  case TOKENIZER_ENDIF:		endif_statement();
      break;
  case TOKENIZER_SELECT:	select_statement();
      break;
  case TOKENIZER_CASE:		case_statement();
      break;
  case TOKENIZER_CASE_ELSE: case_statement();
     break;
  case TOKENIZER_END_SELECT:end_select_statement();
     break;
  // string addition
  case TOKENIZER_STRINGVARIABLE:
  case TOKENIZER_LEFT$:
  case TOKENIZER_MID$:
  case TOKENIZER_RIGHT$:
  case TOKENIZER_UP$:
  case TOKENIZER_LOW$:
  case TOKENIZER_REV$:
  case TOKENIZER_REP$:
  case TOKENIZER_READ$:
  case TOKENIZER_TRIM$:
  case TOKENIZER_STROL$:
  case TOKENIZER_STROR$:
  case TOKENIZER_STR$:
  case TOKENIZER_SPC$:
  case TOKENIZER_CHR$:
  // end of string addition	
  case TOKENIZER_LET:		let_statement();
	if (ended) token=TOKENIZER_ERROR;
     break;
  case TOKENIZER_CR:		return; // -> line_statement 
  case TOKENIZER_DEF_FN: 	tokenizer_def_fn();  return;
  case TOKENIZER_UBASIC_READER:
      if (traconly>9) {     stopwatch(-stm_ctr); return;}
  case TOKENIZER_EXIT:
  case TOKENIZER_GOSUB:
  case TOKENIZER_GOTO:		goto_statement(token);
      break;
  case TOKENIZER_RETURN:    return_statement();
      break;
  case TOKENIZER_BREAK:		break_statement();
      break;
  case TOKENIZER_CONT:		break_statement();
      break;
  case TOKENIZER_FOR:		for_statement();
      break;
  case TOKENIZER_NEXT:		next_statement();
      break;
  case TOKENIZER_DO:	 		do_statement();
      break;
  case TOKENIZER_UNTIL:		until_statement();
      break;
  case TOKENIZER_WHILE:		while_statement();
      break;
  case TOKENIZER_WEND:		wend_statement();
      break;
  case TOKENIZER_EXEC:		tokenizer_exec();
      if (eos()) accept_cr();
	  break;
  case TOKENIZER_SKIP:
	  accept(TOKENIZER_SKIP);
	  l=tokenizer_token(); accept(0);
	  tokenizer_skip(l);
	  break; // >> Crazy Jump fwd; be careful!
  case TOKENIZER_ECHO:	
  case TOKENIZER_WRITE:	
  case TOKENIZER_STRLEN:
  case TOKENIZER_PRINT:     print_statement(token);
      break;
  case TOKENIZER_SHIFT:		shift_statement();
      break;

  case TOKENIZER_SHOOT:		shoot_statement();
      break;
  case TOKENIZER_CLICK:		click_statement();
      break;
  case TOKENIZER_PRESS:		press_statement();
      break;
  case TOKENIZER_RELEASE:	release_statement();
      break;
  case TOKENIZER_IS_KEY: 	is_key_statement();
      break; // syntax synced :
  case TOKENIZER_IS_PRESSED:is_key_statement();
      break;
  case TOKENIZER_IFKEY:
      accept(TOKENIZER_IFKEY);
      ifkey=expr();
	  accept_cr();
	  break;
  case TOKENIZER_WAIT_CLICK:
    accept(token);           wait_click_statement();
      break;
#ifdef INCLUDE_OLD_GET__SYNTAX
  case TOKENIZER_GET_PROP:
      get_prop_statement();
      break;
  case TOKENIZER_GET_TV96:
      get_short_var_statement(token, shooting_get_tv96);
      break;
  case TOKENIZER_GET_USER_TV96:
      get_short_var_statement(token, shooting_get_user_tv96);
      break;
  case TOKENIZER_GET_USER_TV_ID:
      get_int_var_statement(token,   shooting_get_user_tv_id);
      break;
  case TOKENIZER_GET_AV96:
      get_short_var_statement(token, shooting_get_av96);
      break;
  case TOKENIZER_GET_USER_AV96:
      get_short_var_statement(token, shooting_get_user_av96);
      break;
  case TOKENIZER_GET_USER_AV_ID:
      get_int_var_statement(token,   shooting_get_user_av_id);
      break;
//#endif
  case TOKENIZER_GET_ZOOM:
      get_int_var_statement(token,   shooting_get_zoom);
      break;
  case TOKENIZER_GET_FOCUS:
      get_int_var_statement(token,   shooting_get_subject_distance);
      break;
  case TOKENIZER_GET_NEAR_LIMIT:
      get_int_var_statement(token,   shooting_get_near_limit_of_acceptable_sharpness);
      break;
  case TOKENIZER_GET_FAR_LIMIT:
      get_int_var_statement(token,   shooting_get_far_limit_of_acceptable_sharpness);
      break;
  case TOKENIZER_GET_DOF:
      get_int_var_statement(token,   shooting_get_depth_of_field);
      break;
  case TOKENIZER_GET_HYPERFOCAL_DIST:
      get_int_var_statement(token,   shooting_get_hyperfocal_distance);
      break;
  case TOKENIZER_GET_ISO_MARKET:
      get_short_var_statement(token, shooting_get_iso_market);
      break;
  case TOKENIZER_GET_ISO_REAL:
      get_short_var_statement(token, shooting_get_iso_real);
      break;
  case TOKENIZER_GET_BV96:
      get_short_var_statement(token, shooting_get_bv96);
      break;
  case TOKENIZER_GET_SV96:
      get_short_var_statement(token, shooting_get_sv96_real);
      break;
  case TOKENIZER_GET_ISO_MODE:
      get_int_var_statement(token,   shooting_get_iso_mode);
      break;
  case TOKENIZER_GET_VBATT:
      get_int_var_statement(token, (int (*)(void))stat_get_vbatt);
      break;
  case TOKENIZER_GET_DISK_SIZE:
      get_int_var_statement(token, (int (*)(void))GetTotalCardSpaceKb);
      break;
  case TOKENIZER_GET_FREE_DISK_SPACE:
      get_int_var_statement(token, (int (*)(void))GetFreeCardSpaceKb);
      break;
  case TOKENIZER_GET_JPG_COUNT:
      get_int_var_statement(token, (int (*)(void))GetJpgCount);
      break;
  case TOKENIZER_GET_RAW_COUNT:
      get_int_var_statement(token, (int (*)(void))GetRawCount);
      break;
#endif // old get syntax

  case TOKENIZER_SET_ZOOM:
	if (camera_info.state.gui_mode)
      l=one_int_param_function(token, shooting_set_zoom);
      break;
  case TOKENIZER_SET_ZOOM_REL:
      one_int_param_function(token,   shooting_set_zoom_rel);
      break;
  case TOKENIZER_SET_ZOOM_SPEED:
      one_int_param_function(token,   shooting_set_zoom_speed);
      break;
  case TOKENIZER_SET_MF:
      accept(TOKENIZER_SET_MF);    set_mf_statement();
      break;
  case TOKENIZER_SET_AFLOCK:
      on_off_statement(token, DoAFLock, UnlockAF);
      break;
  case TOKENIZER_SET_FOCUS: set_focus_statement();
      break;
  case TOKENIZER_SET_AELOCK:
      on_off_statement(token, DoAELock, UnlockAE);
      break;
  case TOKENIZER_SET_EV: set_ev_statement();
     break;
	   
  case TOKENIZER_SET_TV96_DIRECT:
      one_int_param_function(token,   uB_set_tv96_direct);
      break;
  case TOKENIZER_SET_TV96:
      one_int_param_function(token,   uB_set_tv96);
      break;
  case TOKENIZER_SET_USER_TV96:
      one_short_param_function(token, shooting_set_user_tv96);
      break;
  case TOKENIZER_SET_USER_TV_BY_ID:
      one_int_param_function(token,   shooting_set_user_tv_by_id);
      break;
  case TOKENIZER_SET_USER_TV_BY_ID_REL:
      one_int_param_function(token,   shooting_set_user_tv_by_id_rel);
      break;
  case TOKENIZER_SET_AV96_DIRECT:
      one_int_param_function(token,   uB_set_av96_direct );
      break;
  case TOKENIZER_SET_AV96:
      one_int_param_function(token,   uB_set_av96 );
      break;
  case TOKENIZER_SET_USER_AV96:
      one_short_param_function(token, shooting_set_user_av96);
      break;
  case TOKENIZER_SET_USER_AV_BY_ID:
      one_int_param_function(token,   shooting_set_user_av_by_id);
      break;
  case TOKENIZER_SET_USER_AV_BY_ID_REL:
      one_int_param_function(token,   shooting_set_user_av_by_id_rel);
      break;
  case TOKENIZER_SET_SV96:
      one_int_param_function(token,   uB_set_sv96 );
      break;
  case TOKENIZER_SET_ISO_REAL:
      one_int_param_function(token,   uB_set_iso_real );
      break;
  case TOKENIZER_SET_ISO_MODE:
      one_int_param_function(token,   shooting_set_iso_mode);
      break;
  case TOKENIZER_SHOT_HISTO_ENABLE:
      one_int_param_function(token, (void (*)(int))_shot_histogram_set);
      break;
  case TOKENIZER_SET_RECORD:
      one_int_param_function(token,   shooting_set_playrec_mode);
      break;
  case TOKENIZER_SET_ND_FILTER:
      one_int_param_function(token,   uB_set_nd_filter_state );
      break;
  case TOKENIZER_SET_CAPTURE_MODE:
      one_int_param_function(token, (void (*)(int))shooting_set_mode_chdk);
      break;
  case TOKENIZER_SET_CAPTURE_MODE_CANON:
      set_capture_mode_canon_statement();
      break;
  case TOKENIZER_SET_CANON_IMAGE_FORMAT:
      one_int_param_function(token, (void (*)(int)) shooting_set_canon_image_format);
      break;
/*--------------------v-------------------------*/
  case TOKENIZER_EXIT_ALT:
      switch_function(token, exit_alt);
      break;
  case TOKENIZER_ENTER_ALT:
      accept(token);
      enter_alt(1);
      accept_cr();
      break;
/*--------------------^-------------------------*/ 

      // >> mx3 . motion detector
  case   TOKENIZER_MD_DETECT_MOTION:
      md_detect_motion_statement();
      break;
  case  TOKENIZER_MD_GET_CELL_DIFF:
      md_get_cell_diff_statement();
      break;
  case  TOKENIZER_MD_GET_CELL_VAL:
      md_get_cell_val_statement();
      break;
  case TOKENIZER_MD_AF_LED_CONTROL:{
      accept(token);
      camera_info.perf.md_af_on_delay = expr();
	  accept(-TOKENIZER_COMMA);
      camera_info.perf.md_af_on_time = expr();
      if ((camera_info.perf.md_af_on_time > 0) && (camera_info.cam_af_led >= 0)) {
          camera_info.perf.md_af_tuning = 1;        // Enable MD testing with AF LED
          // clear previous values
          memset(&camera_info.perf.af_led,0,sizeof(camera_info.perf.af_led));
          camera_info.perf.af_led.min = 0xFFFFFFFF;
      } else {
          camera_info.perf.md_af_tuning = 0;        // Disable MD testing
      }
      accept_cr();
      break;}
      // << mx3 . motion_detector

/*
  case TOKENIZER_SET_ISO_MARKET:
      one_int_param_function(token, shooting_set_iso_market);
      break;
  case TOKENIZER_SET_ISO_DL_F_B:
      one_int_param_function(token, shooting_set_iso_real_delta_from_base);
      break;
*/  
  //ARM End
  
  case TOKENIZER_SET_PROP: set_prop_statement();
      break;
  case TOKENIZER_SET_LED:  set_led_statement();
      break;

  case TOKENIZER_SET_MOVIE_STATUS:
      set_movie_status_statement();
      break;
  case TOKENIZER_SET_RESOLUTION:
      set_propcase_statement(token, camera_info.props.resolution);
      break;
  case TOKENIZER_SET_QUALITY:
      set_propcase_statement(token, camera_info.props.quality);
      break;
  case TOKENIZER_SET_METERING_MODE:
 	  set_propcase_statement(token, camera_info.props.metering_mode);
      break;
  case TOKENIZER_SET_AF_FRAME: //?? 
	  accept(TOKENIZER_SET_AF_FRAME);
	  l=expr(); // %3
      if (camera_info.props.propset==1) shooting_set_prop(64,l);
	  else shooting_set_prop(8,l);
//      shooting_set_prop(camera_info.props.af_frame, l); // 1 = FlexiZone, 2 = Face AiAF / Tracking AF); NO change!
      accept_cr();
    break;

  case TOKENIZER_SET_RAW: accept(token);
      conf.save_raw = expr();
      accept_cr();
      break;
  case TOKENIZER_SET_RAW_NR: accept(token);
      conf.raw_nr = expr();
      accept_cr();
      break;
  case TOKENIZER_SET_SCRIPT_AUTOSTART:
      set_autostart_statement();
      break;
      
  case TOKENIZER_GET_CONFIG_VALUE:
    get_config_value_statement();
    break;

  case TOKENIZER_SET_CONFIG_VALUE:
    set_config_value_statement();
    break;
  case TOKENIZER_SET_YIELD:
    set_yield_statement();
    break;

  case TOKENIZER_USB_SYNC_WAIT:
    accept(TOKENIZER_USB_SYNC_WAIT);
#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    if (expr()) usb_sync_wait_flag = 1;
    else        usb_sync_wait_flag = 0;
#endif
    accept_cr();
    break;

  case TOKENIZER_SET_REMOTE_TIMING:
    accept(TOKENIZER_SET_REMOTE_TIMING);
#ifdef UBASIC_TEST
    // TODO: add something here for ubasic_test
#else
    int hpenable= expr();
    if ( hpenable > 0) start_usb_HPtimer(hpenable);
    else stop_usb_HPtimer();
#endif
    accept_cr();
    break;

  case TOKENIZER_SET_EXIT_KEY:
      set_exit_key_statement();
      break;

  case TOKENIZER_WHEEL_LEFT:
      accept(token);
      JogDial_CCW();
      accept_cr();
      break;
  case TOKENIZER_WHEEL_RIGHT:
      accept(token);
      JogDial_CW();
      accept_cr();
      break;

  case TOKENIZER_CLS: cls_statement();
      break;
  case TOKENIZER_SET_CONSOLE_LAYOUT:
      set_console_layout();
      break;
  case TOKENIZER_SET_CONSOLE_AUTOREDRAW:
      set_console_autoredraw();
      break;
  case TOKENIZER_CONSOLE_REDRAW:
      console_redraw_statement();
      break;
  case TOKENIZER_SET_DRAW_TITLE_LINE:
      accept(token);
      camera_info.state.osd_title_line=expr()?1:0;
      accept_cr();
      break;
  case TOKENIZER_SET_BACKLIGHT:
      on_off_statement(token, TurnOnBackLight, TurnOffBackLight);
      break;
  case TOKENIZER_SET_LCDDISPLAY:
      on_off_statement(token, TurnOnDisplay, TurnOffDisplay);
      break;
  case TOKENIZER_PRINT_SCREEN:
      l=one_int_param_function(token, script_print_screen_statement);
	  if (l) {logfile=l;
	  sprintf(writefile, "A/CHDK/LOGS/Log_%04d.Log", abs(logfile));}
	  flag_yield=1;
      break;
  case TOKENIZER_PLAY_SOUND:
      one_int_param_function(token, (void (*)(int))play_sound);
      break;
  case TOKENIZER_BEEP:
	  {play_sound(4); accept(TOKENIZER_BEEP);}
	  break;
  case TOKENIZER_SLEEP: sleep_statement();
      break;
  case TOKENIZER_SWAP: swap_statement();
      break;
  case TOKENIZER_LIST:
//	  accept(TOKENIZER_LIST);
      list_statement();
	  break;
  case TOKENIZER_VARLIST:
		accept(TOKENIZER_VARLIST);
		if (is_token(TOKENIZER_STRING)) list_statement();
      varlist();
      break;
  case TOKENIZER_POKE: poke_statement();
      break;
  case TOKENIZER_MEM_DUMP: // can crash
		accept(TOKENIZER_MEM_DUMP);
		mem_dump_statement();
      break;
	  
  case TOKENIZER_LOOP: loop_statement();
	  break;
  case TOKENIZER_TRONLY:
	tokenizer_next(); trace=on; traconly=on; break;
  case TOKENIZER_TRON:
	tokenizer_next(); trace=on; traconly=off;break;
  case TOKENIZER_TROFF: trace=off; traconly=off;
	tokenizer_next();
      break;
  case TOKENIZER_OPEN: factor();
	  break;
  case TOKENIZER_CLOSE:
    if (readfile!=NULL) fclose(readfile);
    readfile=NULL;
    accept_cr(); break;
  case TOKENIZER_FILE_BROWSER:
    accept(TOKENIZER_FILE_BROWSER); break;
  case TOKENIZER_RESET: 
	accept(TOKENIZER_RESET);  data_statement();
	  ended=-1;
      break;
  case TOKENIZER_REBOOT:      reboot_statement();
      break;
  case TOKENIZER_SHUT_DOWN:
      shutdown_statement();
  case TOKENIZER_ENDOFINPUT:  ended=1; traconly=off;
	break;
  case TOKENIZER_EXCEPTION: // Romlog on demand? 20-60! secs later
	ended=0; while (ended<1); {ended++;}
  case TOKENIZER_DEBUG:{
    accept(TOKENIZER_DEBUG);
    accept(-TOKENIZER_EQ);
      if (eos()) debug=on; else debug=expr();
	  accept_cr();
	  break;}
  case TOKENIZER_TEST:
    accept(TOKENIZER_TEST);	test_statement();
      break;
  case TOKENIZER_AVERAGE:  {stopwatch(expr());}
	  break;
  case TOKENIZER_CLEAN:    clean_statement();
	  break;
  case TOKENIZER_MEMORY:
	accept(TOKENIZER_MEMORY); garbage_collect();
	  break;
  case TOKENIZER_BEGIN:    tokenizer_begin();
	reset();
      break;
  case TOKENIZER_BUG:
    accept(TOKENIZER_CR); ended=3;
      break;
  case TOKENIZER_END: end_statement();
      break;
  case TOKENIZER_SEMICOLON: tokenizer_next();
    stopwatch(-100000-stm_ptr);
      break;
//case TOKENIZER_REM: tokenizer_eol(); break;
	   
  default: {
    stopwatch(1000000+stm_ptr);
    if (stm_ctr<3 && !is_eof()) {reader_init();}
    if (traconly>9) return; ///{ stopwatch(ubasic_linenumber());}

      DEBUG_PRINTF("ubasic.c: statement(): not implemented %d\n", token);
      ended = 3;  // Letters eaten in let_statement
      print_statement(TOKENIZER_PRINT);
      ubasic_error = UBASIC_E_UNK_STATEMENT;
    }
  }// break>>
  l=tokenizer_token();
  if (l==TOKENIZER_ELSE || l==TOKENIZER_ELSEIF) accept_cr(); /// eat singleline if ...[exec {...}]
  
  sec=(get_tick_count()-first_tick)/1000;
}

/*---------------------------------------------------------------------------*/
static void line_statement(void)
{
  /* line numbers have been removed */
  DEBUG_PRINTF("----------- Line number %d ---------\n", tokenizer_line_number());
  /*    current_linenum = tokenizer_num();*/
#if 0
  if (is_token(TOKENIZER_LABEL)) {
#ifdef DEBUG
  if (traconly<9)      tokenizer_label(string, sizeof(string));
      DEBUG_PRINTF("line_statement: label: %s\n", string );
#endif
      //accept(TOKENIZER_LABEL);

  if (traconly>9) { tokenizer_eol(); accept(0);
      return; }
  }
#endif 
  /* reyalp - eat up to 100 labels or rems at a time so they don't cost 10ms each */
  /* Caefix - 10ms are a hollyday for a Canon, lets use that! */
  short count = 100;

  if (traconly>9) { // ubasic_reader
     if (!camera_info.state.gui_mode) { 
         pause=-2; return;}
     if (pause==-2 && camera_info.state.gui_mode) { enter_alt(1);
         kbd_get_autoclicked_key();
         console_redraw(0); flag_yield=1; pause=-1; return;}
  }
  else do {
    short r =  tokenizer_token();
  //  if (ended) break;
    if ( r == TOKENIZER_LABEL ) {
      /* hit limit and we are on a label, return */
      if( count == 1 ) return;
#ifdef DEBUG
      tokenizer_label(string, sizeof(string));
      DEBUG_PRINTF("line_statement: label: %s\n", string );
#endif
	  tokenizer_eol();
    }
	else if ( r == TOKENIZER_NUMBER)  tokenizer_next();//ignore: numeric label for goto|gosub 10*x
    else if ( r == TOKENIZER_MINUS) { tokenizer_next();
		 if ( is_token(TOKENIZER_MINUS)) tokenizer_eol();} // skip Lua-commentline 
    r =  tokenizer_token();
    if ( r == TOKENIZER_VARIABLE || // # & Lets nice here!
         r == TOKENIZER_STRINGVARIABLE || 
		 r == TOKENIZER_LET) { let_statement(); }
    else if ( r == TOKENIZER_DEF_FN ) {tokenizer_def_fn();} 
	// lets eat what we dont need on 1. Position.
    else if ( r == TOKENIZER_SLASH )  {tokenizer_eol();} // skip C-commentline
    else if ( r == TOKENIZER_STRING ) {list_statement(); tokenizer_eol();} // show "commentline"
	else if ( r == TOKENIZER_SEP )    {tokenizer_next(); break;} // no Yield
    else if ( r == TOKENIZER_ASTR ) { accept(0); int x=0;
		switch(tokenizer_token()) {  // '*' 
			case TOKENIZER_SEP: tokenizer_next(); break;
			case TOKENIZER_MOMENT: ended=1; return;
			case TOKENIZER_LABEL: accept(0); get_time(7); break;
			case TOKENIZER_PLUS: accept(0); get_time(42); break;
			case TOKENIZER_LNOT: accept(0); traconly=off; break;
			case TOKENIZER_GT: { accept(0);
			   x=tokenizer_token(); accept(0);
				 tokenizer_skip(x); break;} // Crazy Jump; be careful!
			case TOKENIZER_LEFTPAREN: { // *(
			  do { tokenizer_eol(); accept(0); // ignore multilineblock *( <# # #> \n*)
				 if (is_token(TOKENIZER_ASTR)) {accept(0);
				 x=is_token(TOKENIZER_RIGHTPAREN);}
				} while (!(x|ended));
				if (x) tokenizer_sep(); } // *)
				
			default: {tokenizer_eol();}
		}
	} // if '*' 
	else if ( r ==TOKENIZER_LEFTPAREN) {
		tokenizer_skip(TOKENIZER_RIGHTPAREN);
		accept(TOKENIZER_RIGHTPAREN);}
	else if ( r ==TOKENIZER_RIGHTPAREN) {tokenizer_next();}
	else if ( r == TOKENIZER_ENDOFINPUT) {return;} // -> ubasic_run
	else {	
		if (is_cr()) { accept(0); continue; }
		break; }
  } while(--count);
  
 do { statement();	if (flag_yield|ended) break;
	} while (is_token(TOKENIZER_SEP));
	
 if (debug) {
	tokenizer_statement(string,120);
    sprintf(buf, "@%d: '%s'",ubasic_linenumber(), string);
    script_console_add_error((long)buf);
	}
  sec=(get_tick_count()-first_tick)/1000;

  // ubasic_reader
  if (traconly>9) {
	 if (pause>0) { pause--;
	    if (!is_eof()) {list_statement(); ended=0;}
     }
  } /**/
  return;
}
/*---------------------------------------------------------------------------*/
int ubasic_run(void)
{
	unsigned short li,i,lines = 0;
    flag_yield = 0;
    start_tick = get_tick_count();
	li=ubasic_linenumber(); // catch jumpline for error
 //   subj_dist=shooting_get_subject_distance();
      if (debug>2) {beep(); stopwatch(li); msleep(debug*100);}
  
    do
    {     
        if ( ended )
        {
		//  Generate [error] message
            char buf[123];
            const char *msg;
            DEBUG_PRINTF("%ds uBASIC program finished\n",sec);
			i = (start_sd-GetFreeCardSpaceKb())/1024;
			sprintf(buf, "-%dMb %d @%d Log %d",i, GetFreeCardSpaceKb()/1024, stm_ctr,logfile);
			script_console_add_error((long)buf);
			get_time(8);
			if ( ended==-1) {reset(); break;}
            if (readfile!=NULL) fclose(readfile);
            if (ubasic_error)
            {
                if (ubasic_error >= UBASIC_E_ENDMARK)
                {
                    msg = ubasic_errstrings[UBASIC_E_UNKNOWN_ERROR];
                }
                else
                {
                    msg = ubasic_errstrings[ubasic_error];
                }
                sprintf(buf, "%d:%d uBASIC:%s", ubasic_linenumber(), tokenizer_token(), msg);
                // Show error message
                script_console_add_error((long)buf);
				
                // Show error line, Caefix
				tokenizer_statement(string,50);
				sprintf(buf, "'%s'", string);
                script_console_add_error((long)string);
			//	if (ended==3)
				li = ubasic_linenumber();
				jump_linenum(li);
				list_statement();
			script_console_add_error(LANG_CONSOLE_TEXT_TERMINATED); buf[0]='\0';
				for (i=0; i<=5; i++) moment_statement(990);
				if (loop) {reset(); return SCRIPT_RUN_RUNNING;}
			console_set_layout(0,0,25,5);
            return SCRIPT_RUN_ERROR;
            }
            // Show 'Finished' message//
			sprintf(buf, "%d:%d uBASIC: %04d",
              ubasic_linenumber(), tokenizer_token(), get_exposure_counter());
			script_console_add_error((long)buf);
            script_console_add_error(LANG_CONSOLE_TEXT_FINISHED);
			if (loop) {reset();
			for (i=0; i<=end_wait; i++) {moment_statement(990);}
			return SCRIPT_RUN_RUNNING;}
			console_set_layout(0,0,25,5);
            return SCRIPT_RUN_ENDED;
        }
		while (is_cr())  {tokenizer_next();}

        if (svar_selected>0) { // textbox & file_browser 
            ubasic_set_stringvariable(svar_selected-1,store(string_selected));
            svar_selected=0;}
        line_statement();

		if (traconly==TOKENIZER_UBASIC_READER) { go_fwd=1;
            pause=ubasic_linenumber();
            if (!jump_label("BEGIN")) { tokenizer_init(program_ptr);
            if (!jump_label("BEGIN")) { jump_line(pause-2);
               stopwatch(-ubasic_linenumber());}
            }
		    traconly=pause=ended=0; break;
        }

        if ( flag_yield && !ended) break; // eat eol here.
		if (is_cr() || ended) continue;

        // Return control to CHDK only if external processing required  
		// && to get updated values like get_tick_count ? //Caefix
        if ( flag_yield )
            return SCRIPT_RUN_RUNNING;

        lines++;
    } while (lines < yield_max_lines && get_tick_count() - start_tick < yield_max_ms);
	lsec=(get_tick_count()-first_tick)/1000-lsec;
    return SCRIPT_RUN_RUNNING;
}
/*---------------------------------------------------------------------------* /
static long ubasic_get_variable(short varnum)
{
  if (varnum >= 0 && varnum < MAX_VARNUM) 
    return variables[varnum];

  return 0;
}
/ *----- 4051 -----------------------------------------------------------------*/
// read a key name and return key id, 
// set error and return 0 if key invalid
static short ubasic_get_key_arg()
{
  short k, par=0;
  if (is_paren()) {par=1; tokenizer_next();}
  if (is_token(TOKENIZER_STRING)) {
     tokenizer_string(string, sizeof(string));
     accept(TOKENIZER_STRING);
     if (strncmp(string,".", 2)==0) {k=mem_key;}
     else {k=script_keyid_by_name(string); mem_key=k;}
  } else k=expr();  // NO errortest!
  if (par) accept(TOKENIZER_RIGHTPAREN);
  if (k <= 0) ubasic_error = UBASIC_E_UNK_KEY;
  return k;
}
/*---------------------------------------------------------------------------*/

// Save Action Stack 'return' value in selected variable
void ubasic_set_as_ret(int md_ret);
inline void ubasic_set_as_ret(int md_ret)
{
 //   if (ubasic_as_ret_var_num >= 0)  // Letters only
        ubasic_set_variable(ubasic_as_ret_var_num, md_ret);
}
/*---------------------------------------------------------------------------*/
void ubasic_end(void)
{
    reset();
	svar_init();
	tokenizer_end();
}
// https://app.assembla.com/spaces/chdk/subversion/commits/5792

/* following .bas script starts with same variables[], clean @() field, null $ .
/*----- 5135 -------------- 6087 ---------------------------------------------*/